#include <linux/types.h>
#include <linux/ip.h>
#include <linux/timer.h>
#include <linux/modreglist.h>
#include <linux/netfilter.h>
#include <linux/ip_netfilter.h>
#include <linux/module.h>
#include <linux/kmod.h>
#include <net/protocol.h>
#include <net/checksum.h>
#include <linux/bitops.h>
#include <linux/proc_fs.h>
#include <asm/spinlock.h>

#include "ip_nat.h"
#include "ip_nat_perconn.h"
#include "ip_nat_setup.h"

#ifdef DEBUG
#define DEBUGP printk
#else
#define DEBUGP(format, args...)
#endif

/* Regarding locking: we use a bit spinlock for the whole lot.  We can
 * improve this by using similar techniques to modreglist.c. */
spinlock_t ip_nat_lock = SPIN_LOCK_UNLOCKED;

/* From ip_nat_perconn.c */
static struct ip_nat_perconn *
find_perconn(const struct ip_nat_tuple_proto *tuplepr, 
	     enum ip_nat_direction direction,
	     long now, int unexpired_only);

/* From ip_nat_protos.c */
static struct ip_nat_proto *find_proto(unsigned int protocol);

/* From ip_nat_setup.c */
static unsigned int ip_nat_do_setup(const struct iphdr *iph,
				    enum ip_nat_direction direction,
				    const struct ip_nat_pktinfo *pkt,
				    struct ip_nat_perconn **perconn);

/* Setup types. */
static struct ip_nat_setup_type *setup_types;
/* One setup table for each direction. */
static struct ip_nat_setup *setup_table[2];

/* Hash table of registered protocols */
static struct ip_nat_proto *ip_nat_prototable[MAX_INET_PROTOS];

/* Five hash tables for per_connection information.  Every struct
 * ip_nat_perconn is in all of them. */

/* For tuple lookups; one each direction. */
static struct ip_nat_perconn *perconn_by_tuple[2][IP_NAT_HTABLE_SIZE];
/* These both all refer to IP_NAT_OUTGOING tuples. */
/* For finding least-used srcip/dstip/proto set... */
static struct ip_nat_perconn *perconn_by_ips_proto[IP_NAT_HTABLE_SIZE];
/* For finding other mappings of this srcip/proto/per-proto-src... */
static struct ip_nat_perconn *perconn_by_source[IP_NAT_HTABLE_SIZE];

static struct ip_nat_perconn **
hash_by_tuple(const struct ip_nat_tuple_proto *tuplepr,
	      enum ip_nat_direction dir)
{
	return &perconn_by_tuple[dir][csum_partial((char *)tuplepr, 
						   sizeof(*tuplepr), 0) 
				     % IP_NAT_HTABLE_SIZE];
}

static struct ip_nat_perconn **
hash_by_ips_proto(u_int32_t src, u_int32_t dst, u_int8_t proto)
{
	return &perconn_by_ips_proto[(src + dst + proto) % IP_NAT_HTABLE_SIZE];
}

static struct ip_nat_perconn **
hash_by_source(const struct ip_nat_tuple_proto *tuplepr)
{
	return &perconn_by_source[csum_partial((char *)&tuplepr->tuple.src, 
					       sizeof(tuplepr->tuple.src),
					       tuplepr->proto)
				 % IP_NAT_HTABLE_SIZE];
}

static int ip_nat_in_range(const struct ip_nat_tuple_proto *tuplepr,
			   const struct ip_nat_address_range *source_range)
{
	struct ip_nat_proto *proto = find_proto(tuplepr->proto);

	for (; source_range; source_range = source_range->next) {
		if (ntohl(tuplepr->tuple.src.ip) >= ntohl(source_range->ip_min)
		    && (ntohl(tuplepr->tuple.src.ip) 
			<= ntohl(source_range->ip_max))
		    && proto->in_range(&tuplepr->tuple, source_range))
			return 1;
	}
	return 0;
}

static int
ip_nat_find_src(struct ip_nat_tuple_proto *tuplepr,
		const struct ip_nat_address_range *source_range)
{
	static struct ip_nat_perconn *perconn;

	for (perconn = *hash_by_source(tuplepr); 
	     perconn; 
	     perconn = perconn->by_source) {
		const struct ip_nat_tuple_proto *conn 
			= &perconn->dir[IP_NAT_OUTGOING].connection;
		if (conn->proto == tuplepr->proto
		    && memcmp(&tuplepr->tuple.src, &conn->tuple.src,
			      sizeof(tuplepr->tuple.src)) == 0
		    && ip_nat_in_range(conn, source_range)) {
			tuplepr->tuple.src
				= perconn->dir[IP_NAT_OUTGOING].manip;
			return 1;
		}
	}
	return 0;
}

/* For fragmentation handling, we want the least-used
   src-ip/dst-ip/proto triple.  Fairness doesn't come into it.  Thus
   if the range specifies 1.2.3.4 ports 10000-10005 and 1.2.3.5 ports
   1-65535, we don't do pro-rata allocation based on ports; we choose
   the source ip with the lowest src-ip/dst-ip/proto usage.

   If an allocation then fails (eg. all 6 ports used in the 1.2.3.4
   range), we eliminate that and try again.  This is not the most
   efficient approach, but if you're worried about that, don't hand us
   ranges you don't really have.
*/
static struct ip_nat_address_range *
ip_nat_find_best_ips_proto(struct ip_nat_tuple_proto *tuplepr,
			   const struct ip_nat_address_range *source_range)
			   
{
	const struct ip_nat_address_range *srcptr;
	unsigned int best_score = 0xFFFFFFFF;
	u_int32_t best_src = 0;
	const struct ip_nat_address_range *best_range = NULL;

	for (srcptr = source_range; srcptr; srcptr = srcptr->next) {
		u_int32_t srcip;

		/* Don't do ranges which are already eliminated. */
		if (srcptr->range_full) continue;

		DEBUGP("ip_nat_find_best: %u.%u.%u.%u - %u.%u.%u.%u\n",
		       IP_PARTS_NATIVE(srcptr->ip_min),
		       IP_PARTS_NATIVE(srcptr->ip_max));
		for (srcip = srcptr->ip_min; 
		     srcip <= srcptr->ip_max; 
		     srcip++){
			static struct ip_nat_perconn *perconn;
			unsigned int score = 0;

			DEBUGP("ips_proto_hash seeking src=%u.%u.%u.%u "
			       "dst=%u.%u.%u.%u proto=%u\n",
			       IP_PARTS_NATIVE(srcip), 
			       IP_PARTS(tuplepr->tuple.dst.ip),
			       tuplepr->proto);
			
			for (perconn= *hash_by_ips_proto(htonl(srcip), 
							 tuplepr->tuple.dst.ip,
							 tuplepr->proto);
			     perconn; 
			     perconn = perconn->by_ips_proto) {
				if (perconn->dir[IP_NAT_OUTGOING].manip.ip
				    == htonl(srcip)
				    && (perconn->dir[IP_NAT_OUTGOING]
					.connection.tuple.dst.ip
					== tuplepr->tuple.dst.ip)
				    && (perconn->dir[IP_NAT_OUTGOING]
					.connection.proto == tuplepr->proto))
					score++;
			}

			DEBUGP("ip_nat_find_best: score for %u.%u.%u.%u: %u\n",
			       IP_PARTS_NATIVE(srcip), score);
			if (score < best_score) {
				DEBUGP("ip_nat_find_best_ips_proto: score %u "
				       "beats %u srcip=%u.%u.%u.%u\n",
				       score, best_score, 
				       IP_PARTS_NATIVE(srcip));
				best_score = score;
				best_src = htonl(srcip);
				best_range = srcptr;
				if (score == 0)
					break;
			}
		}
	}

	tuplepr->tuple.src.ip = best_src;
	return (struct ip_nat_address_range *)best_range;
}

static int
get_unique_tuple(struct ip_nat_tuple_proto *tuplepr, 
		 struct ip_nat_address_range *source_range)
{
	struct ip_nat_tuple_proto tmp_tpr;
	struct ip_nat_address_range *source_ptr;

	/* 1) If this srcip/proto/src-proto-part is currently mapped,
	   and that same mapping gives a unique tuple within the given
	   range, use that.  */
	tmp_tpr = *tuplepr;
	if (ip_nat_find_src(&tmp_tpr, source_range)
	    && !find_perconn(&tmp_tpr, IP_NAT_OUTGOING, jiffies, 0)) {
		DEBUGP("get_unique_tuple: Found current mapping\n");
		*tuplepr = tmp_tpr;
		return 1;
	}

	/* 2) Select the least-used IP/proto combination in the given
	   range.
	*/
	while ((source_ptr = ip_nat_find_best_ips_proto(tuplepr, source_range))
	       != NULL) {
		/* 3) The per-protocol part of the tuple is mapped into the
		   range to make a unique tuple.
		*/
		if (find_proto(tuplepr->proto)->unique_tuple(tuplepr, 
							     source_ptr))
			return 1;

		/* Eliminate that from range, and try again. */
		source_ptr->range_full = 1;
	}

	return 0;
}

static void ip_nat_invert_tuple_proto(struct ip_nat_tuple_proto *tuplepr)
{
	u_int32_t tmp = tuplepr->tuple.src.ip;
	
	tuplepr->tuple.src.ip = tuplepr->tuple.dst.ip;
	tuplepr->tuple.dst.ip = tmp;

	find_proto(tuplepr->proto)->invert_tuple(&tuplepr->tuple);
}

/* Check for tuple existence */
int ip_nat_used_tuple(const struct ip_nat_tuple_proto *tuplepr)
{
	struct ip_nat_tuple_proto inv = *tuplepr;
	ip_nat_invert_tuple_proto(&inv);
	return (int)find_perconn(&inv, IP_NAT_INCOMING, jiffies, 0);
}

static int get_tuple_proto(const struct iphdr *iph, 
			   size_t len, 
			   struct ip_nat_tuple_proto *tuplepr)
{
	IP_NAT_TUPLE_PROT_INIT(tuplepr, iph->saddr, iph->daddr, iph->protocol);

	return find_proto(tuplepr->proto)
		->pkt_to_tuple((u_int32_t *)iph + iph->ihl, 
			       len - 4*iph->ihl, &tuplepr->tuple);
}

/* We do checksum mangling, so if they were wrong before they're still
 * wrong.  Also works for incomplete packets (eg. ICMP dest
 * unreachables.) */
u_int16_t ip_nat_cheat_check(u_int32_t oldvalinv, u_int32_t newval,
			     u_int16_t oldcheck)
{
	u_int32_t diffs[] = { oldvalinv, newval };
	return csum_fold(csum_partial((char *)diffs, sizeof(diffs), 
				      oldcheck^0xFFFF));
}

static void ip_nat_manip_pkt(struct iphdr *iph, size_t len, 
			     const struct ip_nat_tuple_manip *manip,
			     enum ip_nat_direction direction)
{
	find_proto(iph->protocol)->manip_pkt(iph, len, manip, direction);

	if (direction == IP_NAT_OUTGOING) {
		/* Replace source ip */
		DEBUGP("ip_nat_manip_pkt: Replacing src ip %u.%u.%u.%u with "
		       "%u.%u.%u.%u\n", IP_PARTS(iph->saddr), 
		       IP_PARTS(manip->ip));

		iph->check = ip_nat_cheat_check(~iph->saddr, manip->ip,
						iph->check);
		iph->saddr = manip->ip;
	} else {
		DEBUGP("ip_nat_manip_pkt: Replacing dst ip %u.%u.%u.%u with "
		       "%u.%u.%u.%u\n", IP_PARTS(iph->daddr), 
		       IP_PARTS(manip->ip));

		/* Replace destination ip */
		iph->check = ip_nat_cheat_check(~iph->daddr, manip->ip,
						iph->check);
		iph->daddr = manip->ip;
	}
}

static int icmp_reply_translation(struct iphdr *iph, 
				  size_t len,
				  enum ip_nat_direction direction)
{
	struct icmphdr *hdr = (struct icmphdr *)((u_int32_t *)iph + iph->ihl);
	size_t datalen = len - iph->ihl*4 - sizeof(*hdr);
	struct ip_nat_perconn *perconn;

	if (hdr->type == ICMP_DEST_UNREACH
	    || hdr->type == ICMP_SOURCE_QUENCH
	    || hdr->type == ICMP_TIME_EXCEEDED
	    || hdr->type == ICMP_PARAMETERPROB
	    || hdr->type == ICMP_REDIRECT) {
		/* Are they talking about one of our connections? */
		struct ip_nat_tuple_proto innertuplepr;
		struct iphdr *inner = (struct iphdr *)(hdr + 1);

		if (!get_tuple_proto(inner, datalen, &innertuplepr)) {
			DEBUGP("icmp_reply_translation: Failed to get tuple\n");
			return 0;
		}

		/* Ordinarily, we'd expect the inverted tupleproto,
		   but it's been preserved inside the ICMP. */
		ip_nat_invert_tuple_proto(&innertuplepr);
		perconn = find_perconn(&innertuplepr, direction, jiffies, 1);
		if (perconn) {
			struct ip_nat_tuple_manip manip;

			DEBUGP("icmp_reply_translation: Found perconn\n");
			/* Map inner packet first; see NAT.txt. */
			ip_nat_manip_pkt(inner, datalen, 
					 &perconn->dir[direction].manip,
					 !direction);

			/* Now outer packet; NULL ICMP translation.*/
			manip.u.icmp.id = hdr->un.echo.id;
			if (direction == IP_NAT_INCOMING) {
				/* If direction == INCOMING, replace dest. */
				manip.ip = perconn->dir[!direction]
					.connection.tuple.src.ip;
			} else {
				/* Replace source.. */
				manip.ip = perconn->dir[!direction]
					.connection.tuple.dst.ip;
			}
			ip_nat_manip_pkt(iph, len, &manip, direction);
			perconn->relatedpacket(perconn, iph, direction);
			return 1;
		}
		else DEBUGP("icmp_reply_translation: no match\n");
	}
	return 0;
}

/* Helper for masquerading: set range to match appropriate interface */
int ip_nat_masq_set(const struct iphdr *iph, 
		    struct ip_nat_address_range *range)
{
	struct rtable *rt;

	/* Pass 0 instead of saddr, since it's going to be changed anyway. */
	/* FIXME: Maybe pass dev->ifindex (or dev->iflink?) instead of
           0? --RR */
	if (ip_route_output(&rt, iph->daddr, 0, IPTOS_TOS(iph->tos), 0) != 0)
		return 0;

	range->ip_min
		= range->ip_max 
		= inet_select_addr(rt->u.dst.dev, rt->rt_gateway, 
				   RT_SCOPE_UNIVERSE);
	range->range_full = 0;

	ip_rt_put(rt);
	return 1;
}

/* Helper for redirect: make it head for that interface. */
int ip_nat_interface(const struct device *dev,
		     struct ip_nat_address_range *range)
{
	/* Grab first address on interface. */
	range->ip_max = range->ip_min 
		=ntohl(((struct in_device *)dev->ip_ptr)->ifa_list->ifa_local);
	range->range_full = 0;
	return 1;
}

static unsigned int ipnat(struct sk_buff *skb,
			  const struct device *device,
			  enum ip_nat_direction direction)
{
	struct iphdr *iph = skb->nh.iph;
	struct ip_nat_pktinfo pkt;
	struct ip_nat_perconn *perconn;
	unsigned int ret;
	unsigned long flags;

	DEBUGP("ipnat: Looking for direction %s.\n",
	       direction == IP_NAT_OUTGOING ? "OUT" : "IN");
	pkt.device = device;
	/* FIXME: One day, fill in properly. --RR */
	skb->nfcache |= NFC_UNKNOWN;

	spin_lock_irqsave(&ip_nat_lock, flags);
	if (!get_tuple_proto(iph, skb->len, &pkt.tuplepr)) {
		DEBUGP("ipnat: Can't get tuple.\n");
		spin_unlock_irqrestore(&ip_nat_lock, flags);
		return NF_ACCEPT;
	}

	/* ICMP replies tied to other protocols are handled separately: 
	 * don't clutter perconn table for such rare occurrances. */
	if (iph->protocol == IPPROTO_ICMP) {
		DEBUGP("ipnat: Trying ICMP reply translation.\n");
		if (icmp_reply_translation(iph, skb->len, direction)) {
			skb->csum = iph->check;
			spin_unlock_irqrestore(&ip_nat_lock, flags);
			skb->nfcache |= NFC_ALTERED;
			return NF_ACCEPT;
		}
	}

	/* First, seek unexpired per-connection match. */
	perconn = find_perconn(&pkt.tuplepr, direction, jiffies, 1);
	if (perconn) {
		DEBUGP("ipnat: Found perconn.\n");
		goto got_perconn;
	}

	/* Look for setup match. */
	switch (ip_nat_do_setup(iph, direction, &pkt, &perconn)) {
	default:
		if (perconn) {
			DEBUGP("ipnat: Setup gave a perconn.\n");
			goto got_perconn;
		} else {
			DEBUGP("ipnat: No setup.\n");
			spin_unlock_irqrestore(&ip_nat_lock, flags);
			return NF_ACCEPT;
		}

	case NF_DROP:
		DEBUGP("ipnat: Rejected by setup.\n");
		spin_unlock_irqrestore(&ip_nat_lock, flags);
		return NF_DROP;

	case NF_USERSPACE:
		DEBUGP("ipnat: USERSPACE by setup.\n");
		spin_unlock_irqrestore(&ip_nat_lock, flags);
		skb->nfreason = NF_SPECIAL_SET_BY_NAT;
		return NF_USERSPACE;
	}

got_perconn:
	ip_nat_manip_pkt(iph, skb->len, &perconn->dir[direction].manip,
			 direction);
	perconn->dir[direction].hitcount++;
	perconn->dir[direction].lasthit = jiffies;
	ret = perconn->postmangle(perconn, iph, direction);
	skb->csum = iph->check;

	spin_unlock_irqrestore(&ip_nat_lock, flags);

	skb->nfcache |= NFC_ALTERED;
	return ret;
}

static unsigned int ipnat_infn(unsigned int hooknum,
			       struct sk_buff **pskb,
			       const struct device *in,
			       const struct device *out)
{
	DEBUGP("ipnat_infn\n");
	return ipnat(*pskb, in, IP_NAT_INCOMING);
}

/* FIXME: Think about choice of interface here. --RR */
static unsigned int ipnat_localfn(unsigned int hooknum,
				  struct sk_buff **pskb,
				  const struct device *in,
				  const struct device *out)
{
	DEBUGP("ipnat_localfn\n");
	/* in will be NULL for local packets. */
	return ipnat(*pskb, out, IP_NAT_INCOMING);
}

static unsigned int ipnat_outfn(unsigned int hooknum,
				struct sk_buff **pskb,
				const struct device *in,
				const struct device *out)
{
	DEBUGP("ipnat_outfn\n");
	return ipnat(*pskb, out, IP_NAT_OUTGOING);
}

static struct nf_hook_ops ipnat_in_ops = { NULL, 0, ipnat_infn, NULL, 
					   PF_INET, NF_IP_PRE_ROUTING };
static struct nf_hook_ops ipnat_local_out_ops = { NULL, 0, ipnat_localfn, NULL,
						 PF_INET, NF_IP_OUTPUT };
static struct nf_hook_ops ipnat_out_ops = { NULL, 0, ipnat_outfn, NULL, 
					    PF_INET, NF_IP_OUTGOING };

#include "ip_nat_protos.c"
#include "ip_nat_standard_protos.c"
#include "ip_nat_perconn.c"
#include "ip_nat_setup.c"

#ifdef CONFIG_PROC_FS
struct proc_dir_entry ip_nat_conns_procinfo = {
	PROC_NET_NAT_CONNS, 18, "ip_nat_connections", S_IFREG | S_IRUGO,
	1, 0, 0, 0, &proc_net_inode_operations, nat_get_perconns
};
struct proc_dir_entry ip_nat_setup_procinfo = {
	PROC_NET_NAT_SETUPS, 13, "ip_nat_setups", S_IFREG | S_IRUGO,
	1, 0, 0, 0, &proc_net_inode_operations, nat_get_setups
};
#endif

/* We need to know when devices go up and down so we can correct the
   device pointers. */
static int ipnat_device_event(struct notifier_block *this, 
			      unsigned long event, void *voiddev)
{
	struct device *dev = voiddev;

	switch (event) {
	case NETDEV_CHANGENAME:
	case NETDEV_DOWN:
		ipnat_setup_down_device(dev);
		if (event == NETDEV_DOWN)
			break;

		/* Fall thru... iff CHANGENAME. */
	case NETDEV_UP:
		ipnat_setup_up_device(dev, dev->name);
		break;
	}
	return NOTIFY_DONE;
}

static struct notifier_block ipnat_dev_notifier = {
	ipnat_device_event,
	NULL,
	0
};

static int init_or_cleanup(int init)
{
	if (!init) goto cleanup;

#ifdef CONFIG_PROC_FS
	proc_net_register(&ip_nat_setup_procinfo);
	proc_net_register(&ip_nat_conns_procinfo);
#endif
	if (nf_register_hook(&ipnat_in_ops) < 0) {
		printk("ip_nat_init: can't register in hook.\n");
		goto cleanup_nothing;
	}
	if (nf_register_hook(&ipnat_out_ops) < 0) {
		printk("ip_nat_init: can't register out hook.\n");
		goto cleanup_inops;
	}
	if (nf_register_hook(&ipnat_local_out_ops) < 0) {
		printk("ip_nat_init: can't register in hook.\n");
		goto cleanup_outops;
	}
	if (!init_ip_nat_setup()) {
		printk("ip_nat_init: can't setup.\n");
		goto cleanup_localoutops;
	}
	init_standard_protos();	
	register_netdevice_notifier(&ipnat_dev_notifier);

	return 1;

 cleanup:
	cleanup_standard_protos();
	cleanup_ip_nat_setup();
	unregister_netdevice_notifier(&ipnat_dev_notifier);
 cleanup_localoutops:
	nf_unregister_hook(&ipnat_local_out_ops);
 cleanup_outops:
	nf_unregister_hook(&ipnat_out_ops);
 cleanup_inops:
	nf_unregister_hook(&ipnat_in_ops);
 cleanup_nothing:
	return 0;
}

#ifdef MODULE
int init_module(void)
{
	if (!init_or_cleanup(1))
		return -EINVAL;
	else
		return 0;
}

void cleanup_module(void)
{
#ifdef CONFIG_PROC_FS
	proc_net_unregister(PROC_NET_NAT_SETUPS);
	proc_net_unregister(PROC_NET_NAT_CONNS);
#endif
	init_or_cleanup(0);
}
#else
__initfunc(void ip_nat_init(void))
{
	if (!init_or_cleanup(1)) 
		panic("Can't set up ip_nat.\n");
}
#endif /*MODULE*/
