#ifndef _IP_NAT_H
#define _IP_NAT_H

struct sk_buff;
struct device;
struct iphdr;

/* These are the parts of the tuple we can manipulate (for
 * outgoing tuple allocation) */
struct ip_nat_tuple_manip {
	u_int32_t ip;
	union {
		struct {
			u_int16_t port;
		} tcp;
		struct {
			u_int16_t port;
		} udp;
		struct {
			u_int16_t id;
		} icmp;
		/* Add other protocols here. */
		u_int32_t all[0];
	} u;
};

/* This contains the information to distinguish a connection. */
struct ip_nat_tuple
{
	struct ip_nat_tuple_manip src;

	/* These are the parts of the tuple which are fixed for
	 * outgoing tuple allocation. */
	struct {
		u_int32_t ip;
		union {
			struct {
				u_int16_t port;
			} tcp;
			struct {
				u_int16_t port;
			} udp;
			struct {
				u_int8_t type, code;
			} icmp;
			/* Add other protocols here. */
			u_int32_t all[0];
		} u;
	} dst;
};

struct ip_nat_tuple_proto
{
	u_int32_t proto;
	struct ip_nat_tuple tuple;
};

#define IP_PARTS_NATIVE(n)			\
(unsigned int)((n)>>24)&0xFF,			\
(unsigned int)((n)>>16)&0xFF,			\
(unsigned int)((n)>>8)&0xFF,			\
(unsigned int)((n)&0xFF)

#define IP_PARTS(n) IP_PARTS_NATIVE(ntohl(n))

#define IP_NAT_TUPLE_PROT_INIT(t,_src,_dst,pr)		\
do {							\
	memset((t), 0, sizeof(*(t)));			\
	(t)->proto = (pr);				\
	(t)->tuple.src.ip = (_src);			\
	(t)->tuple.dst.ip = (_dst);			\
} while(0)

struct ip_nat_pktinfo
{
	const struct device *device;
	struct ip_nat_tuple_proto tuplepr;
};

enum ip_nat_direction
{
	IP_NAT_INCOMING,
	IP_NAT_OUTGOING
};

#ifdef __KERNEL__
static inline void dump_tuplepr(const char *fn,
				enum ip_nat_direction dir, 
				const struct ip_nat_tuple_proto *tuplepr)
{
#ifdef DEBUG
	printk("%s: %s proto=%u src=%u.%u.%u.%u id=%u "
	       "dst=%u.%u.%u.%u type=%u code=%u\n",
	       fn, dir == IP_NAT_OUTGOING ? "OUT" : "IN", 
	       tuplepr->proto, IP_PARTS(tuplepr->tuple.src.ip),
	       ntohs(tuplepr->tuple.src.u.icmp.id), 
	       IP_PARTS(tuplepr->tuple.dst.ip),
	       tuplepr->tuple.dst.u.icmp.type,
	       tuplepr->tuple.dst.u.icmp.code);
#endif
}

static inline void dump_iph(const char *fn, const struct iphdr *iph)
{
#ifdef DEBUG
	printk("%s: proto=%u src=%u.%u.%u.%u dst=%u.%u.%u.%u\n",
	       fn, iph->protocol, IP_PARTS(iph->saddr), IP_PARTS(iph->daddr));
#endif
}

static inline void dump_manip(const char *fn,
			      enum ip_nat_direction dir, 
			      const struct ip_nat_tuple_manip *manip)
{
#ifdef DEBUG
	printk("%s manip: %s=%u.%u.%u.%u id=%u\n",
	       fn, dir == IP_NAT_OUTGOING ? "src" : "dst",
	       IP_PARTS(manip->ip), manip->u.icmp.id);
#endif
}
#endif

enum ip_nat_address_alloc
{
	IP_NAT_FIXED_SRC_IP = 0x01,
	IP_NAT_FIXED_DST_IP = 0x02,
	IP_NAT_FIXED_SRC_PER_PROTO = 0x04,
	IP_NAT_FIXED_DST_PER_PROTO = 0x08,
};

/* A range consists of a linked list of these */
struct ip_nat_address_range
{
	struct ip_nat_address_range *next;
	/* Inclusive: host order. */
	u_int32_t ip_min, ip_max;
	/* Set to 0.  alloc_perconn sets this to 1 if it was
           preferred, but that protocol's unique_tuple() failed. */
	u_int32_t range_full;
	/* Per-proto restrictions hang off the end. */
};

struct ip_nat_proto
{
	/* Next pointer. */
	struct ip_nat_proto *next;
	/* Protocol number. */
	unsigned int proto;
	/* Protocol name */
	const char *name;
	/* If you want your module count modified as perconns use this
	   protocol, set this. */
	struct module *me;

	/* Try to fill in the third arg; return true if possible. */
	int (*pkt_to_tuple)(const void *datah, 
			    size_t datalen, 
			    struct ip_nat_tuple *tuple);
	/* Do a packet translation according to the ip_nat_tuple_manip
	 * and direction (OUTGOING => src, INCOMING => dst). */
	void (*manip_pkt)(struct iphdr *hdr, 
			  size_t len,
			  const struct ip_nat_tuple_manip *manip,
			  enum ip_nat_direction direction);
	/* Is the per-proto part of the tuple in the per-proto part of
	 * this range? */
	int (*in_range)(const struct ip_nat_tuple *tuple, 
			const struct ip_nat_address_range *range);

	/* Allocate a unique IP_NAT_OUTGOING tuple if possible; return
           false if not. */
	int (*unique_tuple)(struct ip_nat_tuple_proto *tuplepr, 
			    const struct ip_nat_address_range *source_range);

	/* Invert the per-proto part of the tuple: turn xmit into reply. */
	void (*invert_tuple)(struct ip_nat_tuple *tuple);

	/* Print out the per-protocol part of the tuple. */
	unsigned int (*print_tuple)(char *buffer, const struct ip_nat_tuple *);

	/* Print out the per-protocol part of the manip. */
	unsigned int (*print_manip)(char *buffer, 
				    const struct ip_nat_tuple_manip *);
};

/* Helper for masquerading: set range to match appropriate interface */
extern int ip_nat_masq_set(const struct iphdr *iph, 
			   struct ip_nat_address_range *range);

/* Helper for redirect: make it head for that interface. */
extern int ip_nat_interface(const struct device *device,
			    struct ip_nat_address_range *range);

/* Calculates relative checksums. */
extern u_int16_t ip_nat_cheat_check(u_int32_t oldvalinv, u_int32_t newval,
				    u_int16_t oldcheck);

/* Check for tuple existence. */
extern int ip_nat_used_tuple(const struct ip_nat_tuple_proto *tuplepr);

/* Protocol registration. */
extern void ip_nat_register_proto(struct ip_nat_proto *proto);
extern void ip_nat_unregister_proto(struct ip_nat_proto *proto);
#endif /*_IP_NAT_H*/
