/* Simple program to create IP Nat setup entries */

#include <stdio.h>
#include <errno.h>
#include <sys/types.h>
#include <net/if.h>
#include <netinet/in.h>
#include <getopt.h>
#include <string.h>
#include <stdlib.h>
#include <limits.h>
#include <stdarg.h>
#include <netdb.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include "ip_nat_setup.h"

#define BASE_OPTIONS				\
{ "source", 1, 0, 's' },			\
{ "destination", 1, 0, 'd' },			\
{ "interface", 1, 0, 'i' },			\
{ "version", 1, 0, 'v' }

#define VERSION "0.1"

#define IP_PARTS_NATIVE(n)			\
(unsigned int)((n)>>24)&0xFF,			\
(unsigned int)((n)>>16)&0xFF,			\
(unsigned int)((n)>>8)&0xFF,			\
(unsigned int)((n)&0xFF)

#define IP_PARTS(n) IP_PARTS_NATIVE(ntohl(n))

static void print_usage(char *msg, ...) __attribute__((noreturn));

static int get_address_range(const char *string, 
			     u_int32_t *ipmin, u_int32_t *ipmax)
{
	struct in_addr minaddr;
	char *delim;

	delim = strchr(string, ':');
	if (delim) {
		struct in_addr maxaddr;

		if (!inet_aton(delim+1, &maxaddr))
			return 0;
		*ipmax = maxaddr.s_addr;
	}
	else if ((delim = strchr(string, '/')) != NULL) {
		if (strchr(delim+1, '.')) {
			struct in_addr addrmask;
			if (!inet_aton(delim+1, &addrmask))
				return 0;

			*delim = '\0';
			if (inet_aton(string, &minaddr) == 0)
				return 0;

			/* These must be in HOST order. */
			*ipmin = ntohl(minaddr.s_addr & addrmask.s_addr);
			*ipmax = ntohl(minaddr.s_addr | ~addrmask.s_addr);
		} else {
			unsigned long mask;
			errno = 0;
			mask = strtoul(delim+1, NULL, 10);
			if (mask == ULONG_MAX && errno)
				return 0;

			mask = htonl((~0) << (32 - mask));
			*delim = '\0';
			if (inet_aton(string, &minaddr) == 0)
				return 0;

			*ipmin = ntohl(minaddr.s_addr) & mask;
			*ipmax = ntohl(minaddr.s_addr) | ~mask;
		}
	} else {
		if (!inet_aton(string, &minaddr))
			return 0;
		*ipmax = *ipmin = ntohl(minaddr.s_addr);
	}

	return (*ipmax >= *ipmin);
}

static int get_address(const char *string, u_int32_t *ip, u_int32_t *mask)
{
	struct in_addr addr;
	char *delim;

	delim = strchr(string, '/');
	if (delim) {
		struct in_addr addrmask;

		if (strchr(delim+1, '.')) {
			if (!inet_aton(delim+1, &addrmask))
				return 0;
			*mask = addrmask.s_addr;

			*delim = '\0';
			if (!inet_aton(string, &addr))
				return 0;
			*ip = addr.s_addr & *mask;

			return 1;
		}
		else {
			unsigned long tmpmask;

			errno = 0;
			tmpmask = strtoul(delim+1, NULL, 10);
			if (tmpmask == ULONG_MAX && errno)
				return 0;

			*mask = htonl((~0) << (32 - tmpmask));
			*delim = '\0';
			if (!inet_aton(string, &addr))
				return 0;

			*ip = addr.s_addr & *mask;

			return 1;
		}
	}
	else {
		if (!inet_aton(string, &addr))
			return 0;
		*ip = addr.s_addr;
	}

	*mask = ~0;
	return 1;
}

static void 
get_source_address(struct ip_nat_setup_user *base, const char *optarg)
{
	if (!get_address(optarg, &base->tupleproto.tuple.src.ip,
			 &base->mask.tuple.src.ip))
		print_usage("Bad source address `%s'", optarg);
	printf("Source address = %08X\n", base->tupleproto.tuple.src.ip);
}

static void 
get_destination_address(struct ip_nat_setup_user *base, 
			     const char *optarg)
{
	if (!get_address(optarg, &base->tupleproto.tuple.dst.ip,
			 &base->mask.tuple.dst.ip))
		print_usage("Bad destination address `%s'", optarg);
}

static void 
get_interface(struct ip_nat_setup_user *base, const char *optarg)
{
	if (strlen(optarg) > IFNAMSIZ-1)
		print_usage("Interface name `%s' too long", optarg);
	strcpy(base->ifname, optarg);
}

void try_base_option(int c, struct ip_nat_setup_user *base, const char *optarg)
{
	switch(c) {
	case 's': 
		get_source_address(base, optarg);
		break;

	case 'd': 
		get_destination_address(base, optarg);
		break;

	case 'i': 
		get_interface(base, optarg);
		break;

	case 'v':
		print_usage(VERSION);
		break;

	default:
		print_usage("Unknown option");
	}
}

enum ip_nat_transtype
{
	IP_NAT_TRANS_MASQUERADE,
	IP_NAT_TRANS_NAT,
	IP_NAT_TRANS_RNAT,
	IP_NAT_TRANS_REDIRECT,
};

struct ip_nat_icmp
{
	enum ip_nat_transtype type;
	/* These unused if type == IP_NAT_TRANS_MASQUERADE/REDIRECT */
	u_int32_t ipmin, ipmax;
};

static int get_port(const char *port, const char *proto, u_int16_t *ret)
{
	struct servent *service;
	int pnum;

	service = getservbyname(port, proto);
	if (service)
		*ret = ntohs(service->s_port);
	else {
		pnum = atoi(port);
		if (pnum < 1 || pnum > 65535) 
			return 0;
		*ret = (u_int16_t)pnum;
	}
	return 1;
}

static struct ip_nat_setup_user *
icmp_nat(int argc, char *argv[], size_t *size)
{
	int c;
	static struct option opts[] = {
		BASE_OPTIONS,
		{ "masquerade", 0, 0, 'm' },
		{ "nat", 1, 0, 'n' },
		{ "rnat", 1, 0, 'r' },
		{ "redirect", 0, 0, 'l' },
		{ 0 }
	};
	static struct 
	{
		struct ip_nat_setup_user base;
		struct ip_nat_icmp icmp;
	} ret;
	int type = -1;

	while ((c = getopt_long(argc, argv, "s:d:i:", opts, NULL)) != -1) {
		switch (c) {
		case 'm':
			type = IP_NAT_TRANS_MASQUERADE;
			ret.base.direction = IP_NAT_OUTGOING;
			break;

		case 'l':
			type = IP_NAT_TRANS_REDIRECT;
			ret.base.direction = IP_NAT_INCOMING;
			break;

		case 'n':
			if (!get_address_range(optarg, &ret.icmp.ipmin, 
					       &ret.icmp.ipmax)) 
				print_usage("Invalid argument to --nat: %s",
					    optarg);
			type = IP_NAT_TRANS_NAT;
			ret.base.direction = IP_NAT_OUTGOING;
			break;

		case 'r':
			if (!get_address_range(optarg, &ret.icmp.ipmin, 
					       &ret.icmp.ipmax)) 
				print_usage("Invalid argument to --rnat: %s",
					    optarg);
			type = IP_NAT_TRANS_RNAT;
			ret.base.direction = IP_NAT_INCOMING;
			break;

		default:
			try_base_option(c, &ret.base, optarg);
		}
	}
	if (type == -1)
		print_usage("Need --masquerade, --nat, --rnat or --redirect");
	ret.icmp.type = type;
	ret.base.tupleproto.proto = IPPROTO_ICMP;
	ret.base.mask.proto = 0xFF;

	strcpy(ret.base.typename, "icmp-nat");

	*size = sizeof(ret);
	return &ret.base;
}

struct ip_nat_tcp_udp
{
	enum ip_nat_transtype type;
	/* These unused if type == IP_NAT_TRANS_MASQUERADE/REDIRECT */
	u_int32_t ipmin, ipmax;
	/* Port to map to if IP_NAT_TRANS_RNAT/REDIRECT */
	u_int16_t mapport;
};

static struct ip_nat_setup_user *
tcpudp_nat(int argc, char *argv[], size_t *size, const char *proto)
{
	int c;
	static struct option opts[] = {
		BASE_OPTIONS,
		{ "masquerade", 0, 0, 'm' },
		{ "nat", 1, 0, 'n' },
		{ "rnat", 1, 0, 'r' },
		{ "redirect", 0, 0, 'l' },
		{ "dport", 1, 0, '1' },
		{ "mapport", 1, 0, '2' },
		{ 0 }
	};
	static struct 
	{
		struct ip_nat_setup_user base;
		struct ip_nat_tcp_udp tcpudp;
	} ret;
	int type = -1;

	while ((c = getopt_long(argc, argv, "s:d:i:", opts, NULL)) != -1) {
		switch (c) {
		case 'm':
			type = IP_NAT_TRANS_MASQUERADE;
			ret.base.direction = IP_NAT_OUTGOING;
			break;

		case 'l':
			type = IP_NAT_TRANS_REDIRECT;
			ret.base.direction = IP_NAT_INCOMING;
			break;

		case 'n':
			if (!get_address_range(optarg, &ret.tcpudp.ipmin, 
					       &ret.tcpudp.ipmax)) 
				print_usage("Invalid argument to --nat: %s",
					    optarg);
			type = IP_NAT_TRANS_NAT;
			ret.base.direction = IP_NAT_OUTGOING;
			break;

		case 'r':
			if (!get_address_range(optarg, &ret.tcpudp.ipmin, 
					       &ret.tcpudp.ipmax)) 
				print_usage("Invalid argument to --rnat: %s",
					    optarg);
			type = IP_NAT_TRANS_RNAT;
			ret.base.direction = IP_NAT_INCOMING;
			break;

		case '1':
			if (type != IP_NAT_TRANS_RNAT
			    && type != IP_NAT_TRANS_REDIRECT)
				print_usage("--dport only valid with --rnat"
					    " or --redirect");

			if (!get_port(optarg, proto, 
				      &ret.base.tupleproto.tuple.dst
				      .u.tcp.port))
				print_usage("Invalid argument to --dport: %s",
					    optarg);
			/* Because it's compared directly with packet,
			 * port must be in network order. */
			ret.base.tupleproto.tuple.dst.u.tcp.port
				= htons(ret.base.tupleproto.tuple.dst
					.u.tcp.port);
			ret.base.mask.tuple.dst.u.tcp.port = 0xFFFF;
			break;

		case '2':
			if (type != IP_NAT_TRANS_RNAT
			    && type != IP_NAT_TRANS_REDIRECT)
				print_usage("--mapport only valid with --rnat"
					    " or --redirect");

			if (!get_port(optarg, proto, &ret.tcpudp.mapport))
			       print_usage("Invalid argument to --mapport: %s",
					   optarg);
			break;

		default:
			try_base_option(c, &ret.base, optarg);
		}
	}
	if (optind < argc)
		print_usage("unknown argument found at end");

	if (type == -1)
		print_usage("Need --masquerade, --nat, --rnat or --redirect");
	ret.tcpudp.type = type;
	ret.base.mask.proto = 0xFF;

	*size = sizeof(ret);
	return &ret.base;
}

static struct ip_nat_setup_user *tcp_nat(int argc, char *argv[], size_t *size)
{
	struct ip_nat_setup_user *ret = tcpudp_nat(argc, argv, size, "tcp");

	ret->tupleproto.proto = IPPROTO_TCP;
	strcpy(ret->typename, "tcp-nat");

	return ret;
}

static struct ip_nat_setup_user *udp_nat(int argc, char *argv[], size_t *size)
{
	struct ip_nat_setup_user *ret = tcpudp_nat(argc, argv, size, "udp");

	ret->tupleproto.proto = IPPROTO_UDP;
	strcpy(ret->typename, "udp-nat");

	return ret;
}

static struct ip_nat_setup_user *
udp_nat_single(int argc, char *argv[], size_t *size)
{
	struct ip_nat_setup_user *ret = tcpudp_nat(argc, argv, size, "udp");

	ret->tupleproto.proto = IPPROTO_UDP;
	strcpy(ret->typename, "udp-nat-single");

	return ret;
}

static struct ip_nat_setup_user *user_nat(int argc, char *argv[], size_t *size)
{
	int c;
	static struct option opts[] = {	
		BASE_OPTIONS,
		{ "protocol", 1, 0, 'p' },
		{ 0 }
	};
	static struct 
	{
		struct ip_nat_setup_user base;
		u_int32_t protocol;
	} ret;

	while ((c = getopt_long(argc, argv, "s:d:i:", opts, NULL)) != -1) {
		switch (c) {
		case 'p': {
			struct protoent *pent;
			pent = getprotobyname(optarg);
			if (!pent)
				print_usage("Invalid argument to --protocol: %s",
					    optarg);
			ret.base.tupleproto.proto = pent->p_proto;
			ret.base.mask.proto = 0xFF;
			break;
		}

		default:
			try_base_option(c, &ret.base, optarg);
		}
	}
	strcpy(ret.base.typename, "user-nat");

	*size = sizeof(ret);
	return &ret.base;
}

static struct 
{
	const char *typename;
	struct ip_nat_setup_user *(*activate)(int, char *[], size_t *);
	const char *usage;
} registration[] 
= { { "icmp-nat", icmp_nat, 
      "  --masquerade\n"
      "  --nat address/mask|address-address\n"
      "  --rnat address/mask|address-address\n"
      "  --redirect\n" },
    { "tcp-nat", tcp_nat, 
      "  --masquerade\n"
      "  --nat address/mask|address-address\n"
      "  --rnat address/mask|address-address [--dport port] [--mapport port]\n"
      "  --redirect [--dport port] [--mapport port]\n" },
    { "udp-nat", udp_nat, 
      "  --masquerade\n"
      "  --nat address/mask|address-address\n"
      "  --rnat address/mask|address-address [--dport port] [--mapport port]\n"
      "  --redirect [--dport port] [--mapport port]\n" },
    { "udp-nat-single", udp_nat_single, 
      "  --masquerade\n"
      "  --nat address/mask|address-address\n"
      "  --rnat address/mask|address-address [--dport port] [--mapport port]\n"
      "  --redirect [--dport port] [--mapport port]\n" },
    { "user-nat", user_nat, 
      "  [NOT YET IMPLEMENTED]\n"
      "  --protocol proto\n"
      "  --mark markvalue\n" }
};

static void
print_usage(char *msg, ...)
{
	unsigned int i;
	va_list args;

	va_start(args, msg);
	vfprintf(stderr, msg, args);
	va_end(args);

	fprintf(stderr, 
		"\nUsage: ip_nat_ctl [insert|delete] [type] [options]...\n\n");
	fprintf(stderr, 
		"General options:\n"
		"  --source/-s [address/mask | address-address]\n"
		"				Source address\n"
		"  --destination/-d [address/mask | address-address]\n"
		"				Destination address\n"
		"  --interface/-i name\n"
		"				Interface to match\n\n");
	for (i = 0; i < sizeof(registration)/sizeof(registration[0]); i++) {
		fprintf(stderr, "Options for %s:\n", registration[i].typename);
		fprintf(stderr, "%s", registration[i].usage);
	}
	exit(1);
}

int main(int argc, char *argv[])
{
	int sockfd = socket(AF_INET, SOCK_RAW, IPPROTO_RAW);
	static struct ip_nat_setup_user *setup;
	size_t size;
	int op;
	unsigned int i;

	if (argc < 3) 
		print_usage("Too few options.");

	if (strcmp(argv[1], "insert") == 0)
		op = IP_NAT_SO_INSERT;
	else if (strcmp(argv[1], "delete") == 0) 
		op = IP_NAT_SO_DELETE;
	else
		print_usage("Expect `insert' or `delete'.");

	for (i = 0; i < sizeof(registration)/sizeof(registration[0]); i++) {
		if (strcmp(registration[i].typename, argv[2]) == 0)
			break;
	}

	if (i == sizeof(registration)/sizeof(registration[0]))
		print_usage("Unknown type `%s'", argv[2]);

	setup = registration[i].activate(argc-2, argv+2, &size);

	if (setsockopt(sockfd, IPPROTO_IP, op, setup, size) < 0) {
		perror("setsockopt problem");
		return 1;
	}
	return 0;
}
