/*	icmp-nat	Creates perconns to do a single translation of
			reply if possible.  Perconn expires after first hit.
			Translates using a scheme given by args.
*/

#include <linux/types.h>
#include <linux/ip.h>
#include <linux/icmp.h>
#include <linux/netdevice.h>
#include <linux/inetdevice.h>
#include <linux/netfilter.h>
#include <linux/rtnetlink.h>
#include <linux/module.h>

#include "ip_nat.h"
#include "ip_nat_setup.h"
#include "ip_nat_perconn.h"

#define DEBUGP printk

enum ip_nat_transtype
{
	IP_NAT_TRANS_MASQUERADE,
	IP_NAT_TRANS_NAT,
	IP_NAT_TRANS_RNAT,
	IP_NAT_TRANS_REDIRECT,
};

struct ip_nat_icmp
{
	enum ip_nat_transtype type;
	/* These unused if type == IP_NAT_TRANS_MASQUERADE or REDIRECT */
	u_int32_t ipmin, ipmax;
};

struct ip_nat_address_range_icmp
{
	struct ip_nat_address_range range;
	u_int16_t idmin, idmax;
};

static int check(const struct ip_nat_setup *setup, size_t extrasize)
{
	struct ip_nat_icmp *info = (struct ip_nat_icmp *)(setup + 1);

	if (extrasize != sizeof(struct ip_nat_icmp)) {
		DEBUGP("icmp_nat_check: extra size %u != %u.\n",
		       extrasize, sizeof(struct ip_nat_icmp));
		return 0;
	}

	/* Must specify proto == ICMP. */
	if (((struct ip_nat_pktinfo *)&setup->mask)->tuplepr.proto != 0xFF
	    || setup->info.tuplepr.proto != IPPROTO_ICMP) {
		DEBUGP("icmp_nat_check: mask %u, proto = %u.\n",
		       ((struct ip_nat_pktinfo *)&setup->mask)->tuplepr.proto,
		       setup->info.tuplepr.proto);
		return 0;
	}

	switch (info->type) {
	case IP_NAT_TRANS_MASQUERADE:
	case IP_NAT_TRANS_NAT:
		if (setup->direction != IP_NAT_OUTGOING) {
			DEBUGP("icmp_nat_check: bad dir for masq/NAT.\n");
			return 0;
		}
		break;

	case IP_NAT_TRANS_RNAT:
	case IP_NAT_TRANS_REDIRECT:
		if (setup->direction != IP_NAT_INCOMING) {
			DEBUGP("icmp_nat_check: bad dir for RNAT/REDIRECT.\n");
			return 0;
		}
		break;

	default:
		DEBUGP("icmp_nat_check: bad type %u.\n", info->type);
		return 0;
	}
	return 1;
}

/* Function to call if we get a match */
unsigned int setup(const struct iphdr *iph, 
		   const struct ip_nat_pktinfo *pkt,
		   enum ip_nat_direction direction,
		   const struct ip_nat_setup *cause,
		   struct ip_nat_perconn **perconn)
{
	struct icmphdr *hdr = (struct icmphdr *)((u_int32_t *)iph + iph->ihl);
	struct ip_nat_icmp *info = (struct ip_nat_icmp *)cause->private;
	struct ip_nat_address_range_icmp range;

	DEBUGP("icmp_nat setup: Doing it to %u.\n", hdr->type);
	/* Can only deal with these types. */
	if (hdr->type != ICMP_ECHO
	    && hdr->type != ICMP_TIMESTAMP
	    && hdr->type != ICMP_INFO_REQUEST
	    && hdr->type != ICMP_ADDRESS) {
		DEBUGP("icmp_nat setup: Can't NAT ICMP type %u.\n", hdr->type);
		return NF_DROP;
	}

	range.range.next = NULL;
	range.range.range_full = 0;
	range.idmin = 0;
	range.idmax = 65535;
	
	/* Set range. */
	switch (info->type) {
	case IP_NAT_TRANS_MASQUERADE:
		if (!ip_nat_masq_set(iph, &range.range)) {
			DEBUGP("icmp-nat: unroutable packet.\n");
			dump_iph(__FUNCTION__, iph);
			return NF_DROP;
		}
		break;

	case IP_NAT_TRANS_REDIRECT:
		if (!ip_nat_interface(pkt->device, &range.range))
			return NF_DROP;
		DEBUGP("icmp-nat: REDIRECT to %u.%u.%u.%u\n",
		       IP_PARTS_NATIVE(range.range.ip_min));
		break;

	case IP_NAT_TRANS_NAT:
	case IP_NAT_TRANS_RNAT:
		DEBUGP("icmp-nat: ip_min = %u.%u.%u.%u ip_max = %u.%u.%u.%u\n",
		       IP_PARTS_NATIVE(info->ipmin), 
		       IP_PARTS_NATIVE(info->ipmax));
		range.range.ip_min = info->ipmin;
		range.range.ip_max = info->ipmax;

		break;
	}

	dump_tuplepr(__FUNCTION__, direction, &pkt->tuplepr);
	*perconn = ip_nat_alloc_perconn(cause, &pkt->tuplepr, 
					direction, &range.range,
					ip_nat_delete_maybe, ip_nat_accept,
					HZ*60, &ip_nat_delete_perconn_timeout,
					NULL, 0);
	if (!*perconn) {
		DEBUGP("icmp-nat: ip_nat_alloc_perconn returned NULL.\n");
		return NF_DROP;
	}
	else DEBUGP("icmp-nat: ip_nat_alloc_perconn returned ok.\n");
	return NF_ACCEPT;
}

static int print_setup(char *buffer, const struct ip_nat_setup *setup)
{
	struct ip_nat_icmp *info = (struct ip_nat_icmp *)setup->private;
	int len;

	switch (info->type) {
	case IP_NAT_TRANS_MASQUERADE:
		len = sprintf(buffer, " MASQ");
		break;

	case IP_NAT_TRANS_REDIRECT:
		len = sprintf(buffer, " REDIR");
		break;

	default:
		len = sprintf(buffer, " %s %u.%u.%u.%u",
			      info->type == IP_NAT_TRANS_NAT ? "NAT" : "RNAT",
			      IP_PARTS_NATIVE(info->ipmin));
		if (info->ipmin != info->ipmax)
			len += sprintf(buffer + len, "-%u.%u.%u.%u",
				       IP_PARTS_NATIVE(info->ipmax));
	}
	return len;
}

/* Print out the private part of the perconn (if any). */
static int print_perconn(char *buffer, const struct ip_nat_perconn *perconn)
{
	return 0;
}

static struct ip_nat_setup_type icmp_nat
= { NULL, "icmp-nat", 0, &__this_module, check, setup, print_setup, 
    print_perconn };

int init_module(void)
{
	DEBUGP("icmp-nat: setup = %p\n", setup);
	return ip_nat_register_setup_type(&icmp_nat);
}

void cleanup_module()
{
	ip_nat_unregister_setup_type(&icmp_nat);
}
