/* Outgoing is simple: compare with outgoing connection */
static struct ip_nat_perconn *
find_perconn(const struct ip_nat_tuple_proto *tuplepr, 
	     enum ip_nat_direction direction,
	     long now,
	     int unexpired_only)
{
	const struct ip_nat_perconn *ret;

	dump_tuplepr(__FUNCTION__, direction, tuplepr);

	for (ret = *hash_by_tuple(tuplepr, direction);
	     ret; 
	     ret = ret->by_tuple[direction]) {
		dump_tuplepr(__FUNCTION__ "looking at", direction, 
			     &ret->dir[direction].connection);
		if (memcmp(tuplepr, &ret->dir[direction].connection, 
			   sizeof(*tuplepr)) == 0) {
			if (!unexpired_only 
			    || !atomic_read(&ret->dir[direction].expired))
				break;
		}
	}
	return (struct ip_nat_perconn *)ret;
}

void ip_nat_delete_perconn_timeout(struct ip_nat_perconn *perconn)
{
	struct ip_nat_perconn **list;
	const struct ip_nat_tuple_proto *tpr
		= &perconn->dir[IP_NAT_OUTGOING].connection;
	struct ip_nat_proto *proto
		= find_proto(perconn->allocated_tuple.proto);

	/* Kill timer: interrupts are blocked, so it's no race. */
	if (perconn->timerfn)
		del_timer(&perconn->timeout);

	/* Remove from the hashes */
	list = hash_by_tuple(&perconn->dir[IP_NAT_OUTGOING].connection,
			     IP_NAT_OUTGOING);
	while (*list != perconn) 
		list = &(*list)->by_tuple[IP_NAT_OUTGOING];
	*list = perconn->by_tuple[IP_NAT_OUTGOING];

	list = hash_by_tuple(&perconn->dir[IP_NAT_INCOMING].connection,
			     IP_NAT_INCOMING);
	while (*list != perconn) 
		list = &(*list)->by_tuple[IP_NAT_INCOMING];
	*list = perconn->by_tuple[IP_NAT_INCOMING];

	list = hash_by_ips_proto(perconn->dir[IP_NAT_OUTGOING].manip.ip, 
				 tpr->tuple.dst.ip, tpr->proto);
	while (*list != perconn) 
		list = &(*list)->by_ips_proto;
	*list = perconn->by_ips_proto;

	list = hash_by_source(tpr);
	while (*list != perconn) 
		list = &(*list)->by_source;
	*list = perconn->by_source;

	/* Decrement module usage for setup type and protocol. */
	if (perconn->cause->type->me)
		__MOD_DEC_USE_COUNT(perconn->cause->type->me);
	if (proto->me)
		__MOD_DEC_USE_COUNT(proto->me);

	kfree(perconn);
}

/* Never fails, returns NF_ACCEPT.  Suitable for use as ip_nat_fn. 
 * You need lock to use this. */
unsigned int ip_nat_delete_perconn(struct ip_nat_perconn *perconn,
				   const struct iphdr *iph,
				   enum ip_nat_direction dir)
{
	ip_nat_delete_perconn_timeout(perconn);
	return NF_ACCEPT;
}

unsigned int ip_nat_delete_maybe(struct ip_nat_perconn *perconn,
				 const struct iphdr *iph,
				 enum ip_nat_direction dir)
{
	/* Expire this direction.  If both expired, delete. */
	atomic_set(&perconn->dir[dir].expired, 1);
	if (atomic_read(&perconn->dir[!dir].expired)) 
		ip_nat_delete_perconn(perconn, iph, dir);

	return NF_ACCEPT;
}

/* Null ip_nat_fn: returns NF_ACCEPT. */
unsigned int ip_nat_accept(struct ip_nat_perconn *perconn,
			   const struct iphdr *iph,
			   enum ip_nat_direction direction)
{
	return NF_ACCEPT;
}

static void ipnat_timer_wrapper(unsigned long perconn_ul)
{
	struct ip_nat_perconn *perconn = (struct ip_nat_perconn *)perconn_ul;
	unsigned long flags;

	spin_lock_irqsave(&ip_nat_lock, flags);
	perconn->timerfn(perconn);
	spin_unlock_irqrestore(&ip_nat_lock, flags);
}

/* Returns NULL if can't alloc. */
struct ip_nat_perconn *
ip_nat_alloc_perconn(const struct ip_nat_setup *cause,
		     const struct ip_nat_tuple_proto *tuplepr,
		     enum ip_nat_direction direction,
		     struct ip_nat_address_range *source_range,
		     ip_nat_fn *postmangle, ip_nat_fn *relatedpacket,
		     unsigned long timertime, 
		     void (*timerfn)(struct ip_nat_perconn *),
		     void *extra, size_t extrasize)
{
	struct ip_nat_perconn **list;
	struct ip_nat_perconn *perconn;
	struct ip_nat_proto *proto = find_proto(tuplepr->proto);
	const struct ip_nat_tuple_proto *tpr;

	perconn = kmalloc(sizeof(struct ip_nat_perconn) + extrasize, 
			  GFP_ATOMIC);
	if (!perconn) {
		DEBUGP("ip_nat_alloc_perconn: Can't allocate %u bytes\n",
		       sizeof(struct ip_nat_perconn) + extrasize);
		return NULL;
	}

	perconn->cause = cause;
	perconn->dir[IP_NAT_INCOMING].lasthit 
		= perconn->dir[IP_NAT_OUTGOING].lasthit 
		= jiffies;
	perconn->dir[IP_NAT_INCOMING].hitcount 
		= perconn->dir[IP_NAT_OUTGOING].hitcount
		= 0;
	perconn->postmangle = postmangle;
	perconn->relatedpacket = relatedpacket;
	atomic_set(&perconn->dir[IP_NAT_OUTGOING].expired, 0);
	atomic_set(&perconn->dir[IP_NAT_INCOMING].expired, 0);
	memcpy(perconn+1, extra, extrasize);

	/* New tuple: start at current IP_NAT_OUTGOING tuple. */
	perconn->allocated_tuple = *tuplepr;
	if (direction == IP_NAT_INCOMING)
		ip_nat_invert_tuple_proto(&perconn->allocated_tuple);

	if (!get_unique_tuple(&perconn->allocated_tuple, source_range)) {
		DEBUGP("ip_nat_alloc_perconn: Can't get unique tuple.\n");
		kfree(perconn);
		return NULL;
	}

	dump_tuplepr("alloc_perconn's new OUTGOING perconn", direction, 
		     &perconn->allocated_tuple);
	if (direction == IP_NAT_OUTGOING) {
		/* A=Internal src, B=External dest, C=NAT address.
		   Outgoing: A/B -> C/B
		   Incoming: B/C -> B/A

		   tuplepr = A/B, allocated = C/B.
		*/

		/* Outgoing match = A/B.  Outgoing manip = C. */
		perconn->dir[IP_NAT_OUTGOING].connection = *tuplepr;
		perconn->dir[IP_NAT_OUTGOING].manip
			= perconn->allocated_tuple.tuple.src;

		/* Incoming match = B/C
		 * Incoming manip = A. */
		perconn->dir[IP_NAT_INCOMING].connection
			= perconn->allocated_tuple;
		ip_nat_invert_tuple_proto(&perconn->dir[IP_NAT_INCOMING]
					  .connection);
		perconn->dir[IP_NAT_INCOMING].manip = tuplepr->tuple.src;
	}
	else {
		/* A=External src, B=Internal dest, C=RNAT address.
		   Incoming: A/B -> A/C
		   Outgoing: C/A -> B/A

		   tuplepr = A/B, allocated = C/A.
		*/
		/* Incoming match = A/B. */
		perconn->dir[IP_NAT_INCOMING].connection = *tuplepr;
		/* Outgoing match = C/A. */
		perconn->dir[IP_NAT_OUTGOING].connection 
			= perconn->allocated_tuple;

		/* Incoming manip = C */
		perconn->dir[IP_NAT_INCOMING].manip
			= perconn->allocated_tuple.tuple.src;
		/* Outgoing manip = B. */
		{
			struct ip_nat_tuple_proto tmp_tpr = *tuplepr;
			ip_nat_invert_tuple_proto(&tmp_tpr);		
			perconn->dir[IP_NAT_OUTGOING].manip
				= tmp_tpr.tuple.src;
		}
	}
#if 1
	dump_tuplepr("alloc_perconn", IP_NAT_OUTGOING, 
		     &perconn->dir[IP_NAT_OUTGOING].connection);
	dump_manip("alloc_perconn", IP_NAT_OUTGOING, 
		     &perconn->dir[IP_NAT_OUTGOING].manip);
	dump_tuplepr("alloc_perconn", IP_NAT_INCOMING,
		     &perconn->dir[IP_NAT_INCOMING].connection);
	dump_manip("alloc_perconn", IP_NAT_INCOMING, 
		     &perconn->dir[IP_NAT_INCOMING].manip);
#endif

	/* Insert into the various hashes */ 
	list = hash_by_tuple(&perconn->dir[IP_NAT_OUTGOING].connection,
			     IP_NAT_OUTGOING);
	perconn->by_tuple[IP_NAT_OUTGOING] = *list;
	*list = perconn;

	list = hash_by_tuple(&perconn->dir[IP_NAT_INCOMING].connection,
			     IP_NAT_INCOMING);
	perconn->by_tuple[IP_NAT_INCOMING] = *list;
	*list = perconn;

	/* Convenience variable */
	tpr = &perconn->dir[IP_NAT_OUTGOING].connection;
	DEBUGP("ips_proto_hash add src=%u.%u.%u.%u dst=%u.%u.%u.%u proto=%u\n",
	       IP_PARTS(perconn->dir[IP_NAT_OUTGOING].manip.ip), 
	       IP_PARTS(tpr->tuple.dst.ip), tpr->proto);

	list = hash_by_ips_proto(perconn->dir[IP_NAT_OUTGOING].manip.ip, 
				 tpr->tuple.dst.ip, tpr->proto);
	perconn->by_ips_proto = *list;
	*list = perconn;

	list = hash_by_source(tpr);
	perconn->by_source = *list;
	*list = perconn;

	/* Increment module usage for setup type and protocol. */
	if (perconn->cause->type->me)
		__MOD_INC_USE_COUNT(perconn->cause->type->me);
	if (proto->me)
		__MOD_INC_USE_COUNT(proto->me);

	/* Setup expiry timer. */
	perconn->timerfn = timerfn;
	if (timerfn) {
		init_timer(&perconn->timeout);
		perconn->timeout.function = &ipnat_timer_wrapper;
		perconn->timeout.data = (unsigned long)perconn;
		perconn->timeout.expires = jiffies + timertime;
		add_timer(&perconn->timeout);
	}

	return perconn;
}

/* Delete all perconns created by a setup. */
static void delete_perconns(const struct ip_nat_setup *setup)
{
	unsigned int i;

	/* We can traverse any of these hashes; chosen at random. */
	for (i = 0; i < IP_NAT_HTABLE_SIZE; i++) {
		struct ip_nat_perconn *perconn;

		for (perconn = perconn_by_tuple[IP_NAT_INCOMING][i];perconn;) {
			struct ip_nat_perconn *next 
				= perconn->by_tuple[IP_NAT_INCOMING];

			if (perconn->cause == setup)
				ip_nat_delete_perconn_timeout(perconn);

			perconn = next;
		}
	}
}

#ifdef CONFIG_PROC_FS
static unsigned int dump_dir(char *buffer,
			     const struct ip_nat_perconn *perconn,
			     int dir)
			     
{
	unsigned int len;
	struct ip_nat_proto *proto 
		= find_proto(perconn->dir[dir].connection.proto);
	unsigned long diff, sec, usec;

	diff = jiffies - perconn->dir[dir].lasthit;
	usec = (diff % HZ) * 1000000 / HZ;
	sec = diff / HZ;

	len = sprintf(buffer, " %s:", dir == IP_NAT_INCOMING ? "in" : "out");
	len += sprintf(buffer + len, "%slast=%lu.%06lu hits=%u", 
		       atomic_read(&perconn->dir[dir].expired) ? "EXPIRED ":"",
		       sec, usec, perconn->dir[dir].hitcount);
	len += sprintf(buffer + len, " src=%u.%u.%u.%u dst=%u.%u.%u.%u", 
		       IP_PARTS(perconn->dir[dir].connection.tuple.src.ip),
		       IP_PARTS(perconn->dir[dir].connection.tuple.dst.ip));
	len += proto->print_tuple(buffer + len, 
				  &perconn->dir[dir].connection.tuple);
	len += sprintf(buffer + len, " new %s=%u.%u.%u.%u",
		       (dir == IP_NAT_INCOMING ? "dst" : "src"),
		       IP_PARTS(perconn->dir[dir].manip.ip));

	len += proto->print_manip(buffer + len, &perconn->dir[dir].manip);

	return len;
}

static unsigned int print_perconn(char *buffer,
				  const struct ip_nat_perconn *perconn)
{
	unsigned int len;
	len = sprintf(buffer, "%s", perconn->cause->type->name);

	len += dump_dir(buffer + len, perconn, IP_NAT_INCOMING);
	len += dump_dir(buffer + len, perconn, IP_NAT_OUTGOING);

	len += perconn->cause->type->print_perconn(buffer+len, perconn);
	len += sprintf(buffer+len, "\n");

	return len;
}

/* List of perconns; offset in records not bytes. */
static int nat_get_perconns(char *buffer, char **start, off_t offset,
			    int length, int dummy)
{
	unsigned int i;
	int len = 0,last_len = 0;
	off_t upto = 0;
	unsigned long flags;

	spin_lock_irqsave(&ip_nat_lock, flags);
	/* We can traverse any of these hashes; chosen at random. */
	for (i = 0; i < IP_NAT_HTABLE_SIZE; i++) {
		struct ip_nat_perconn *perconn;

		for (perconn = perconn_by_tuple[IP_NAT_INCOMING][i];
		     perconn;
		     perconn = perconn->by_tuple[IP_NAT_INCOMING], upto++) {

			DEBUGP("nat_get_perconns: printing entry %lu.\n", upto);
			if (upto < offset) continue;

			last_len = len;
			len += print_perconn(buffer + len, perconn);
			if (len > length) {
				len = last_len;
				goto finished;
			}
		}
	}
finished:
	spin_unlock_irqrestore(&ip_nat_lock, flags);
	buffer[len] = '\0';

	DEBUGP("nat_get_perconns: Length = %i (of %i).  Offset = %li.\n",
	       len, length, upto);
	/* `start' hack - see fs/proc/generic.c line ~165 */
	*start = (char *)((unsigned int)upto - offset);
	return len;
}
#endif

