static const struct ip_nat_setup *
find_setup(enum ip_nat_direction direction,
	   const struct ip_nat_pktinfo *pkt)
{
	const struct ip_nat_setup *ptr;

	for (ptr = setup_table[direction]; ptr; ptr = ptr->next) {
		struct ip_nat_pktinfo mangled = *pkt;
		unsigned int i;

#if 1
		DEBUGP("pkt     setup    mask\n");    
#endif
		/* Apply mask. */
		for (i = 0; i < sizeof(ptr->mask)/sizeof(NICE_TYPE); i++) {
#if 1
			DEBUGP("%08X %08X %08X\n",
			       ((NICE_TYPE *)pkt)[i], 
			       ((NICE_TYPE *)&ptr->info)[i],
			       ptr->mask[i]);
#endif
			((NICE_TYPE *)&mangled)[i] &= ptr->mask[i];
		}

		if (memcmp(&ptr->info, &mangled, sizeof(mangled)) == 0)
			break;
	}

	return ptr;
}

/* Returns NF_* or IP_NAT_NO_SETUP */
static unsigned int ip_nat_do_setup(const struct iphdr *iph,
				    enum ip_nat_direction direction,
				    const struct ip_nat_pktinfo *pkt,
				    struct ip_nat_perconn **perconn)
{
	const struct ip_nat_setup *setup = find_setup(direction, pkt);

	if (!setup) {
		DEBUGP("ip_nat_do_setup: No setup in direction %s.\n",
		       direction == IP_NAT_OUTGOING ? "OUT" : "IN");
		return NF_ACCEPT;
	}
	else {
		DEBUGP("ip_nat_do_setup: setup in direction %s: %s.\n",
		       direction == IP_NAT_OUTGOING ? "OUT" : "IN",
		       setup->type->name);
		return setup->type->setup(iph, pkt, direction, setup, perconn);
	}
}

static int setup_type_cmpfn(const void *arg, const void *curr)
{
	return strcmp(arg, ((struct ip_nat_setup_type *)curr)->name) == 0;
}

static struct ip_nat_setup_type *find_setup_type(const char *name, int load)
{
	struct ip_nat_setup_type *ret;

	ret = modreg_find(&setup_types, name, setup_type_cmpfn);
	if (!ret && load) {
		char modulename[IP_NAT_SETUP_MAX_NAMELEN + sizeof("ip_nat_")];
		strcpy(modulename, "ip_nat_");
		strcat(modulename, name);
		request_module(modulename);
		return find_setup_type(name, 0);
	}
	else return ret;
}

int ip_nat_register_setup_type(struct ip_nat_setup_type *me)
{
	if (find_setup_type(me->name, 0))
		return -EEXIST;

	me->usage = 0;
	modreg_prepend(&setup_types, me);
	DEBUGP("ip_nat_register_setup_type: reg %s\n", me->name);
	MOD_INC_USE_COUNT;
	return 0;
}

int ip_nat_unregister_setup_type(struct ip_nat_setup_type *me)
{
	if (me->usage) {
		DEBUGP("ip_nat_unregister_setup_type: Can't dereg %s\n",
		       me->name);
		return -EBUSY;
	}

	modreg_delete(&setup_types, me);
	MOD_DEC_USE_COUNT;
	DEBUGP("ip_nat_unregister_setup_type: dereg %s\n", me->name);
	return 0;
}

/* NAT setup tables are ordered, similar to routing tables: most
 * specific to least specific.  I've tried (and maybe failed) to be
 * intuitive here.  Higher the better.  This is a partial ordering.  */
static u_int32_t ip_nat_precedence(const struct ip_nat_pktinfo *mask)
{
	unsigned int i;
	u_int32_t prec = 0;

	/* We build this up most important to least. */

	/* Do we care about device?: 1 bit */
	prec |= mask->device ? 1 : 0;

	/* Source IP bits + destination IP bits: 7 bits */
	prec <<= 7;
	prec |= hweight32(mask->tuplepr.tuple.src.ip) 
		+ hweight32(mask->tuplepr.tuple.dst.ip);

	/* Do we care about protocol?: 1 bit */
	prec <<= 1;
	prec |= mask->tuplepr.proto ? 1 : 0;

	/* Per-protocol information: 1+5n bits */
	{
		u_int32_t perprotobits = 1;
		u_int32_t perprotoprec = 0;

		for (i = 0; 
		     i < sizeof(mask->tuplepr.tuple.src.u)/sizeof(u_int32_t); 
		     i++) {
			perprotoprec
				+= hweight32(mask->tuplepr.tuple.src.u.all[i]);
			perprotobits += 5;
		}

		for (i = 0; 
		     i < sizeof(mask->tuplepr.tuple.dst.u)/sizeof(u_int32_t); 
		     i++) {
			perprotoprec
				+= hweight32(mask->tuplepr.tuple.dst.u.all[i]);
			perprotobits += 5;
		}

		prec <<= perprotobits;
		prec |= perprotoprec;
	}

	return prec;
}

static int prec_cmpfn(const void *new, const void *curr)
{
	return ((struct ip_nat_setup *)new)->precedence
		> ((struct ip_nat_setup *)curr)->precedence;
}

/* Return 0 or -errno */
static int setup_user_to_setup(const struct ip_nat_setup_user *user,
			       struct ip_nat_setup **setup,
			       size_t extrasize)
{
	struct ip_nat_setup_type *type;
	struct ip_nat_pktinfo *pktinfomask;

	type = find_setup_type(user->typename, 1);
	if (!type) {
		DEBUGP("setup_user_to_setup: can't find type `%s'\n", 
		       user->typename);
		return -ENOENT;
	}

	if (user->direction != IP_NAT_OUTGOING
	    && user->direction != IP_NAT_INCOMING) {
		DEBUGP("setup_user_to_setup: bad direction `%i'\n", 
		       user->direction);
		return -EINVAL;
	}

	*setup = kmalloc(sizeof(struct ip_nat_setup)+extrasize, GFP_USER);
	if (!*setup)
		return -ENOMEM;

	memset((*setup)->mask, 0, sizeof((*setup)->mask));
	pktinfomask = (struct ip_nat_pktinfo *)(*setup)->mask;
	pktinfomask->tuplepr = user->mask;
	/* If they specify an interface, include it in mask. */
	if (user->ifname[0]) 
		memset(&pktinfomask->device, 0xFF,sizeof(pktinfomask->device));
	/* FIXME: Non-zeroes in the holes will screw us here --RR */
	(*setup)->info.tuplepr = user->tupleproto;
	(*setup)->info.device = dev_get(user->ifname);
	(*setup)->type = type;
	(*setup)->precedence = ip_nat_precedence(pktinfomask);
	memcpy((*setup)->ifname, user->ifname, sizeof((*setup)->ifname));
	(*setup)->direction = user->direction;
	memcpy((*setup)->private, user->private, extrasize);
	return 0;
}

static int insert_setup(struct ip_nat_setup_user *setupuser, unsigned int len)
{
	int ret;
	struct ip_nat_setup *setup;

	ret = setup_user_to_setup(setupuser, &setup, 
				  len - sizeof(struct ip_nat_setup_user));
	if (ret == 0) {
		if (setup->type->check(setup, 
				       len - sizeof(struct ip_nat_setup_user))) {
			modreg_insert(&setup_table[setup->direction], setup, 
				      prec_cmpfn);
			setup->type->usage++;
			if (setup->type->me)
				__MOD_INC_USE_COUNT(setup->type->me);
		} else {
			DEBUGP("insert_setup: type check for %s failed\n", 
			       setupuser->typename);
			kfree(setup);
			ret = -EINVAL;
		}
	}
	return ret;
}

static int setup_cmpfn(const void *find, const void *curr)
{
	/* Compare everything after next ptr and device ptr. */
	return memcmp(find+sizeof(void *)*2, curr+sizeof(void *)*2, 
		      sizeof(struct ip_nat_setup) - sizeof(void *)*2) == 0;
}

static int delete_setup(const struct ip_nat_setup_user *setupuser)
{
	int ret;
	struct ip_nat_setup *setup;

	ret = setup_user_to_setup(setupuser, &setup, 0);
	if (ret == 0) {
		struct ip_nat_setup *match;
		match = modreg_find(&setup_table[setup->direction], setup,
				    setup_cmpfn);
		if (!match)
			ret = -ENOENT;
		else {
			/* Blow up all the perconns created by this setup. */
			delete_perconns(match);

			modreg_delete(&setup_table[setup->direction], match);
			setup->type->usage--;
			if (setup->type->me)
				__MOD_DEC_USE_COUNT(setup->type->me);
		}

		kfree(setup);
	}
	return ret;
}

static void ipnat_setup_down_device(const struct device *dev)
{
	struct ip_nat_setup *setup;
	int i;
	unsigned long flags;

	spin_lock_irqsave(&ip_nat_lock, flags);
	for (i = 0; i < 2; i++) {
		for (setup = setup_table[i]; setup; setup = setup->next) {
			if (setup->info.device == dev)
				setup->info.device = NULL;
		}
	}
	spin_unlock_irqrestore(&ip_nat_lock, flags);
}
	
static void ipnat_setup_up_device(const struct device *dev, const char *name)
{
	struct ip_nat_setup *setup;
	int i;
	unsigned long flags;

	DEBUGP("Bringing up device %s (%p).\n", name, dev);
	spin_lock_irqsave(&ip_nat_lock, flags);
	for (i = 0; i < 2; i++) {
		for (setup = setup_table[i]; setup; setup = setup->next) {
			if (strcmp(setup->ifname, name) == 0)
				setup->info.device = dev;
		}
	}
	spin_unlock_irqrestore(&ip_nat_lock, flags);
}

static int do_ip_nat_ctl(int cmd, void *user, unsigned int len)
{
	int ret;

	switch (cmd) {
	case IP_NAT_SO_INSERT: {
		struct ip_nat_setup_user *setupuser;

		if (len < sizeof(struct ip_nat_setup)) {
			DEBUGP("IP_NAT_SO_INSERT: len %u < %u\n", 
			       len, sizeof(struct ip_nat_setup));
			ret = -EINVAL;
		} else if (!(setupuser = kmalloc(len, GFP_USER)))
			ret = -ENOMEM;
		else if (copy_from_user(setupuser, user, len)) {
			kfree(setupuser);
			ret = -EFAULT;
		}
		else {
			ret = insert_setup(setupuser, len);
			kfree(setupuser);
		}
		break;
	}

	case IP_NAT_SO_DELETE: {
		struct ip_nat_setup_user setupuser;

		DEBUGP("IP_NAT_SO_DELETE: doing it.\n");
		if (len < sizeof(setupuser))
			ret = -EINVAL;
		else if (copy_from_user(&setupuser, user, sizeof(setupuser)))
			ret = -EFAULT;
		else ret = delete_setup(&setupuser);

		break;
	}

	default:
		panic("ip_nat_setup: Someone wake Rusty.\n");
	}

	return ret;
}

#ifdef CONFIG_PROC_FS
/* List of perconns; offset in records not bytes. */
static int nat_get_setups(char *buffer, char **start, off_t offset,
			  int length, int dummy)
{
	struct ip_nat_setup *setup;
	off_t upto = 0;
	unsigned long flags;
	int len = 0,last_len = 0;
	int i;

	spin_lock_irqsave(&ip_nat_lock, flags);
	for (i = 0; i < 2; i++) {
		for (setup = setup_table[i]; setup; setup=setup->next, upto++){
			struct ip_nat_pktinfo *maskpr
				= (struct ip_nat_pktinfo *)&setup->mask;
			if (upto < offset) continue;

			last_len = len;
			len += sprintf(buffer + len, 
				       "%3s %16s %9s %u.%u.%u.%u/%lu ->"
				       " %u.%u.%u.%u/%lu",
				       i == IP_NAT_OUTGOING 
				       ? "OUT" : "IN",
				       setup->type->name,
				       setup->ifname, 
				       IP_PARTS(setup->info.tuplepr.tuple
						.src.ip),
#if 1
				       maskpr->tuplepr.tuple.src.ip
				       == 0xFFFFFFFF ? 32 
				       : ffz(maskpr->tuplepr.tuple.src.ip),
#else
				       ntohl(maskpr->tuplepr.tuple.src.ip),
#endif
				       IP_PARTS(setup->info.tuplepr.tuple
						.dst.ip),
#if 1
				       maskpr->tuplepr.tuple.dst.ip 
				       == 0xFFFFFFFF ? 32 
				       : ffz(maskpr->tuplepr.tuple.dst.ip)
#else
				       ntohl(maskpr->tuplepr.tuple.dst.ip)
#endif
					       );
			if (maskpr->tuplepr.proto)
				len += sprintf(buffer + len,
					       " proto=%u", 
					       setup->info.tuplepr.proto);
			len += setup->type->print_setup(buffer+len, setup);
			len += sprintf(buffer + len, "\n");
			if (len > length) {
				len = last_len;
				goto finished;
			}
		}
	}
 finished:
	spin_unlock_irqrestore(&ip_nat_lock, flags);
	buffer[len] = '\0';

	DEBUGP("nat_get_setups: Length = %i (of %i).  Offset = %li.\n",
	       len, length, upto);
	/* `start' hack - see fs/proc/generic.c line ~165 */
	*start = (char *)((unsigned int)upto - offset);
	return len;
}
#endif

static struct nf_setsockopt_ops ip_nat_sockopts 
= { NULL, PF_INET, IP_NAT_BASE_CTL, IP_NAT_SO_MAX, do_ip_nat_ctl };

static int init_ip_nat_setup(void)
{
	return nf_register_sockopt(&ip_nat_sockopts) >= 0;
}

static void cleanup_ip_nat_setup(void)
{
	nf_unregister_sockopt(&ip_nat_sockopts);
}

