/* It's not worth having separate modules for each of these, and we need
 * ICMP anyway. */
#include <linux/types.h>
#include <linux/icmp.h>
#include <linux/udp.h>
#include <linux/tcp.h>
#include <net/tcp.h>
#include <net/udp.h>
#include <net/protocol.h>
#include <net/checksum.h>
#include "ip_nat.h"

struct tcpudpicmp_address_range
{
	/* Inclusive range: host order */
	u_int16_t min, max;
};

static int icmp_pkt_to_tuple(const void *datah, 
			     size_t datalen, 
			     struct ip_nat_tuple *tuple)
{
	const struct icmphdr *hdr = datah;
	if ((char *)&hdr->un.echo.id + sizeof(hdr->un.echo.id) - (char *)datah
	    > datalen)
		return 0;
	tuple->dst.u.icmp.type = hdr->type;
	tuple->dst.u.icmp.code = hdr->code;
	tuple->src.u.icmp.id = hdr->un.echo.id;

	return 1;
}

static void icmp_manip_pkt(struct iphdr *iph, size_t len,
			   const struct ip_nat_tuple_manip *manip,
			   enum ip_nat_direction direction)
{
	struct icmphdr *hdr = (struct icmphdr *)((u_int32_t *)iph + iph->ihl);
	
	hdr->checksum = ip_nat_cheat_check(hdr->un.echo.id ^ 0xFFFF,
					   manip->u.icmp.id,
					   hdr->checksum);
	DEBUGP("Replacing icmp id %u with %u.\n", 
	       hdr->un.echo.id, manip->u.icmp.id);
	hdr->un.echo.id = manip->u.icmp.id;
}

/* Because TCP, UDP and ICMP fields of struct ip_nat_address_range (hangs
   off end) and ip_nat_manip_tuple are structurally equivalent, these
   next two functions are identical for all three, so we cheat and use
   one for all */
static int tcpudpicmp_in_range(const struct ip_nat_tuple *tuple, 
			       const struct ip_nat_address_range *srcrange)
{
	const struct tcpudpicmp_address_range *range = (void *)(srcrange + 1);

	return ntohs(tuple->src.u.tcp.port) >= range->min 
		&& ntohs(tuple->src.u.tcp.port) <= range->max;
}

static int tcpudpicmp_unique_tuple(struct ip_nat_tuple_proto *tuplepr, 
				   const struct ip_nat_address_range *srcrange)
{
	static u_int16_t id = 0;
	const struct tcpudpicmp_address_range *range = (void *)(srcrange + 1);
	unsigned int range_size = (unsigned int)range->max - range->min + 1;
	unsigned int i;

	/* Try no change, if in range. */
	if (ntohs(tuplepr->tuple.src.u.tcp.port) >= range->min
	    && ntohs(tuplepr->tuple.src.u.tcp.port) <= range->max
	    && !ip_nat_used_tuple(tuplepr)) {
		DEBUGP("Straight thru ID %hu.\n",
		       ntohs(tuplepr->tuple.src.u.tcp.port));
		return 1;
	}

	for (i = 0; i < range_size; i++, id++) {
		tuplepr->tuple.src.u.tcp.port
			= htons(range->min + (id % range_size));
		if (!ip_nat_used_tuple(tuplepr)) {
			DEBUGP("Allocated ID %u.\n", 
			       ntohs(tuplepr->tuple.src.u.tcp.port));
			return 1;
		}
		else DEBUGP("ID %u already taken.\n", 
			    ntohs(tuplepr->tuple.src.u.tcp.port));
	}
	return 0;
}

/*
  0 Echo Reply <=> 8 Echo
  13 Timestamp <=> 14 Timestamp Reply
  15 Information Request <=> 16 Information Reply
  17 Address Mask Request <=> 18 Address Mask Reply */
static void icmp_invert_tuple(struct ip_nat_tuple *tuple)
{
	static u_int8_t invmap[] = { 8, 255, 255, 255, 255, 255, 255, 255,
				     0, 255, 255, 255, 255, 14, 13, 16,
				     15, 18, 17 };

	/* Should never set up perconn to handle non-invertable types */
	if (tuple->dst.u.icmp.type >= sizeof(invmap) 
	   || invmap[tuple->dst.u.icmp.type] == 255)
		panic("icmp_invert_tuple: can't handle type %u.\n",
		      tuple->dst.u.icmp.type);

	tuple->dst.u.icmp.type = invmap[tuple->dst.u.icmp.type];
}

/* TCP and UDP packets are structurally equivalent as far as ports go,
   so use the same function.
*/
static int tcpudp_pkt_to_tuple(const void *datah, 
			       size_t datalen, 
			       struct ip_nat_tuple *tuple)
{
	const struct tcphdr *hdr = datah;
	if ((char *)&hdr->dest + sizeof(hdr->dest) - (char *)datah > datalen)
		return 0;
	tuple->src.u.tcp.port = hdr->source;
	tuple->dst.u.tcp.port = hdr->dest;

	return 1;
}

static void tcp_manip_pkt(struct iphdr *iph, size_t len,
			  const struct ip_nat_tuple_manip *manip,
			  enum ip_nat_direction direction)
{
	struct tcphdr *hdr = (struct tcphdr *)((u_int32_t *)iph + iph->ihl);
	u_int32_t oldip;
	u_int16_t *portptr;

	if (direction == IP_NAT_OUTGOING) {
		/* Get rid of src ip and src pt */
		oldip = iph->saddr;
		portptr = &hdr->source;
	}
	else {
		/* Get rid of dst ip and dst pt */
		oldip = iph->daddr;
		portptr = &hdr->dest;
	}

	hdr->check = ip_nat_cheat_check(~oldip, manip->ip,
					ip_nat_cheat_check(*portptr ^ 0xFFFF, 
							   manip->u.tcp.port,
							   hdr->check));
	*portptr = manip->u.tcp.port;
}

static void udp_manip_pkt(struct iphdr *iph, size_t len,
			  const struct ip_nat_tuple_manip *manip,
			  enum ip_nat_direction direction)
{
	struct udphdr *hdr = (struct udphdr *)((u_int32_t *)iph + iph->ihl);
	u_int32_t oldip, srcip, dstip;
	u_int16_t *portptr;

	if (direction == IP_NAT_OUTGOING) {
		/* Get rid of src ip and src pt */
		oldip = iph->saddr;
		portptr = &hdr->source;
		srcip = manip->ip;
		dstip = iph->daddr;
	}
	else {
		/* Get rid of dst ip and dst pt */
		oldip = iph->daddr;
		portptr = &hdr->dest;
		srcip = iph->saddr;
		dstip = manip->ip;
	}

	/* Don't redo checksums if it's not set. */
	if (hdr->check == 0) {
		*portptr = manip->u.udp.port;
		return;
	}

	/* Get rid of old src and dest ips, and old src and dest ports. */
	hdr->check = ip_nat_cheat_check(~oldip, manip->ip,
					ip_nat_cheat_check(*portptr ^ 0xFFFF, 
							   manip->u.udp.port,
							   hdr->check));
	if (hdr->check == 0) hdr->check = 0xFFFF;
	*portptr = manip->u.udp.port;
}

/* TCP and UDP packets are structurally equivalent as far as ports go,
   so use the same function.
*/
static void tcpudp_invert_tuple(struct ip_nat_tuple *tuple)
{
	/* Switch source and destination ports */
	u_int16_t tmp;

	tmp = tuple->src.u.tcp.port;
	tuple->src.u.tcp.port = tuple->dst.u.tcp.port;
	tuple->dst.u.tcp.port = tmp;
}

static unsigned int 
icmp_print_tuple(char *buffer, const struct ip_nat_tuple *tuple)
{
	return sprintf(buffer, " type=%u code=%u id=%u",
		       tuple->dst.u.icmp.type,
		       tuple->dst.u.icmp.code,
		       ntohs(tuple->src.u.icmp.id));
}

static unsigned int 
icmp_print_manip(char *buffer, const struct ip_nat_tuple_manip *manip)
{
	return sprintf(buffer, " id %u", ntohs(manip->u.icmp.id));
}

static unsigned int 
tcpudp_print_tuple(char *buffer, const struct ip_nat_tuple *tuple)
{
	return sprintf(buffer, " srcpt=%u dstpt=%u",
		       ntohs(tuple->src.u.tcp.port), 
		       ntohs(tuple->dst.u.tcp.port));
}

static unsigned int
tcpudp_print_manip(char *buffer, const struct ip_nat_tuple_manip *manip)
{
	return sprintf(buffer, " port=%u",
		       ntohs(manip->u.tcp.port));
}

static struct ip_nat_proto ipnat_tcp 
= { NULL, IPPROTO_TCP, "TCP", NULL,
    tcpudp_pkt_to_tuple, tcp_manip_pkt, tcpudpicmp_in_range, 
    tcpudpicmp_unique_tuple, tcpudp_invert_tuple, 
    tcpudp_print_tuple, tcpudp_print_manip };
static struct ip_nat_proto ipnat_udp
= { NULL, IPPROTO_UDP, "UDP", NULL,
    tcpudp_pkt_to_tuple, udp_manip_pkt, tcpudpicmp_in_range, 
    tcpudpicmp_unique_tuple, tcpudp_invert_tuple, 
    tcpudp_print_tuple, tcpudp_print_manip };
static struct ip_nat_proto ipnat_icmp
= { NULL, IPPROTO_ICMP, "ICMP", NULL,
    icmp_pkt_to_tuple, icmp_manip_pkt, tcpudpicmp_in_range, 
    tcpudpicmp_unique_tuple, icmp_invert_tuple, 
    icmp_print_tuple, icmp_print_manip };

void init_standard_protos(void)
{
	ip_nat_register_proto(&ipnat_tcp);
	ip_nat_register_proto(&ipnat_udp);
	ip_nat_register_proto(&ipnat_icmp);
	/* We don't really want a mod use count of 3 */
	MOD_DEC_USE_COUNT; MOD_DEC_USE_COUNT; MOD_DEC_USE_COUNT;
}

void cleanup_standard_protos(void)
{
	MOD_INC_USE_COUNT; MOD_INC_USE_COUNT; MOD_INC_USE_COUNT;
	ip_nat_unregister_proto(&ipnat_tcp);
	ip_nat_unregister_proto(&ipnat_udp);
	ip_nat_unregister_proto(&ipnat_icmp);
}
