/* tcp-nat      Creates perconns to do a single translation of reply if
		possible.  Perconns expire after IP_NAT_TCP_TIMEOUT.
		Translates using a scheme given by args.  */

#include <linux/types.h>
#include <net/ip.h>
#include <linux/tcp.h>
#include <linux/icmp.h>
#include <linux/netdevice.h>
#include <linux/inetdevice.h>
#include <linux/netfilter.h>
#include <linux/rtnetlink.h>
#include <linux/module.h>

#include "ip_nat.h"
#include "ip_nat_setup.h"
#include "ip_nat_perconn.h"

#define DEBUGP printk

#define IP_NAT_TCP_OPENING_TIMEOUT (75*HZ)
#define IP_NAT_TCP_ESTABLISHED_TIMEOUT (6*3600*HZ)
#define IP_NAT_TCP_CLOSED_TIMEOUT (75*HZ)

enum ip_nat_transtype
{
	IP_NAT_TRANS_MASQUERADE,
	IP_NAT_TRANS_NAT,
	IP_NAT_TRANS_RNAT,
	IP_NAT_TRANS_REDIRECT,
};

enum ip_nat_tcpstate {
	IP_NAT_TCP_OPENING,
	IP_NAT_TCP_ESTABLISHED,
	IP_NAT_TCP_CLOSING,
};

/* We tack this on the end of the perconn to track TCP state. */
struct ip_nat_perconn_tcp
{
	enum ip_nat_tcpstate state;
};

struct ip_nat_tcp
{
	enum ip_nat_transtype type;
	/* These unused if type == IP_NAT_TRANS_MASQUERADE/REDIRECT */
	u_int32_t ipmin, ipmax;
	/* Port to map to if IP_NAT_TRANS_RNAT/IP_NAT_TRANS_REDIRECT */
	u_int16_t mapport;
};

struct ip_nat_address_range_tcp
{
	struct ip_nat_address_range range;
	u_int16_t portmin, portmax;
};

static int check(const struct ip_nat_setup *setup, size_t extrasize)
{
	struct ip_nat_tcp *info = (struct ip_nat_tcp *)(setup + 1);

	if (extrasize != sizeof(struct ip_nat_tcp)) {
		DEBUGP("tcp_nat_check: extra size %u != %u.\n",
		       extrasize, sizeof(struct ip_nat_tcp));
		return 0;
	}

	/* Must specify proto == TCP. */
	if (((struct ip_nat_pktinfo *)&setup->mask)->tuplepr.proto != 0xFF
	    || setup->info.tuplepr.proto != IPPROTO_TCP) {
		DEBUGP("tcp_nat_check: mask %u, proto = %u.\n",
		       ((struct ip_nat_pktinfo *)&setup->mask)->tuplepr.proto,
		       setup->info.tuplepr.proto);
		return 0;
	}

	switch (info->type) {
	case IP_NAT_TRANS_NAT:
	case IP_NAT_TRANS_MASQUERADE:
		if (setup->direction != IP_NAT_OUTGOING) {
			DEBUGP("tcp_nat_check: bad dir for masq/NAT.\n");
			return 0;
		}
		break;

	case IP_NAT_TRANS_RNAT:
	case IP_NAT_TRANS_REDIRECT:
		if (setup->direction != IP_NAT_INCOMING) {
			DEBUGP("tcp_nat_check: bad dir for RNAT/REDIRECT.\n");
			return 0;
		}
		break;

	default:
		DEBUGP("tcp_nat_check: bad type %u.\n", info->type);
		return 0;
	}
	return 1;
}

static unsigned int process_icmp(struct ip_nat_perconn *perconn,
				 const struct iphdr *iph,
				 enum ip_nat_direction dir)
{
	struct icmphdr *hdr = (struct icmphdr *)((u_int32_t *)iph + iph->ihl);
	struct ip_nat_perconn_tcp *perconn_tcp = (void *)(perconn + 1);
	struct ip_nat_tcp *info = (struct ip_nat_tcp *)perconn->cause->private;
	enum ip_nat_direction reply
		= (info->type == IP_NAT_TRANS_RNAT 
		   ? IP_NAT_OUTGOING : IP_NAT_INCOMING);

	if (perconn_tcp->state == IP_NAT_TCP_OPENING
	    && dir == reply
	    && hdr->type == ICMP_DEST_UNREACH) {
		mod_timer(&perconn->timeout, 
			  jiffies + IP_NAT_TCP_CLOSED_TIMEOUT);
		perconn_tcp->state = IP_NAT_TCP_CLOSING;
	}
	return NF_ACCEPT;
}

static unsigned int process(struct ip_nat_perconn *perconn,
			    const struct iphdr *iph,
			    enum ip_nat_direction dir)
{
	struct tcphdr *hdr = (struct tcphdr *)((u_int32_t *)iph + iph->ihl);
	struct ip_nat_perconn_tcp *perconn_tcp = (void *)(perconn + 1);
	struct ip_nat_tcp *info = (struct ip_nat_tcp *)perconn->cause->private;
	enum ip_nat_direction reply
		= (info->type == IP_NAT_TRANS_RNAT 
		   ? IP_NAT_OUTGOING : IP_NAT_INCOMING);

	switch (perconn_tcp->state) {
	case IP_NAT_TCP_OPENING:
		if (dir == reply) {
			/* First reply packet: RST -> move to CLOSING,
			 * Other -> move to ESTABLISHED. */
			if (!(ntohs(iph->frag_off) & IP_OFFSET)) {
				if (hdr->rst) {
					mod_timer(&perconn->timeout, 
						  jiffies
						  + IP_NAT_TCP_CLOSED_TIMEOUT);
					perconn_tcp->state 
						= IP_NAT_TCP_CLOSING;
				} else {
					mod_timer(&perconn->timeout, 
						  jiffies 
						  + IP_NAT_TCP_ESTABLISHED_TIMEOUT);
					perconn_tcp->state
						= IP_NAT_TCP_ESTABLISHED;
				}
			}
		} else {
			/* Refresh timer. */
			mod_timer(&perconn->timeout, 
				  jiffies + IP_NAT_TCP_OPENING_TIMEOUT);
		}
		break;

	case IP_NAT_TCP_CLOSING:
		/* Nothing can change it now; we're doomed. */
		return NF_ACCEPT;

	case IP_NAT_TCP_ESTABLISHED:
		/* If we see a RST or a FIN, enter expiry cycle. */
		if (!(ntohs(iph->frag_off) & IP_OFFSET) 
		    && (hdr->fin || hdr->rst)) {
			mod_timer(&perconn->timeout, 
				  jiffies + IP_NAT_TCP_CLOSED_TIMEOUT);
			perconn_tcp->state = IP_NAT_TCP_CLOSING;
		} else {
			/* Refresh timer */
			mod_timer(&perconn->timeout, 
				  jiffies + IP_NAT_TCP_ESTABLISHED_TIMEOUT);
		}
	}

	return NF_ACCEPT;
}

/* Function to call if we get a match */
unsigned int setup(const struct iphdr *iph, 
		   const struct ip_nat_pktinfo *pkt,
		   enum ip_nat_direction direction,
		   const struct ip_nat_setup *cause,
		   struct ip_nat_perconn **perconn)
{
	struct tcphdr *hdr = (struct tcphdr *)((u_int32_t *)iph + iph->ihl);
	struct ip_nat_tcp *info = (struct ip_nat_tcp *)cause->private;
	struct ip_nat_address_range_tcp range;
	struct ip_nat_perconn_tcp tcpinfo = { IP_NAT_TCP_OPENING };

	DEBUGP("tcp_nat setup: Doing it to %u.\n", ntohs(hdr->source));

	range.range.next = NULL;
	range.range.range_full = 0;
	/* Keep low ports mapped to low ports, and high ports mapped to
	 * high ports. */
	if (ntohs(hdr->source) < 1024) {
		range.portmin = 1;
		range.portmax = 1023;
	} else {
		range.portmin = 1024;
		range.portmax = 65535;
	}

	/* Set range. */
	switch (info->type) {
	case IP_NAT_TRANS_MASQUERADE:
		if (!ip_nat_masq_set(iph, &range.range)) {
			DEBUGP("icmp-nat: unroutable packet.\n");
			dump_iph(__FUNCTION__, iph);
			return NF_DROP;
		}
		break;

	case IP_NAT_TRANS_REDIRECT:
		if (info->mapport) 
			range.portmin = range.portmax = info->mapport;
		if (!ip_nat_interface(pkt->device, &range.range))
			return NF_DROP;
		break;

	case IP_NAT_TRANS_RNAT:
		if (info->mapport) 
			range.portmin = range.portmax = info->mapport;
		/* Fall thru ... */
	case IP_NAT_TRANS_NAT:
		DEBUGP("tcp-nat: ip_min = %u.%u.%u.%u ip_max = %u.%u.%u.%u\n",
		       IP_PARTS_NATIVE(info->ipmin), 
		       IP_PARTS_NATIVE(info->ipmax));
		range.range.ip_min = info->ipmin;
		range.range.ip_max = info->ipmax;

		break;
	}

	dump_tuplepr(__FUNCTION__, direction, &pkt->tuplepr);
	*perconn = ip_nat_alloc_perconn(cause, &pkt->tuplepr, direction,
					&range.range, 
					&process,
					&process_icmp,
					IP_NAT_TCP_OPENING_TIMEOUT,
					&ip_nat_delete_perconn_timeout,
					&tcpinfo, sizeof(tcpinfo));
	if (!*perconn) {
		DEBUGP("tcp-nat: ip_nat_alloc_perconn returned NULL.\n");
		return NF_DROP;
	}
	else DEBUGP("tcp-nat: ip_nat_alloc_perconn returned ok.\n");
	return NF_ACCEPT;
}

static int print_setup(char *buffer, const struct ip_nat_setup *setup)
{
	struct ip_nat_tcp *info = (struct ip_nat_tcp *)setup->private;
	int len;

	switch (info->type) {
	case IP_NAT_TRANS_MASQUERADE:
		len = sprintf(buffer, " MASQ");
		break;

	case IP_NAT_TRANS_REDIRECT:
		len = sprintf(buffer, " REDIR");
		if (info->mapport)
			len += sprintf(buffer + len, " to port %hu",
				       info->mapport);
		break;

	case IP_NAT_TRANS_NAT:
		len = sprintf(buffer, " NAT %u.%u.%u.%u",
			      IP_PARTS_NATIVE(info->ipmin));
		if (info->ipmin != info->ipmax)
			len += sprintf(buffer + len, "-%u.%u.%u.%u",
				       IP_PARTS_NATIVE(info->ipmax));
		break;

	default:
		len = sprintf(buffer," RNAT %u.%u.%u.%u", 
			      IP_PARTS_NATIVE(info->ipmin));
		if (info->ipmin != info->ipmax)
			len += sprintf(buffer + len, "-%u.%u.%u.%u",
				       IP_PARTS_NATIVE(info->ipmax));
		if (info->mapport)
			len += sprintf(buffer + len, " to port %hu",
				       info->mapport);
		break;

	}
	return len;
}

/* Print out the private part of the perconn (if any). */
static int print_perconn(char *buffer, const struct ip_nat_perconn *perconn)
{
	struct ip_nat_perconn_tcp *perconn_tcp = (void *)(perconn + 1);

	switch (perconn_tcp->state) {
	case IP_NAT_TCP_CLOSING:
		return sprintf(buffer, " CLOSING");
	case IP_NAT_TCP_OPENING:
		return sprintf(buffer, " OPENING");
	default:
		return 0;
	}
}

static struct ip_nat_setup_type tcp_nat
= { NULL, "tcp-nat", 0, &__this_module, check, setup, print_setup, 
    print_perconn };

int init_module(void)
{
	DEBUGP("tcp-nat: setup = %p\n", setup);
	return ip_nat_register_setup_type(&tcp_nat);
}

void cleanup_module()
{
	ip_nat_unregister_setup_type(&tcp_nat);
}
