/* udp-nat      Creates perconns to do a single translation of reply if
		possible.  Perconn expires after first hit.
		Translates using a scheme given by args.  */

#include <linux/types.h>
#include <linux/ip.h>
#include <linux/udp.h>
#include <linux/netdevice.h>
#include <linux/inetdevice.h>
#include <linux/netfilter.h>
#include <linux/rtnetlink.h>
#include <linux/module.h>

#include "ip_nat.h"
#include "ip_nat_setup.h"
#include "ip_nat_perconn.h"

#define DEBUGP printk

#define IP_NAT_UDP_TIMEOUT (60*HZ)

enum ip_nat_transtype
{
	IP_NAT_TRANS_MASQUERADE,
	IP_NAT_TRANS_NAT,
	IP_NAT_TRANS_RNAT,
	IP_NAT_TRANS_REDIRECT,
};

struct ip_nat_udp
{
	enum ip_nat_transtype type;
	/* These unused if type == IP_NAT_TRANS_MASQUERADE/REDIRECT */
	u_int32_t ipmin, ipmax;
	/* Port to map to if IP_NAT_TRANS_RNAT */
	u_int16_t mapport;	
};

struct ip_nat_address_range_udp
{
	struct ip_nat_address_range range;
	u_int16_t portmin, portmax;
};

static int check(const struct ip_nat_setup *setup, size_t extrasize)
{
	struct ip_nat_udp *info = (struct ip_nat_udp *)(setup + 1);

	if (extrasize != sizeof(struct ip_nat_udp)) {
		DEBUGP("udp_nat_check: extra size %u != %u.\n",
		       extrasize, sizeof(struct ip_nat_udp));
		return 0;
	}

	/* Must specify proto == UDP. */
	if (((struct ip_nat_pktinfo *)&setup->mask)->tuplepr.proto != 0xFF
	    || setup->info.tuplepr.proto != IPPROTO_UDP) {
		DEBUGP("udp_nat_check: mask %u, proto = %u.\n",
		       ((struct ip_nat_pktinfo *)&setup->mask)->tuplepr.proto,
		       setup->info.tuplepr.proto);
		return 0;
	}

	switch (info->type) {
	case IP_NAT_TRANS_MASQUERADE:
	case IP_NAT_TRANS_NAT:
		if (setup->direction != IP_NAT_OUTGOING) {
			DEBUGP("udp_nat_check: bad dir for masq/NAT.\n");
			return 0;
		}
		break;

	case IP_NAT_TRANS_RNAT:
	case IP_NAT_TRANS_REDIRECT:
		if (setup->direction != IP_NAT_INCOMING) {
			DEBUGP("udp_nat_check: bad dir for RNAT/REDIRECT.\n");
			return 0;
		}
		break;

	default:
		DEBUGP("udp_nat_check: bad type %u.\n", info->type);
		return 0;
	}
	return 1;
}

extern struct ip_nat_setup_type udp_nat_single;

/* Returns, refreshes timer. */
static unsigned int process(struct ip_nat_perconn *perconn,
			    const struct iphdr *iph,
			    enum ip_nat_direction dir)
{
	/* Refresh timer */
	mod_timer(&perconn->timeout, jiffies + IP_NAT_UDP_TIMEOUT);
	return NF_ACCEPT;
}

/* Function to call if we get a match */
unsigned int setup(const struct iphdr *iph, 
		   const struct ip_nat_pktinfo *pkt,
		   enum ip_nat_direction direction,
		   const struct ip_nat_setup *cause,
		   struct ip_nat_perconn **perconn)
{
	struct udphdr *hdr = (struct udphdr *)((u_int32_t *)iph + iph->ihl);
	struct ip_nat_udp *info = (struct ip_nat_udp *)cause->private;
	struct ip_nat_address_range_udp range;
	ip_nat_fn *matchfn;

	DEBUGP("udp_nat setup: Doing it to %u.\n", ntohs(hdr->source));

	range.range.next = NULL;
	range.range.range_full = 0;

	/* Keep low ports mapped to low ports, and high ports mapped to
	 * high ports. */
	if (ntohs(hdr->source) < 1024) {
		range.portmin = 1;
		range.portmax = 1023;
	} else {
		range.portmin = 1024;
		range.portmax = 65535;
	}

	/* If it's a single-hit type, then expire when a match occurs. */
	if (cause->type == &udp_nat_single) matchfn = ip_nat_delete_maybe;
	/* Otherwise refresh timer. */
	else matchfn = process;

	/* Set range. */
	switch (info->type) {
	case IP_NAT_TRANS_MASQUERADE:
		if (!ip_nat_masq_set(iph, &range.range)) {
			DEBUGP("icmp-nat: unroutable packet.\n");
			dump_iph(__FUNCTION__, iph);
			return NF_DROP;
		}
		break;

	case IP_NAT_TRANS_REDIRECT:
		 if (info->mapport) 
			 range.portmin = range.portmax = info->mapport;
		 if (!ip_nat_interface(pkt->device, &range.range))
			 return NF_DROP;
		 break;

	case IP_NAT_TRANS_RNAT:
		 if (info->mapport) 
			 range.portmin = range.portmax = info->mapport;
		/* Fall thru ... */
	case IP_NAT_TRANS_NAT:
		DEBUGP("udp-nat: ip_min = %u.%u.%u.%u ip_max = %u.%u.%u.%u\n",
		       IP_PARTS_NATIVE(info->ipmin),
		       IP_PARTS_NATIVE(info->ipmax));
		range.range.ip_min = info->ipmin;
		range.range.ip_max = info->ipmax;

		break;
	}

	dump_tuplepr(__FUNCTION__, direction, &pkt->tuplepr);
	*perconn = ip_nat_alloc_perconn(cause, &pkt->tuplepr, direction,
					&range.range, 
					matchfn, 
					/* ICMP -> expire */
					&ip_nat_delete_perconn,
					IP_NAT_UDP_TIMEOUT,
					&ip_nat_delete_perconn_timeout,
					NULL, 0);
	if (!*perconn) {
		DEBUGP("udp-nat: ip_nat_alloc_perconn returned NULL.\n");
		return NF_DROP;
	}
	else DEBUGP("udp-nat: ip_nat_alloc_perconn returned ok.\n");
	return NF_ACCEPT;
}

static int print_setup(char *buffer, const struct ip_nat_setup *setup)
{
	struct ip_nat_udp *info = (struct ip_nat_udp *)setup->private;
	int len;

	switch (info->type) {
	case IP_NAT_TRANS_MASQUERADE:
		len = sprintf(buffer, " MASQ");
		break;

	case IP_NAT_TRANS_REDIRECT:
		len = sprintf(buffer, " REDIR");
		if (info->mapport)
			len += sprintf(buffer+len, " to port %hu",
				       info->mapport);
		break;

	case IP_NAT_TRANS_NAT:
		len = sprintf(buffer, " NAT %u.%u.%u.%u",
			      IP_PARTS_NATIVE(info->ipmin));
		if (info->ipmin != info->ipmax)
			len += sprintf(buffer + len, "-%u.%u.%u.%u",
				       IP_PARTS_NATIVE(info->ipmax));
		break;

	default:
		len = sprintf(buffer, " RNAT %u.%u.%u.%u", 
			      IP_PARTS_NATIVE(info->ipmin));
		if (info->ipmin != info->ipmax)
			len += sprintf(buffer + len, "-%u.%u.%u.%u",
				       IP_PARTS_NATIVE(info->ipmax));
		if (info->mapport)
			len += sprintf(buffer + len, " to port %hu",
				       info->mapport);
		break;
	}
	return len;
}

/* Print out the private part of the perconn (if any). */
static int print_perconn(char *buffer, const struct ip_nat_perconn *perconn)
{
	return 0;
}

static struct ip_nat_setup_type udp_nat
= { NULL, "udp-nat", 0, &__this_module, check, setup, print_setup, 
    print_perconn };
static struct ip_nat_setup_type udp_nat_single
= { NULL, "udp-nat-single", 0, &__this_module, check, setup, print_setup,
    print_perconn };

int init_module(void)
{
	int ret;
	DEBUGP("udp-nat: setup = %p\n", setup);
	ret = ip_nat_register_setup_type(&udp_nat);
	if (ret == 0) {
		ret = ip_nat_register_setup_type(&udp_nat_single);
		if (ret != 0)
			ip_nat_unregister_setup_type(&udp_nat);
	}
	return ret;
}

void cleanup_module()
{
	ip_nat_unregister_setup_type(&udp_nat);
	ip_nat_unregister_setup_type(&udp_nat_single);
}
