/*
 * This is the 1998 rewrite of IP Firewalling, aiming for kernel 2.3.x.
 *
 * Copyright (C) 1998 Paul F. Russell & Michael J. Neuling
 */
#ifdef __KERNEL__
#include <linux/config.h>

#include <asm/smp.h>
#include <linux/module.h>
#include <linux/skbuff.h>
#include <linux/minmax.h>
#include <linux/kmod.h>
#include <linux/stat.h>
#include <linux/interrupt.h>
#include <asm/uaccess.h>
#include <linux/netdevice.h>
#else
#include "ip_tables_usersupp.h"
#endif

#include <stdarg.h>
#include "ip_tables.h"
#include "ip_tables_standard.h"

#ifndef IP_OFFSET
#define IP_OFFSET 0x1FFF
#endif

/* Understanding locking in this code: (thanks to Alan Cox for using
 * little words to explain this to me). -- PR
 *
 * In UP, there can be two packets traversing the chains:
 * 1) A packet from the current userspace context
 * 2) A packet off the bh handlers (timer or net).
 *
 * For SMP (kernel v2.1+), multiply this by # CPUs.
 *
 * [Note that this in not correct for 2.2 - because the socket code always
 *  uses lock_kernel() to serialize, and bottom halves (timers and net_bhs)
 *  only run on one CPU at a time.  This will probably change for 2.3.
 *  It is still good to use spinlocks because that avoids the global cli() 
 *  for updating the tables, which is rather costly in SMP kernels -AK]
 *
 * This means counters and backchains can get corrupted if no precautions
 * are taken.
 *
 * To actually alter a chain on UP, we need only do a cli(), as this will
 * stop a bh handler firing, as we are in the current userspace context
 * (coming from a setsockopt()).
 *
 * On SMP, we need a write_lock_irqsave(), which is a simple cli() in
 * UP.
 *
 * For backchains and counters, we use an array, indexed by
 * [cpu_number_map[smp_processor_id()]*2 + !in_interrupt()]; the array
 * is of size [smp_num_cpus*2].  For v2.0, smp_num_cpus is effectively
 * 1.  So, confident of uniqueness, we modify counters even though we
 * only have a read lock (to read the counters, you need a write lock,
 * though).  */

/* Why I didn't use straight locking... -- PR
 * 
 * The backchains can be separated out of the ip_chains structure, and
 * allocated as needed inside ipt_check().
 *
 * The counters, however, can't.  Trying to lock these means blocking
 * interrupts every time we want to access them.  This would suck HARD
 * performance-wise.  Not locking them leads to possible corruption,
 * made worse on 32-bit machines (counters are 64-bit).  */

/* A few tricks.  We avoid running off the end by keeping a special entry
 * there which uses ipt_error as its action function.  We avoid underflow
 * (more IPT_RETURN than jumps) by pointing the back pointer from the
 * entry points at this entry too.
 */

/*#define DEBUG_IP_FIREWALL*/
/*#define DEBUG_ALLOW_ALL*/ /* Useful for remote debugging */
/*#define DEBUG_IP_FIREWALL_USER*/
/*#define DEBUG_IP_FIREWALL_LOCKING*/

#ifdef __KERNEL__
#ifdef CONFIG_SMP
#define SLOT_NUMBER() (cpu_number_map[smp_processor_id()]*2 + !in_interrupt())
#else
#define SLOT_NUMBER() (!in_interrupt())
#endif
#define NUM_SLOTS (smp_num_cpus*2)
#else
#define SLOT_NUMBER() 0
#define NUM_SLOTS 1
#endif

#if defined(DEBUG_IP_FIREWALL) || defined (DEBUG_IP_FIREWALL_USER)
#define static
#endif

#ifdef DEBUG_IP_FIREWALL_LOCKING
static unsigned int iptd_rlocks, iptd_wlocks;
#define IPTD_DEBUG_LOCK(d)			\
do {						\
	IPTD_DONT_HAVE_LOCK(d);			\
	d |= (1 << SLOT_NUMBER());		\
} while (0)

#define IPTD_DEBUG_UNLOCK(d)			\
do {						\
	IPTD_HAVE_LOCK(d);			\
	d &= ~(1 << SLOT_NUMBER());		\
} while (0)

#define IPTD_DONT_HAVE_LOCK(d)					\
do {								\
	if ((d) & (1 << SLOT_NUMBER()))				\
		printk("%s:%i: Got lock on %i already!\n", 	\
		       __FILE__, __LINE__, SLOT_NUMBER());	\
} while(0)

#define IPTD_HAVE_LOCK(d)				\
do {							\
	if (!((d) & (1 << SLOT_NUMBER())))		\
	printk("%s:%i:No lock on %i!\n", 		\
	       __FILE__, __LINE__, SLOT_NUMBER());	\
} while (0)

#else
#define IPTD_DEBUG_LOCK(d) do { } while(0)
#define IPTD_DEBUG_UNLOCK(d) do { } while(0)
#define IPTD_DONT_HAVE_LOCK(d) do { } while(0)
#define IPTD_HAVE_LOCK(d) do { } while(0)
#endif /*DEBUG_IP_FIREWALL_LOCKING*/

#ifdef __KERNEL__
#define IPTD_READ_LOCK(l) do { IPTD_DEBUG_LOCK(iptd_rlocks); read_lock(l); } while (0)
#define IPTD_WRITE_LOCK(l) do { IPTD_DEBUG_LOCK(iptd_wlocks); write_lock(l); } while (0)
#define IPTD_READ_LOCK_IRQ(l,f) do { IPTD_DEBUG_LOCK(iptd_rlocks); read_lock_irqsave(l,f); } while (0)
#define IPTD_WRITE_LOCK_IRQ(l,f) do { IPTD_DEBUG_LOCK(iptd_wlocks); write_lock_irqsave(l,f); } while (0)
#define IPTD_READ_UNLOCK(l) do { IPTD_DEBUG_UNLOCK(iptd_rlocks); read_unlock(l); } while (0)
#define IPTD_WRITE_UNLOCK(l) do { IPTD_DEBUG_UNLOCK(iptd_wlocks); write_unlock(l); } while (0)
#define IPTD_READ_UNLOCK_IRQ(l,f) do { IPTD_DEBUG_UNLOCK(iptd_rlocks); read_unlock_irqrestore(l,f); } while (0)
#define IPTD_WRITE_UNLOCK_IRQ(l,f) do { IPTD_DEBUG_UNLOCK(iptd_wlocks); write_unlock_irqrestore(l,f); } while (0)
#else
#define IPTD_READ_LOCK(l) do { IPTD_DEBUG_LOCK(iptd_rlocks);  } while (0)
#define IPTD_WRITE_LOCK(l) do { IPTD_DEBUG_LOCK(iptd_wlocks); } while (0)
#define IPTD_READ_LOCK_IRQ(l,f) do { IPTD_DEBUG_LOCK(iptd_rlocks); (f)=1; } while (0)
#define IPTD_WRITE_LOCK_IRQ(l,f) do { IPTD_DEBUG_LOCK(iptd_wlocks); (f)=1; } while (0)
#define IPTD_READ_UNLOCK(l) do { IPTD_DEBUG_UNLOCK(iptd_rlocks); } while (0)
#define IPTD_WRITE_UNLOCK(l) do { IPTD_DEBUG_UNLOCK(iptd_wlocks); } while (0)
#define IPTD_READ_UNLOCK_IRQ(l,f) do { IPTD_DEBUG_UNLOCK(iptd_rlocks); } while (0)
#define IPTD_WRITE_UNLOCK_IRQ(l,f) do { IPTD_DEBUG_UNLOCK(iptd_wlocks); } while (0)
#endif

/*
 *	Implement IP packet firewall
 */

#ifdef DEBUG_IP_FIREWALL 
#define dprintf(format, args...)  printk(format , ## args)
#else
#define dprintf(format, args...)
#endif

#ifdef DEBUG_IP_FIREWALL_USER
#define duprintf(format, args...) printk(format , ## args)
#else
#define duprintf(format, args...)
#endif

/* Lock around ipt_table_info table */
rwlock_t ipt_lock = RW_LOCK_UNLOCKED;

struct ipt_kern_entry
{
	struct ipt_ip ip;
	union ipt_proto proto;
	struct ipt_action *act;
	struct ipt_standard info;
	unsigned int cacheinfo;
};

struct per_slot
{
	unsigned int backpointer;
	struct ipt_counters counters;
};

/* The table itself */
struct ipt_table_info
{
	unsigned int number;
	unsigned int hook_entry[3];
	/* number of these */
	struct ipt_kern_entry *table; 

	/* Actually NUM_SLOTS() of these, each one pointing to [number] */
	struct per_slot *slots[0];
};

static struct ipt_table_info *ipt_table_info;
static struct ipt_action *ipt_action;

/* Pointer / size_t pairs.  NULL terminated. Returns true if all allocated. */
/* Make sure you cast to size_t!  Free every element. */
static int multialloc(int pri, ...)
{
	va_list args;
	struct ll { struct ll *prev; };
	struct ll *ptr = NULL;
	void **newptr;

	va_start(args, pri);
	while ((newptr = va_arg(args, void **)) != NULL) {
		size_t size = va_arg(args, size_t);
		size = UMAX(size, sizeof(void *));
		*newptr = kmalloc(size, pri);
		if (!*newptr) {
			/* Loop back and cleanup */
			while (ptr) {
				struct ll *oldptr = ptr->prev;
				kfree(ptr);
				ptr = oldptr;
			}
			break;
		}
		((struct ll *)*newptr)->prev = ptr;
		ptr = (struct ll *)*newptr;
	}

	va_end(args);
	/* If we hit the end, then it's OK. */
	return !newptr;
}

#define ADD_COUNTER(c,b,p) do { (c)->bcnt += (b); (c)->pcnt += (p); } while(0)

/* Returns 1 if the port is matched by the range, 0 otherwise */
static int port_match(u_int16_t min, u_int16_t max, u_int16_t port, int frag, int invert)
{
	if (frag) /* Fragments fail ANY port test. */
		return (min == 0 && max == 0xFFFF);
	else return (port >= min && port <= max) ^ invert;
}

/* Returns whether matches rule or not. */
static int ip_packet_match(const struct ipt_kern_entry *f, 
			   const char *indev,
			   const char *outdev,
			   struct iphdr *ip, 
			   int isfrag,
			   const void *hdr __attribute__((unused)))
{
#define FWINV(bool,invflg) ((bool) ^ !!(f->ip.invflags & invflg))
	/*
	 *	This is a bit simpler as we don't have to walk
	 *	an interface chain as you do in BSD - same logic
	 *	however.
	 */

	if (FWINV((ip->saddr&f->ip.smsk.s_addr) != f->ip.src.s_addr,
		  IPT_INV_SRCIP)
	    || FWINV((ip->daddr&f->ip.dmsk.s_addr) != f->ip.dst.s_addr,
		     IPT_INV_DSTIP)) {
		dprintf("Source or dest mismatch.\n");

		dprintf("SRC: %u. Mask: %u. Target: %u.%s\n", ip->saddr,
			f->ip.smsk.s_addr, f->ip.src.s_addr,
			f->ip.invflags & IPT_INV_SRCIP ? " (INV)" : "");
		dprintf("DST: %u. Mask: %u. Target: %u.%s\n", ip->daddr,
			f->ip.dmsk.s_addr, f->ip.dst.s_addr,
			f->ip.invflags & IPT_INV_DSTIP ? " (INV)" : "");
		return 0;
	}

	/* Look for ifname match */
	if (FWINV(memcmp(indev, f->ip.iniface, (f->ip.flags&IPT_F_EXACT_VIA_IN)
			 ? IFNAMSIZ : strlen(f->ip.iniface)) != 0,
		  IPT_INV_VIA_IN)) {
		dprintf("VIA in mismatch.%s\n",
			indev, f->ip.iniface, 
			f->ip.invflags&IPT_INV_VIA_IN ?" (INV)":"");
		return 0;
	}
	if (FWINV(memcmp(outdev, &f->ip.outiface, 
			 (f->ip.flags&IPT_F_EXACT_VIA_OUT)
			 ? IFNAMSIZ : strlen(f->ip.outiface)) != 0,
		  IPT_INV_VIA_OUT)) {
		dprintf("VIA out mismatch (%s vs %s).%s\n",
			outdev, f->ip.outiface, 
			f->ip.invflags&IPT_INV_VIA_OUT?" (INV)":"");
		return 0;
	}

	/* If we have a fragment rule but the packet is not a fragment
	 * the we return zero */
	if (FWINV((f->ip.flags&IPT_F_FRAG) && !isfrag, IPT_INV_FRAG)) {
		dprintf("Fragment rule but not fragment.%s\n",
			f->ip.invflags & IPT_INV_FRAG ? " (INV)" : "");
		return 0;
	}

	/* Check TOS match */
	if ((f->ip.flags & IPT_F_TOS)
	    && FWINV(f->ip.tos != ip->tos, IPT_INV_TOS)) {
		dprintf("Packet TOS %u does not match %u.%s\n",
			ip->tos, f->ip.tos,
			f->ip.invflags&IPT_INV_TOS ? " (INV)":"");
		return 0;
	}

	/* Check specific protocol */
	if (f->ip.proto
	    && FWINV(ip->protocol != f->ip.proto, IPT_INV_PROTO)) {
		dprintf("Packet protocol %hi does not match %hi.%s\n",
			ip->protocol, f->ip.proto,
			f->ip.invflags&IPT_INV_PROTO ? " (INV)":"");
		return 0;
	}

	/* Look for IP options match */
	if (f->ip.option) {
		const u_int8_t *opt = (u_int8_t *)(ip+1);
		const u_int8_t *end = opt + ip->ihl*4;
		int found = 0;

		/* FIXME: Run 0-length and > ip->ihl IP options
		 * against various stacks.  Might be fun. --RR */
		while (opt < end && !found) {
			found = (f->ip.option == *opt);
			if (*opt < 2) opt++;
			else opt += opt[1]?:1;
		}

		if (FWINV(!found, IPT_INV_OPTION)) {
			dprintf("Option %u%s found\n", f->ip.option,
			f->ip.invflags&IPT_INV_OPTION ? "" : " not");
			return 0;
		}
	}			

	return 1;
}

static inline int
tcp_option_match(const u_int8_t *opt, const u_int8_t *end, u_int8_t option, 
		 int invert)
{
	while (opt < end) {
		if (option == *opt) return !invert;
		if (*opt < 2) opt++;
		else opt += opt[1]?:1;
	}

	return invert;
}

static int 
tcp_packet_match(const struct ipt_kern_entry *f, 
		 const char *indev,
		 const char *outdev,
		 struct iphdr *ip, 
		 int isfrag, 
		 const void *hdr)
{
	const struct tcphdr *tcp=hdr;
	const struct ipt_tcp *rule = &f->proto.tcp;

	if (!ip_packet_match(f, indev, outdev, ip, isfrag, hdr))
		return 0;

#define FWINVTCP(bool,invflg) ((bool) ^ !!(f->proto.tcp.invflags & invflg))

	/* FIXME: Try tcp doff >> packet len against various stacks --RR */
	return (f->ip.proto != IPPROTO_TCP
		|| ((!rule->option 
		     || tcp_option_match((u_int8_t *)(tcp+1), 
					 (u_int8_t *)tcp + tcp->doff*4, 
					 rule->option,
					 !!(rule->invflags 
					    & IPT_TCP_INV_OPTION)))
		    && port_match(rule->spts[0], rule->spts[1],
				  ntohs(tcp->source), isfrag,
				  !!(rule->invflags & IPT_TCP_INV_SRCPT))
		    && port_match(rule->dpts[0], rule->dpts[1],
				  ntohs(tcp->dest), isfrag,
				  !!(rule->invflags & IPT_TCP_INV_DSTPT))
		    && FWINVTCP((((unsigned char *)tcp)[13] & rule->flg_mask)
				== rule->flg_cmp, 
				IPT_TCP_INV_FLAGS)));
}

static int 
udp_packet_match(const struct ipt_kern_entry *f, 
		 const char *indev,
		 const char *outdev,
		 struct iphdr *ip, 
		 int isfrag,
		 const void *hdr)
{
	const struct udphdr *udp=hdr;
	const struct ipt_udp *rule = &f->proto.udp;

	if (!ip_packet_match(f, indev, outdev, ip, isfrag, hdr))
		return 0;

	return (f->ip.proto != IPPROTO_UDP
		|| (port_match(rule->spts[0], rule->spts[1],
			       ntohs(udp->source), isfrag,
			       !!(rule->invflags & IPT_UDP_INV_SRCPT))
		    && port_match(rule->dpts[0], rule->dpts[1],
				  ntohs(udp->dest), isfrag,
				  !!(rule->invflags & IPT_UDP_INV_DSTPT))));
}

static int 
icmp_packet_match(const struct ipt_kern_entry *f, 
		  const char *indev,
		  const char *outdev,
		  struct iphdr *ip, 
		  int isfrag,
		  const void *hdr)
{
	const struct icmphdr *icmp=hdr;
	const struct ipt_icmp *rule = &f->proto.icmp;

	if (!ip_packet_match(f, indev, outdev, ip, isfrag, hdr))
		return 0;

	dprintf("icmp_packet_match: IP part matches.\n");
	dprintf("type %u (seek %u-%u) code %u (seek %u-%u).\n",
		icmp->type, rule->type[0], rule->type[1],
		icmp->code, rule->code[0], rule->code[1]);
	return (f->ip.proto != IPPROTO_ICMP
		|| ((port_match(rule->type[0], rule->type[1],
				icmp->type, isfrag, 0)
		     && port_match(rule->code[0], rule->code[1],
				   icmp->code, isfrag, 0))
		    ^ !!(rule->invflags & IPT_ICMP_INV)));
}

static int
ipt_standard_action(struct sk_buff **pskb, 
		    const char *indev,
		    const char *outdev,
		    void *info,
		    unsigned int pos)
{
	const struct ipt_standard *std = info;

	if (std->mark) {
		(*pskb)->nfreason = NF_SPECIAL_SET_BY_IPTABLES;
		(*pskb)->nfmark = std->mark;
	}

	return std->verdict;
}

static int
ipt_error(struct sk_buff **skb, 
	  const char *indev,
	  const char *outdev,
	  void *info,
	  unsigned int pos)
{
	if (net_ratelimit()) 
		printk("ip_tables: error: `%s'\n", (char *)info);

	return -NF_DROP - 1;
}

/*
 *	Returns one of the generic firewall policies, like FW_ACCEPT.
 */
static unsigned int 
ipt_check(struct sk_buff **skbp,
	  unsigned int pos,
	  const char *indev,
	  const char *outdev,
	  unsigned int *cache)
{
	u_int16_t offset;
	struct iphdr *ip = (*skbp)->nh.iph;
	void *protohdr = (u_int32_t *)ip + ip->ihl;
	unsigned int size_req;
	unsigned int i = 0;
	int spos = (int)pos;
	int (*match)(const struct ipt_kern_entry *f, 
		     const char *indev,
		     const char *outdev,
		     struct iphdr *ip, 
		     int isfrag,
		     const void *hdr);
	const unsigned int number = ipt_table_info->number;

	/* We handle fragments by dealing with the first fragment as
	 * if it was a normal packet.  All other fragments are treated
	 * normally, except that they will NEVER match rules that ask
	 * things we don't know, ie. tcp syn flag or ports).  If the
	 * rule is also a fragment-specific rule, non-fragments won't
	 * match it. */
	offset = ntohs(ip->frag_off) & IP_OFFSET;

	switch (ip->protocol) {
	case IPPROTO_TCP:
		/* Don't care about things past flags word */
		size_req = 16; 
		/* Don't allow a fragment of TCP 8 bytes in. Nobody
		 * normal causes this. Its a cracker trying to break
		 * in by doing a flag overwrite to pass the direction
		 * checks.
		 */
		if (offset == 1) {
			if (net_ratelimit())
				printk("ip_tables: suspect TCP fragment.\n");
			return NF_DROP;
		}
		match = &tcp_packet_match;
		break;

	case IPPROTO_UDP:
		match = &udp_packet_match;
		size_req = 8;
		break;

	case IPPROTO_ICMP:
		match = &icmp_packet_match;
		size_req = 8;
		break;

	default:
		match = &ip_packet_match;
		size_req = 0;
	}
	/* If we can't investigate ports, treat as fragment.  It's
	 * either a trucated whole packet, or a truncated first
	 * fragment, or a TCP first fragment of length 8-15, in which
	 * case the above rule stops reassembly.
	 */
	offset |= (ntohs(ip->tot_len) < (ip->ihl<<2)+size_req);

	while (1) {
		unsigned int oldpos = (unsigned int)spos;
		struct ipt_kern_entry *e = &ipt_table_info->table[oldpos];
		*cache |= e->cacheinfo;
		if (match(e, indev, outdev, ip, offset, protohdr)) {
			struct per_slot *slots
				= ipt_table_info->slots[SLOT_NUMBER()];
			dprintf("We have a match on %u.\n", oldpos);
			ADD_COUNTER(&slots[oldpos].counters,
				    ntohs(ip->tot_len), 1);

			spos = e->act->action(skbp, indev, outdev, 
					      &e->info, spos);

			/* Pop from stack? */
			if (spos == -IPT_RETURN - 1) {
				dprintf("ipt_check: pop %u.\n", oldpos);
				spos = slots[oldpos].backpointer;
			}
			/* A verdict. */
			else if (spos < 0) break;
			/* Push onto stack? */
			else if (spos != (int)oldpos+1) {
				slots[spos].backpointer = oldpos;
				/* Overflow? */
				if (++i > number) {
					dprintf("ip_tables: Overflow!\n");
					spos = ipt_error(skbp, indev, outdev,
							 NULL, spos);
				}
			}
		}
		else spos++;
	}

#ifdef DEBUG_ALLOW_ALL
	return NF_ACCEPT;
#else
	return (unsigned int)(-spos - 1);
#endif
}

/* One level of recursion never hurt anyone. */
static struct ipt_action *find_action(const char *name, int tryload)
{
	struct ipt_action *act;

	for (act = ipt_action; act; act = act->next) {
		if (strcmp(name, act->name) == 0)
			return act;
	}

	if (tryload) {
		char modulename[IPT_FUNCTION_MAXNAMELEN + sizeof("ipt_")];
		strcpy(modulename, "ipt_");
		strcat(modulename, name);
		request_module(modulename);
		return find_action(name, 0);
	}
	return NULL;
}

static unsigned int get_cacheinfo(const struct ipt_kern_entry *newentry)
{
	unsigned int cache = 0;
	if (newentry->ip.smsk.s_addr) cache |= NFC_IP_SRC;
	if (newentry->ip.dmsk.s_addr) cache |= NFC_IP_DST;
	if (newentry->ip.iniface[0] == '\0' 
	    && !(newentry->ip.flags & IPT_F_EXACT_VIA_IN)) 
		cache |= NFC_IP_IF_IN;
	if (newentry->ip.outiface[0] == '\0' 
	    && !(newentry->ip.flags & IPT_F_EXACT_VIA_OUT)) 
		cache |= NFC_IP_IF_OUT;
	if (newentry->ip.flags & IPT_F_TOS) cache |= NFC_IP_TOS;
	if (newentry->ip.option) cache |= NFC_IP_OPTIONS;

	if (newentry->ip.proto) {
		cache |= NFC_IP_PROTO;
		switch (newentry->ip.proto) {
		case IPPROTO_ICMP:
			if (newentry->proto.icmp.type[0] != 0
			    || newentry->proto.icmp.type[1] != 255)
				cache |= NFC_IP_SRC_PT;
			if (newentry->proto.icmp.code[0] != 0
			    || newentry->proto.icmp.code[1] != 255)
				cache |= NFC_IP_DST_PT;
			break;

		case IPPROTO_UDP:
			if (newentry->proto.udp.spts[0] != 0
			    || newentry->proto.udp.spts[1] != 65535)
				cache |= NFC_IP_SRC_PT;
			if (newentry->proto.udp.dpts[0] != 0
			    || newentry->proto.udp.dpts[1] != 65535)
				cache |= NFC_IP_DST_PT;
			break;

		case IPPROTO_TCP:
			if (newentry->proto.tcp.spts[0] != 0
			    || newentry->proto.tcp.spts[1] != 65535)
				cache |= NFC_IP_SRC_PT;
			if (newentry->proto.tcp.dpts[0] != 0
			    || newentry->proto.tcp.dpts[1] != 65535)
				cache |= NFC_IP_DST_PT;
			if (newentry->proto.tcp.flg_mask)
				cache |= NFC_IP_TCPFLAGS;
			if (newentry->proto.tcp.option)
				cache |= NFC_IP_PROTO_UNKNOWN;
			break;
		}
	}
	return cache;
}

/* Checks the user-supplied table segment. */
static int translate_table(struct ipt_kern_entry *newtable,
			   unsigned int number,
			   struct per_slot firstslot[],
			   const struct ipt_entry entry[])
{
	unsigned int i;

	/* Each action name must be valid.  Each sanity check must pass. */
	for (i = 0; i < number; i++) {
		const struct ipt_entry *e = &entry[i];
		struct ipt_action *act = find_action(e->name, 1);

		if (!act) {
			duprintf("ip_tables: no entry by name `%s'.\n",
				 entry->name);
			return 0;
		}
		if (act->checkentry) {
			if (!act->checkentry(&e->action, i, number)) {
				duprintf("ip_tables: check failed for `%s'.\n",
					 e->name);
				return 0;
			}
		}
		/* Copy the details. */
		newtable[i] = ((struct ipt_kern_entry)
			       { e->ip, e->proto, act, e->action,
				 get_cacheinfo(&newtable[i]) });

		/* Set counters (this is the first slot). */
		firstslot[i].counters = e->counters;
	}

	return 1;
}

/* Gets counters. */
static void get_counters(unsigned int pos, 
			 unsigned int number,
			 struct ipt_counters *counters, 
			 const struct per_slot *slots,
			 unsigned int size)
{
	unsigned int i;

#ifdef DEBUG_IP_FIREWALL_LOCKING
	/* Need lock if operating on live data */
	if (slots == ipt_table_info->slots[0])
		IPTD_HAVE_LOCK(iptd_wlocks);
#endif

	for (i = 0; i < number; i++) {
		unsigned int j;

		counters[i].pcnt = counters[i].bcnt = 0;
		for (j = 0; j < NUM_SLOTS; j++) {
			ADD_COUNTER(&counters[i],
				    (slots + j*size)[pos+i].counters.bcnt,
				    (slots + j*size)[pos+i].counters.pcnt);
		}
	}
}

static void get_entries(unsigned int position, unsigned int number, 
			struct ipt_entry *entries)
{
	unsigned long flags;
	unsigned int i;

	/* FIXME: Check number inside lock --RR*/
	IPTD_WRITE_LOCK_IRQ(&ipt_lock, flags);
	for (i = position; i < position + number; i++) {
		/* Translate back to userland friendly. */
		struct ipt_kern_entry *kent = &ipt_table_info->table[i];
		struct ipt_entry *ent = &entries[i - position];

		memcpy(&ent->ip, &kent->ip, sizeof(ent->ip));
		memcpy(&ent->proto, &kent->proto, sizeof(ent->proto));
		memcpy(ent->name, kent->act->name, sizeof(ent->name));
		get_counters(i, 1, &ent->counters, ipt_table_info->slots[0],
			     ipt_table_info->number);
		memcpy(&ent->action, &kent->info, sizeof(ent->action));
	}
	IPTD_WRITE_UNLOCK_IRQ(&ipt_lock, flags);
}

static void add_counters(struct ipt_counters_info *paddc)
{
	unsigned long flags;
	unsigned int i;

	IPTD_WRITE_LOCK_IRQ(&ipt_lock, flags);
	for (i = 0; i < paddc->number; i++) {
		/* We're lazy, and add to the first slot; overflow
		 * works its fey magic and everything is OK. */
		ADD_COUNTER(&ipt_table_info->slots[0][i + paddc->position]
			    .counters,
			    paddc->counters[i].bcnt,
			    paddc->counters[i].pcnt);
	}
	IPTD_WRITE_UNLOCK_IRQ(&ipt_lock, flags);
}

/* Set error node and entry point backpointers to point to it */
static void setup_table_errors(struct ipt_kern_entry *table,
			       struct per_slot *slots,
			       unsigned int number)
{
	unsigned int i;

	memset(&table[number-1], 0, sizeof(table[number-1]));
	table[number-1].act = find_action(IPT_STANDARD_ERROR, 0);
	strcpy((char *)&table[number-1].info, "over/underflow");

	/* Set up backchains to point to error */
	for (i = 0; i < NUM_SLOTS*number; i++)
		slots[i].backpointer = number-1;
}

/* The error node automatically gets tacked on for them. */
static int ipt_replace_table(struct ipt_replace *repl)
{
	unsigned int i;
	unsigned int oldnumber, newnumber = repl->number + 1;
	struct ipt_kern_entry *newtable, *oldtable;
	struct per_slot *newslots, *oldslots;
	unsigned long flags;

	/* Check hooks. */
	for (i = 0; i < 3; i++) {
		if (repl->hook_entry[i] >= repl->number) {
			duprintf("Bad hook %u->%u.\n", i, repl->hook_entry[i]);
			goto bad_data;
		}
	}

	/* Allocate slots and new entries. */
	if (!multialloc(GFP_USER, 
			&newtable, newnumber * sizeof(struct ipt_kern_entry),
			&newslots, newnumber*NUM_SLOTS*sizeof(struct per_slot),
			(void *)NULL))
		return -ENOMEM;

	/* Zero counters. */
	memset(newslots, 0, newnumber*NUM_SLOTS*sizeof(struct per_slot));

	if (!translate_table(newtable, repl->number, newslots, repl->entries))
		goto bad_data;

	/* Increase module usage counts. */
	for (i = 0; i < repl->number; i++) {
		if (newtable[i].act->me)
			__MOD_INC_USE_COUNT(newtable[i].act->me);
	}

	/* Termination condition: unconditional error. */
	setup_table_errors(newtable, newslots, newnumber);

	/* Do the substitution. */
	IPTD_WRITE_LOCK_IRQ(&ipt_lock, flags);
	oldnumber = ipt_table_info->number;
	oldslots = ipt_table_info->slots[0];
	oldtable = ipt_table_info->table;
	ipt_table_info->number = newnumber;
	ipt_table_info->table = newtable;
	for (i = 0; i < NUM_SLOTS; i++) 
		ipt_table_info->slots[i] = newslots + i*newnumber;
	memcpy(ipt_table_info->hook_entry, repl->hook_entry, 
	       sizeof(ipt_table_info->hook_entry));
	IPTD_WRITE_UNLOCK_IRQ(&ipt_lock, flags);

	/* Get the old counters. */
	get_counters(0, repl->num_counters, repl->counters, 
		     oldslots, oldnumber);

	/* Decrease module usage counts */
	for (i = 0; i < oldnumber; i++) {
		if (oldtable[i].act->me) 
			__MOD_DEC_USE_COUNT(oldtable[i].act->me);
	}
	kfree(oldtable);
	kfree(oldslots);

	return 0;

 bad_data:
	kfree(newtable);
	kfree(newslots);
	return -EINVAL;
}

static int do_ipt_ctl(int cmd, void *user, unsigned int len)
{
	int ret;

	switch (cmd) {
	case IPT_SO_GET_INFO:
		if (len != sizeof(struct ipt_getinfo)) ret = -EINVAL;
		else {
			struct ipt_getinfo info 
				= { { 0, 0, 0 }, ipt_table_info->number };
			memcpy(info.hook_entry, ipt_table_info->hook_entry,
			       sizeof(info.hook_entry));
			if (copy_to_user(user, &info, len))
				ret = -EINVAL;
			else ret = 0;
		}
		break;

	case IPT_SO_REPLACE:
		if (len != sizeof(struct ipt_replace)) ret = -EINVAL;
		else {
			struct ipt_replace repl;
			struct ipt_entry *in;
			struct ipt_counters *out;

			/* One extra entry for error marker. */
			if (copy_from_user(&repl, user, sizeof(repl)))
				ret = -EFAULT;
			else if (!multialloc(GFP_USER, 
					     &in,
					     sizeof(struct ipt_entry)
					     * (repl.number + 1),
					     &out, 
					     sizeof(struct ipt_counters)
					     * repl.num_counters,
					     (void *)NULL))
				ret = -ENOMEM;
			else if (copy_from_user(in, repl.entries,
						sizeof(struct ipt_entry)
						* repl.number)) {
				kfree(in);
				kfree(out);
				ret = -EFAULT;
			}
			else {
				struct ipt_counters *userp = repl.counters;

				/* Replace user pointers with kernel ones. */
				repl.entries = in;
				repl.counters = out;
				ret = ipt_replace_table(&repl);
				if (ret == 0 
				    && copy_to_user(userp, out,
						    sizeof(struct ipt_counters)
						    * repl.num_counters))
					ret = -EFAULT;
				kfree(in);
				kfree(out);
			}
		}
		break;

	case IPT_SO_ADD_COUNTERS:
		if (len < sizeof(struct ipt_counters_info)) {
			duprintf("IPT_SO_*_COUNTERS: %u too small\n", len);
			ret = -EINVAL;
		}
		else {
			struct ipt_counters_info getc;
			struct ipt_counters_info *pgetc;

			/* FIXME: border checks under lock. */
			if (copy_from_user(&getc, user, sizeof(getc))) {
				duprintf("IPT_SO_*_COUNTERS: fault\n");
				ret = -EFAULT;
			}
			else if (getc.position + getc.number 
				 > ipt_table_info->number) {
				duprintf("IPT_SO_*_COUNTERS: over end %u\n",
					 ipt_table_info->number);
				ret = -EINVAL;
			}
			else if (len != sizeof(struct ipt_counters_info)
				 + getc.number*sizeof(struct ipt_counters)) {
				duprintf("IPT_SO_*_COUNTERS: bad length\n");
				ret = -EINVAL;
			}
			else if ((pgetc	= kmalloc(len, GFP_USER)) == NULL)
				ret = -ENOMEM;
			else {
				if (copy_from_user(pgetc, user, len))
					ret = -EFAULT;
				else {
					add_counters(pgetc);
					ret = 0;
				}
				kfree_s(pgetc, len);
			}
		}
		break;

	case IPT_SO_GET_ENTRIES:
		if (len < sizeof(struct ipt_get_entries)) {
			duprintf("IPT_SO_GET_ENTRIES: %u too small\n", len);
			ret = -EINVAL;
		} else {
			struct ipt_get_entries gete;
			struct ipt_entry *entries;

			/* FIXME: border checks under lock. */
			if (copy_from_user(&gete, user, sizeof(gete))) {
				duprintf("IPT_SO_GET_ENTRIES: fault\n");
				ret = -EFAULT;
			}
			else if (gete.position + gete.number 
				 > ipt_table_info->number) {
				duprintf("IPT_SO_GET_ENTRIES: over end\n");
				ret = -EINVAL;
			}
			else if (len != sizeof(struct ipt_get_entries)
				 + gete.number * sizeof(struct ipt_entry)) {
				duprintf("IPT_SO_GET_ENTRIES: bad length\n");
				ret = -EINVAL;
			}
			else if ((entries = kmalloc(gete.number
						    * sizeof(struct ipt_entry),
						    GFP_KERNEL)) == NULL)
				ret = -ENOMEM;
			else {
				get_entries(gete.position, gete.number, 
					    entries);
				if (copy_to_user(user+sizeof(gete), entries, 
						 gete.number 
						 * sizeof(struct ipt_entry)))
					ret = -EFAULT;
				else ret = 0;

				kfree(entries);
			}
		}
		break;

	default:
		duprintf("do_ipt_ctl:  unknown request %i\n", cmd);
		ret = -EINVAL;
	}

	return ret;
}

/* The work comes in here from netfilter.c. */
static unsigned int ipt_hook(unsigned int hook,
			     struct sk_buff **pskb,
			     const struct device *indev,
			     const struct device *outdev)
{
	unsigned int ret;

	dprintf("ipt_hook: hook %u.\n", hook);
	IPTD_READ_LOCK(&ipt_lock);
	ret = ipt_check(pskb, ipt_table_info->hook_entry[hook-NF_IP_INPUT], 
			indev ? indev->name : "", 
			outdev ? outdev->name : "", 
			&(*pskb)->nfcache);
	IPTD_READ_UNLOCK(&ipt_lock);
	dprintf("ipt_hook: returning %u.\n", ret);

	return ret;
}

/* Registration hooks for functions. */
int ipt_register_function(struct ipt_action *action)
{
	unsigned long flags;
	struct ipt_action **i;
	int ret = 0;

	IPTD_WRITE_LOCK_IRQ(&ipt_lock, flags);

	for (i = &ipt_action; *i; i = &(*i)->next) {
		if (strcmp((*i)->name, action->name) == 0) {
			ret = -EINVAL;
			goto out;
		}
	}
	action->next = NULL;
	*i = action;

	MOD_INC_USE_COUNT;

 out:
	IPTD_WRITE_UNLOCK_IRQ(&ipt_lock, flags);
	return ret;
}

int ipt_unregister_function(struct ipt_action *action)
{
	unsigned long flags;
	struct ipt_action **i;
	int ret = 0;

	IPTD_WRITE_LOCK_IRQ(&ipt_lock, flags);

	for (i = &ipt_action; *i; i = &(*i)->next) {
		if (*i == action) 
			break;
	}
	if (!*i) ret = -EINVAL;
	else {
		*i = action->next;
		MOD_DEC_USE_COUNT;
		/* Hmmm.... We could shrink table here maybe. */
	}

	IPTD_WRITE_UNLOCK_IRQ(&ipt_lock, flags);
	return ret;
}

static int ipt_standard_checkentry(const void *infovoid, 
				   unsigned int pos,
				   unsigned int finalpos)
{
	const struct ipt_standard *info = infovoid;

	/* Check standard info. */
	if (info->verdict < -IPT_RETURN - 1
	    || info->verdict > (int)finalpos) {
		duprintf("ipt_standard_check: bad verdict (%i)\n",
			 info->verdict);
		return 0;
	}
	return 1;
}

static struct nf_hook_ops ipt_ops[3];
static struct ipt_action ipt_standard_reg 
= { NULL, IPT_STANDARD_ACTION, ipt_standard_action, ipt_standard_checkentry, 
    NULL };
static struct ipt_action ipt_error_reg 
= { NULL, IPT_STANDARD_ERROR, ipt_error, NULL, NULL };
#ifdef __KERNEL__
static struct nf_setsockopt_ops ipt_sockopts 
= { NULL, PF_INET, IPT_BASE_CTL, IPT_SO_MAX, do_ipt_ctl };
#endif

/* Default to no forward for security reasons. */
static int forward = NF_DROP;
MODULE_PARM(forward, "i");

#define IPT_NR 4

static int init(void)
{
	unsigned int i;
	struct ipt_kern_entry *table;
	struct per_slot *slots;

#ifdef DEBUG_IP_FIREWALL_LOCKING
	iptd_wlocks = iptd_rlocks = 0;
#endif

	if (!multialloc(GFP_ATOMIC, 
			&ipt_table_info, 
			sizeof(*ipt_table_info) 
			+ NUM_SLOTS*sizeof(struct per_slot *),
			&table, 
			IPT_NR * sizeof(struct ipt_kern_entry),
			&slots,
			NUM_SLOTS*IPT_NR*sizeof(struct per_slot),
			NULL)) {
		printk("Unable to allocate IP firewall table.\n");
		return 0;
	}

	memset(ipt_table_info, 0, sizeof(*ipt_table_info));
	memset(table, 0, IPT_NR * sizeof(struct ipt_kern_entry));
	memset(slots, 0, NUM_SLOTS*IPT_NR*sizeof(struct per_slot));
	ipt_table_info->number = IPT_NR;
	ipt_table_info->table = table;

	/* Register standard function; doesn't count as usage. */
	ipt_register_function(&ipt_standard_reg);
	MOD_DEC_USE_COUNT;
	ipt_register_function(&ipt_error_reg);
	MOD_DEC_USE_COUNT;

	/* Hooks are sequential. */
	ipt_table_info->hook_entry[IPT_HOOK_INPUT] = 0;
	ipt_table_info->hook_entry[IPT_HOOK_FORWARD] = 1;
	ipt_table_info->hook_entry[IPT_HOOK_OUTPUT] = 2;

	/* Four rules: accept, forward rule, accept and error rule. */
	table[ipt_table_info->hook_entry[IPT_HOOK_INPUT]].act
		= &ipt_standard_reg;
	table[ipt_table_info->hook_entry[IPT_HOOK_INPUT]].info 
		= ((struct ipt_standard){ -NF_ACCEPT-1, 0, { 0 }});

	table[ipt_table_info->hook_entry[IPT_HOOK_FORWARD]].act
		= &ipt_standard_reg;
	table[ipt_table_info->hook_entry[IPT_HOOK_FORWARD]].info 
		= ((struct ipt_standard){ -forward-1, 0, { 0 } });

	table[ipt_table_info->hook_entry[IPT_HOOK_OUTPUT]].act
		= &ipt_standard_reg;
	table[ipt_table_info->hook_entry[IPT_HOOK_OUTPUT]].info 
		= ((struct ipt_standard){ -NF_ACCEPT-1, 0, { 0 } });

	/* Set up slots. */
	for (i = 0; i < NUM_SLOTS; i++) 
		ipt_table_info->slots[i] = slots + i*IPT_NR;

	setup_table_errors(table, ipt_table_info->slots[0], IPT_NR);

#ifdef __KERNEL__
	/* Register setsockopt */
	if (nf_register_sockopt(&ipt_sockopts) < 0) {
		printk("Unable to register sockopts.\n");
		return 0;
	}
#endif

	/* Register hooks */
	for (i = 0; i < 3; i++) {
		ipt_ops[i] = ((struct nf_hook_ops)
			      { NULL, 0, ipt_hook, NULL, PF_INET, 
				i+NF_IP_INPUT });
		if (nf_register_hook(&ipt_ops[i]) < 0) {
			printk("Unable to register iptable on hook %u.\n", 
			       i+NF_IP_INPUT);
			return 0;
		}
	}

#if defined(DEBUG_IP_FIREWALL) || defined(DEBUG_IP_FIREWALL_USER)
	printk("Firewall tables enabled! ``But I'm a boy''. \n");
#endif
	return 1;
}

#ifdef MODULE
int init_module(void)
{
	if (forward < 0 || forward > NF_MAX_VERDICT) {
		printk("Invalid forward value `%i'.\n", forward);
		return -EINVAL;
	}

	if (!init())
		return -EINVAL;

	return 0;
}

void cleanup_module(void)
{
	unsigned long flags;
	unsigned int i;

	IPTD_WRITE_LOCK_IRQ(&ipt_lock, flags);

	for (i = 0; i < 3; i++) {
		if (nf_unregister_hook(&ipt_ops[i]) != 0)
			printk("ip_tables: Can't unregister hook %u!\n", i);
	}
	nf_unregister_sockopt(&ipt_sockopts);
	kfree(ipt_table_info->table);
	kfree(ipt_table_info->slots[0]);
	kfree(ipt_table_info);

	IPTD_WRITE_UNLOCK_IRQ(&ipt_lock, flags);
}
#else /* !MODULE */
__initfunc(void ip_fw_init(void))
{
	if (!init()) 
		panic("Failed to initialize iptables\n");
}

#endif
