/*
 * 25-Jul-1998 Major changes to allow for ip chain table 
 */

/*
 * 	Format of an IP firewall descriptor
 *
 * 	src, dst, src_mask, dst_mask are always stored in network byte order.
 * 	flags are stored in host byte order (of course).
 * 	Port numbers are stored in HOST byte order.
 */

#ifndef _IP_TABLES_H
#define _IP_TABLES_H

#ifdef __KERNEL__
#include <linux/if.h>
#include <linux/types.h>
#include <linux/icmp.h>
#include <linux/in.h>
#include <linux/ip.h>
#include <linux/tcp.h>
#include <linux/udp.h>
#endif
#include <linux/ip_netfilter.h>
#include "ip_tables_standard.h"

#define IPT_FUNCTION_MAXNAMELEN 16

struct ipt_ip {
	/* Source and destination IP addr */
	struct in_addr src, dst;		
	/* Mask for src and dest IP addr */
	struct in_addr smsk, dmsk;
	char iniface[IFNAMSIZ], outiface[IFNAMSIZ];

	/* Protocol, 0 = ANY */
	u_int16_t proto;
	/* TOS to match iff flags & IPT_F_TOS */
	u_int8_t tos;

	/* IP Option to match: 0 = any or none. */
	u_int8_t option;

	/* Flags word */
	u_int8_t flags;
	/* Inverse flags */
	u_int8_t invflags;
};

/* tcp specific stuff */
struct ipt_tcp
{
	u_int16_t spts[2];			/* Source port range. */
	u_int16_t dpts[2];			/* Destination port range. */
	u_int8_t flg_mask;			/* TCP flags mask byte */
	u_int8_t flg_cmp;			/* TCP flags compare byte */
	u_int8_t option;			/* TCP option to match 0=any */
	u_int8_t invflags;			/* Inverse flags */
};

/* udp specific stuff */
struct ipt_udp
{
	u_int16_t spts[2];			/* Source port range. */
	u_int16_t dpts[2];			/* Destination port range. */
	u_int8_t invflags;			/* Inverse flags */
};

/* icmp specific stuff */
struct ipt_icmp
{
	u_int8_t type[2];			/* range of type */
	u_int8_t code[2];			/* range of code */
	u_int8_t invflags;			/* Inverse flags */
};

struct ipt_counters
{
	u_int64_t pcnt, bcnt;			/* Packet and byte counters */
};

union ipt_proto
{
	struct ipt_tcp tcp;
	struct ipt_udp udp;
	struct ipt_icmp icmp;
};

/* This structure defines each of the firewall rules.  Consists of 3
   parts which are 1) general IP header stuff 2) protocol specific
   stuff 3) the action to perform if the rule matches */
struct ipt_entry
{
	struct ipt_ip ip;
	union ipt_proto proto;
	/* Used by userspace: name of action to take. */
	char name[IPT_FUNCTION_MAXNAMELEN];
	/* Packet and byte counters. */
	struct ipt_counters counters;

	struct ipt_standard action;
};

/* Values for "flag" field in struct ipt_ip (general ip structure). */
#define IPT_F_FRAG		0x01	/* Set if rule is a fragment rule */
#define IPT_F_EXACT_VIA_IN	0x02	/* Exact iniface match */
#define IPT_F_EXACT_VIA_OUT	0x04	/* Exact outiface match */
#define IPT_F_TOS		0x08	/* Match the TOS. */

#define IPT_F_MASK		0x0F	/* All possible flag bits mask. */

/* Values for "inv" field in struct ipt_ip. */
#define IPT_INV_VIA_IN		0x01	/* Invert the sense of IN IFACE. */
#define IPT_INV_VIA_OUT		0x02	/* Invert the sense of OUT IFACE */
#define IPT_INV_TOS		0x04	/* Invert the sense of TOS. */
#define IPT_INV_SRCIP		0x08	/* Invert the sense of SRC IP. */
#define IPT_INV_DSTIP		0x10	/* Invert the sense of DST OP. */
#define IPT_INV_FRAG		0x20	/* Invert the sense of FRAG. */
#define IPT_INV_PROTO		0x40	/* Invert the sense of PROTO. */
#define IPT_INV_OPTION		0x80	/* Invert the sense of option. */

/* Values for "inv" field in struct ipt_tcp. */
#define IPT_TCP_INV_SRCPT	0x01	/* Invert the sense of source ports. */
#define IPT_TCP_INV_DSTPT	0x02	/* Invert the sense of dest ports. */
#define IPT_TCP_INV_OPTION	0x04	/* Invert the sense of TCP option. */
#define IPT_TCP_INV_FLAGS	0x08	/* Invert the sense of TCP flags. */

/* Values for "inv" field in struct ipt_udp. */
#define IPT_UDP_INV_SRCPT	0x01	/* Invert the sense of source ports. */
#define IPT_UDP_INV_DSTPT	0x02	/* Invert the sense of dest ports. */

/* Values for "inv" field for struct ipt_icmp. */
#define IPT_ICMP_INV		0x01	/* Invert the sense of type test */ 

/*
 *	New IP firewall options for [gs]etsockopt at the RAW IP level.
 *	Unlike BSD Linux inherits IP options so you don't have to use
 * a raw socket for this. Instead we check rights in the calls. */

#define IPT_BASE_CTL		64	/* base for firewall socket options */

#define IPT_SO_REPLACE		(IPT_BASE_CTL)
#define IPT_SO_GET_INFO		(IPT_BASE_CTL + 1)
#define IPT_SO_GET_ENTRIES	(IPT_BASE_CTL + 2)
#define IPT_SO_ADD_COUNTERS	(IPT_BASE_CTL + 3)
#define IPT_SO_MAX		IPT_SO_ADD_COUNTERS

/* RETURN verdict */
#define IPT_RETURN (NF_MAX_VERDICT+1)

#define IPT_HOOK_INPUT (NF_IP_INPUT - 2)
#define IPT_HOOK_FORWARD (NF_IP_FORWARD - 2)
#define IPT_HOOK_OUTPUT (NF_IP_OUTPUT - 2)

/* The argument to IPT_SO_GET_INFO */
struct ipt_getinfo
{
	/* Hook entry points: INPUT, FORWARD and OUTPUT */
	unsigned int hook_entry[3];

	/* Number of entries. */
	unsigned int num_entries;
};

/* The argument to IPT_SO_REPLACE. */
struct ipt_replace
{
	/* Hook entry points. */
	unsigned int hook_entry[3];

	/* Information about entries to be inserted: */
	/* Number of entries. */
	unsigned int number;
	/* The entries. */
	struct ipt_entry *entries;

	/* Information about entries' counters to saved: */
	/* Number of entries. */
	unsigned int num_counters;
	/* The counters. */
	struct ipt_counters *counters;
};

/* The argument to IPT_SO_ADD_COUNTERS. */
struct ipt_counters_info
{
	/* Number of entries. */
	unsigned int number;
	/* Where to add them. (0 = start) */
	unsigned int position;
	/* The counters (actually `number' of these). */
	struct ipt_counters counters[0];
};

/* The argument to IPT_SO_GET_ENTRIES. */
struct ipt_get_entries
{
	/* Number of entries. */
	unsigned int number;
	/* Where to get them. (0 = start) */
	unsigned int position;
	/* The entries (actually `number' of these). */
	struct ipt_entry entries[0];
};

/* The argument to IPT_CHECK. */
struct ipt_check
{
	unsigned int insertion_point;	/* Where in table to start packet */
	const char indev[IFNAMSIZ];
	const char outdev[IFNAMSIZ];
	struct iphdr ip;		/* IP header */
	char ipbody[0];			/* Hangs off end. */
};

/*
 *	Main firewall chains definitions and global var's definitions.
 */

#if defined(__KERNEL__) || defined (IP_TABLES_USER)

#include <linux/config.h>
#include <linux/init.h>
extern void ipt_init(void) __init;

/* Registration hooks for actions. */
struct ipt_action
{
	struct ipt_action *next;

	const char name[IPT_FUNCTION_MAXNAMELEN];
	/* Return -verdict - 1 or a position. */
	int (*action)(struct sk_buff **skb, 
		      const char *indev,
		      const char *outdev,
		      void *info, 
		      unsigned int pos);
	/* Called when user tries to insert an entry of this type. */
	/* Should return non-zero for OK. */
	int (*checkentry)(const void *info, unsigned int pos, 
			  unsigned int finalpos);

	/* Set this to __this_module if you are a module, otherwise NULL */
	struct module *me;
};

int ipt_register_function(struct ipt_action *action);
int ipt_unregister_function(struct ipt_action *action);
#endif /*__KERNEL__*/
#endif /* _IP_TABLES_H */
