/*
 * Author: Paul.Russell@rustcorp.com.au and mneuling@radlogic.com.au
 *
 * Based on the ipchains code by Paul Russell and Michael Neuling
 *
 *	iptables -- IP firewall administration for kernels with
 * firewall table (aimed for the 2.3 kernels)
 *
 *	See the accompanying manual page iptables(8) for information
 *	about proper usage of this program.
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation; either version 2 of the License, or
 *	(at your option) any later version.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * History:
 * 1.0: First release -- based on ipchains.c
 */

#include <stdio.h>
#include <stdarg.h>
#include <errno.h>
#include <limits.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <ctype.h>
#include <getopt.h>
#include <netdb.h>
#include <sys/param.h>
#include <pwd.h>
#include <sys/types.h>
#include <dlfcn.h>

#include "libipfwc/libipfwc.h"

#ifndef TRUE
#define TRUE 1
#endif
#ifndef FALSE
#define FALSE 0
#endif

#ifndef IPT_LIB_DIR
#define IPT_LIB_DIR "/usr/local/lib/iptables"
#endif

#define CMD_NONE		0x000U
#define CMD_INSERT		0x001U
#define CMD_DELETE		0x002U
#define CMD_DELETE_NUM		0x004U
#define CMD_REPLACE		0x008U
#define CMD_APPEND		0x010U
#define CMD_LIST		0x020U
#define CMD_FLUSH		0x040U
#define CMD_ZERO		0x080U
#define CMD_NEW_CHAIN		0x100U
#define CMD_DELETE_CHAIN	0x200U
#define CMD_SET_POLICY		0x400U
#define CMD_CHECK		0x800U
#define NUMBER_OF_CMD	12
static const char cmdflags[] = { 'I', 'D', 'D', 'R', 'A', 'L', 'F', 'Z',
				 'N', 'X', 'P', 'M', 'S', 'C' };

#define OPT_NONE	0x00000U
#define OPT_NUMERIC	0x00001U
#define OPT_SOURCE	0x00002U
#define OPT_DESTINATION	0x00004U
#define OPT_PROTOCOL	0x00008U
#define OPT_JUMP	0x00010U
#define OPT_TCPSYN	0x00020U
#define OPT_BIDIR	0x00040U
#define OPT_VERBOSE	0x00080U
#define OPT_PRINTK	0x00100U
#define OPT_EXPANDED	0x00200U
#define OPT_TOS		0x00400U
#define OPT_VIANAMEIN	0x00800U
#define OPT_VIANAMEOUT	0x01000U
#define OPT_FRAGMENT    0x02000U
#define OPT_MARK        0x04000U
#define OPT_SRCPT       0x08000U
#define OPT_DSTPT       0x10000U
#define OPT_NETLINK     0x20000U
#define OPT_LINENUMBERS 0x40000U
#define NUMBER_OF_OPT	19
static const char optflags[] = { 'n', 's', 'd', 'p', 'j', 'y', 'b', 'v',
				 'l', 'x', 't', 'i', 'f', 'm', 's', 'd', 'o', '3'};

#define FMT_NUMERIC	0x0001
#define FMT_NOCOUNTS	0x0002
#define FMT_KILOMEGAGIGA 0x0004
#define FMT_OPTIONS	0x0008
#define FMT_NOTABLE	0x0010
#define FMT_NOTARGET	0x0020
#define FMT_VIA		0x0040
#define FMT_NONEWLINE	0x0080
#define FMT_TOS		0x0100
#define FMT_MARK        0x0200
#define FMT_NETLINK     0x0400
#define FMT_LINENUMBERS 0x0800

#define FMT_PRINT_RULE (FMT_NOCOUNTS | FMT_OPTIONS | FMT_TOS | FMT_VIA \
			| FMT_NUMERIC | FMT_NOTABLE | FMT_MARK)

struct option original_opts[] = {
	{ "append", 1, 0, 'A' },
	{ "delete", 1, 0,  'D' },
	{ "insert", 1, 0,  'I' },
	{ "replace", 1, 0,  'R' },
	{ "list", 2, 0,  'L' },
	{ "flush", 2, 0,  'F' },
	{ "zero", 2, 0,  'Z' },
	{ "check", 1, 0,  'C' },
	{ "new-chain", 1, 0,  'N' },
	{ "delete-chain", 1, 0,  'X' },
	{ "policy", 1, 0,  'P' },
	{ "set", 1, 0,  'S' },
	{ "bidirectional", 0, 0,  'b' },
	{ "source", 1, 0, 's' },
	{ "destination", 1, 0,  'd' },
	{ "src", 1, 0,  's' }, /* synonym */
	{ "dst", 1, 0,  'd' }, /* synonym */
	{ "source-port", 1, 0, '1' },
	{ "destination-port", 1, 0, '2' },
	{ "icmp-type", 1, 0, '1' },
	{ "sport", 1, 0, '1' }, /* synonym */
	{ "dport", 1, 0, '2' }, /* synonym */
	{ "proto", 1, 0,  'p' },
	{ "in", 1, 0, 'i' }, 
	{ "jump", 1, 0, 'j' },
	{ "TOS", 1, 0, 't' },
	{ "mark", 1, 0, 'm' },
	{ "numeric", 0, 0, 'n' },
	{ "log", 0, 0, 'l' },
	{ "out", 2, 0, 'o' }, 
	{ "verbose", 0, 0, 'v' },
	{ "exact", 0, 0, 'x' },
	{ "fragments", 0, 0, 'f' },
	{ "syn", 0, 0, 'y' },
	{ "version", 0, 0, 'V' },
	{ "help", 0, 0, 'h' },
	{ "line-numbers", 0, 0, '3' },
	{ "rawinfo", 0, 0, 'r' },
	{0}
};

/* TOS names and values. */
struct TOS_value
{
	unsigned char TOS;
	const char *name;
} TOS_values[] = {
#ifdef NOT_A_GODDAMN_YANK
	{ 0x10, "Minimise-Delay" },
	{ 0x08, "Minimise-Throughput" },
	{ 0x04, "Maximise-Reliability" },
	{ 0x02, "Minimise-Cost" },
#else
	{ 0x10, "Minimize-Delay" },
	{ 0x08, "Minimize-Throughput" },
	{ 0x04, "Maximize-Reliability" },
	{ 0x02, "Minimize-Cost" },
#endif
	{ 0x00, "Normal Service" },
};

/* Table of legal combinations of commands and options.  If any of the
 * given commands make an option legal, that option is legal (applies to
 * CMD_LIST and CMD_ZERO only).
 * Key:
 *  +  compulsory
 *  x  illegal
 *     optional
 */

static char commands_v_options[NUMBER_OF_CMD][NUMBER_OF_OPT] =
/* Well, it's better than "Re: Linux vs FreeBSD" */
{
	/*     -n  -s  -d  -p  -j  -y  -b  -v  -l  -x  -t  -i  -f  -m  -s  -d  -o  -3*/
/*INSERT*/    {'x',' ',' ',' ',' ',' ',' ',' ',' ','x',' ',' ',' ',' ',' ',' ',' ','x'},
/*DELETE*/    {'x',' ',' ',' ',' ',' ',' ',' ',' ','x',' ',' ',' ',' ',' ',' ',' ','x'},
/*DELETE_NUM*/{'x','x','x','x','x','x','x',' ','x','x','x','x','x','x','x','x','x','x'},
/*REPLACE*/   {'x',' ',' ',' ',' ',' ','x',' ',' ','x',' ',' ',' ',' ',' ',' ',' ','x'},
/*APPEND*/    {'x',' ',' ',' ',' ',' ',' ',' ',' ','x',' ',' ',' ',' ',' ',' ',' ','x'},
/*LIST*/      {' ','x','x','x','x','x','x',' ','x',' ','x','x','x','x','x','x','x',' '},
/*FLUSH*/     {'x','x','x','x','x','x','x',' ','x','x','x','x','x','x','x','x','x','x'},
/*ZERO*/      {'x','x','x','x','x','x','x',' ','x','x','x','x','x','x','x','x','x','x'},
/*NEW_CHAIN*/ {'x','x','x','x','x','x','x',' ','x','x','x','x','x','x','x','x','x','x'},
/*DEL_CHAIN*/ {'x','x','x','x','x','x','x',' ','x','x','x','x','x','x','x','x','x','x'},
/*SET_POLICY*/{'x','x','x','x','x','x','x',' ','x','x','x','x','x','x','x','x','x','x'},
/*CHECK*/     {'x','+','+','+','x',' ',' ',' ','x','x','x','+',' ','x',' ',' ','x','x'}
};

static char package_version[] = "iptables 1.0.1, 13-Dec-1998";

static const char *program;

/* Do we invert the next option? */
static int invert = FALSE;

static int check_inverse(const char option[], int *invert);

/* Parsing functions for command line. */
static struct in_addr *parse_hostnetwork(const char *, unsigned int *);
static void parse_hostnetworkmask(const char *, struct in_addr **,
	struct in_addr *, unsigned int *);
static struct in_addr *parse_mask(char *);
static unsigned short int parse_protocol(const char *);
static unsigned short int parse_tos(const char *);
static const char * parse_policy(const char *chain, const char *policy);
static void parse_icmp(const char *srcportstring, const char *dstportstring,
		       u_int16_t *type, u_int16_t* code, int *options, 
		       u_int16_t invflags, int checking);
static void parse_ports(const char *portstring, u_int16_t *ports, 
			u_int16_t proto);
static u_int16_t parse_port(const char *port, unsigned short proto);
static void parse_mark(const char *markstring, u_int32_t *mark);
static unsigned short parse_interface(const char *ifstring, 
				      char *vianame, 
				      unsigned short flag);
static const char *parse_target(const char *targetname);
static int parse_rulenumber(const char *rule);

/* Various parsing helper functions */
static struct in_addr *host_to_addr(const char *, unsigned int *);
static char *addr_to_host(const struct in_addr *);
static struct in_addr *network_to_addr(const char *);
static char *addr_to_network(const struct in_addr *);
static char *addr_to_anyname(const struct in_addr *);
static struct in_addr *dotted_to_addr(const char *);
static char *addr_to_dotted(const struct in_addr *);
static char *mask_to_dotted(const struct in_addr *);
static int service_to_port(const char *, unsigned short);
static char *port_to_service(int, unsigned short);
static int string_to_number(const char *, int, int);

/* Check the options based on the tables above */
static void generic_opt_check(int command, int options);
static char opt2char(int option);
static char cmd2char(int option);
static void add_command(int *cmd, const int newcmd, const int othercmds);

static void set_option(unsigned int *options,
		       unsigned int option,
		       int invert, 
		       unsigned int *invoptions);

static void inaddrcpy(struct in_addr *, struct in_addr *);
static void *fw_malloc(size_t);
static void *fw_calloc(size_t, size_t);
enum exittype {
	OTHER_PROBLEM = 1,
	PARAMETER_PROBLEM,
	VERSION_PROBLEM
};
static void exit_error(enum exittype, char *, ...)
  __attribute__((noreturn, format(printf,2,3)));
static void exit_tryhelp(int) __attribute__((noreturn));
static void exit_printicmphelp() __attribute__((noreturn));
static void exit_printhelp() __attribute__((noreturn));


struct icmp_names {
	const char *name;
	u_int8_t type;
	u_int8_t code_min, code_max;
	int use_code_min_for_testing;
};

const struct icmp_names icmp_codes[] = {
	{ "echo-reply", 0, 0, 0xFF, TRUE },
	/* Alias */ { "pong", 0, 0, 0xFF, TRUE },

	{ "destination-unreachable", 3, 0, 0xFF, FALSE },
	{   "network-unreachable", 3, 0, 0, TRUE },
	{   "host-unreachable", 3, 1, 1, TRUE },
	{   "protocol-unreachable", 3, 2, 2, TRUE },
	{   "port-unreachable", 3, 3, 3, TRUE },
	{   "fragmentation-needed", 3, 4, 4, TRUE },
	{   "source-route-failed", 3, 5, 5, TRUE },
	{   "network-unknown", 3, 6, 6, TRUE },
	{   "host-unknown", 3, 7, 7, TRUE },
	{   "network-prohibited", 3, 9, 9, TRUE },
	{   "host-prohibited", 3, 10, 10, TRUE },
	{   "TOS-network-unreachable", 3, 11, 11, TRUE },
	{   "TOS-host-unreachable", 3, 12, 12, TRUE },
	{   "communication-prohibited", 3, 13, 13, TRUE },
	{   "host-precedence-violation", 3, 14, 14, TRUE },
	{   "precedence-cutoff", 3, 15, 15, TRUE },

	{ "source-quench", 4, 0, 0xFF, TRUE },

	{ "redirect", 5, 0, 0xFF, FALSE },
	{   "network-redirect", 5, 0, 0, TRUE },
	{   "host-redirect", 5, 1, 1, TRUE },
	{   "TOS-network-redirect", 5, 2, 2, TRUE },
	{   "TOS-host-redirect", 5, 3, 3, TRUE },

	{ "echo-request", 8, 0, 0xFF, TRUE },
	/* Alias */ { "ping", 8, 0, 0xFF, TRUE },

	{ "router-advertisement", 9, 0, 0xFF, TRUE },

	{ "router-solicitation", 10, 0, 0xFF, TRUE },

	{ "time-exceeded", 11, 0, 0xFF, FALSE },
	/* Alias */ { "ttl-exceeded", 11, 0, 0xFF, FALSE },
	{   "ttl-zero-during-transit", 11, 0, 0, TRUE },
	{   "ttl-zero-during-reassembly", 11, 1, 1, TRUE },

	{ "parameter-problem", 12, 0, 0xFF, FALSE },
	{   "ip-header-bad", 12, 0, 0, TRUE },
	{   "required-option-missing", 12, 1, 1, TRUE },

	{ "timestamp-request", 13, 0, 0xFF, TRUE },

	{ "timestamp-reply", 14, 0, 0xFF, TRUE },

	{ "address-mask-request", 17, 0, 0xFF, TRUE },

	{ "address-mask-reply", 18, 0, 0xFF, TRUE }
};

#define FMT(tab,notab) ((format & FMT_NOTABLE) ? notab : tab)

static void
print_header(FILE *fp, unsigned int format, const char *chain, 
	     ipfwc_handle_t *handle)
{
	struct ipt_counters counters;
	const char *pol = ipfwc_get_policy(chain, &counters, handle);
	fprintf(fp, "Chain %s", chain);
	if (pol) {
		fprintf(fp, " (policy %s", pol);
		if (!(format & FMT_NOCOUNTS)) 
			fprintf(fp, " %llu packets, %llu bytes",
				counters.pcnt, counters.bcnt);
		fprintf(fp, ")\n");
	} else {
		unsigned int refs;
		ipfwc_get_references(&refs, chain, handle);
		printf(" (%u references)\n", refs);
	}

	if (format & FMT_LINENUMBERS)
		fprintf(fp, FMT("%-4s ", "%s "), "num");
	if (!(format & FMT_NOCOUNTS)) {
		if (format & FMT_KILOMEGAGIGA) {
			fprintf(fp, FMT("%5s ","%s "), "pkts");
			fprintf(fp, FMT("%5s ","%s "), "bytes");
		} else {
			fprintf(fp, FMT("%8s ","%s "), "pkts");
			fprintf(fp, FMT("%10s ","%s "), "bytes");
		}
	}
	if (!(format & FMT_NOTARGET)) {
		fprintf(fp, FMT("%-9s ","%s "), "target");
	}
	fputs(" prot ", fp);
	if (format & FMT_OPTIONS)
		fputs("opt  ", fp);
	if (format & FMT_TOS)
		fputs(" tos  ", fp);
	if (format & FMT_VIA) {
		fprintf(fp, FMT("%-9s ","%s "), "in");
		fprintf(fp, FMT("%-9s ","%s "), "out");
	}
	if (format & FMT_MARK)
		fprintf(fp, FMT("%-10s ","%s "), "mark");
	fprintf(fp, FMT(" %-20s ","%s "), "source");
	fprintf(fp, FMT(" %-20s ","%s "), "destination");
	fputs(" ports\n", fp);
}

static void
print_num(FILE *fp, u_int64_t number, unsigned int format)
{
	if (format & FMT_KILOMEGAGIGA) {
		if (number > 99999) {
			number = (number + 500) / 1000;
			if (number > 9999) {
				number = (number + 500) / 1000;
				if (number > 9999) {
					number = (number + 500) / 1000;
					fprintf(fp,
						FMT("%4lluG ","%lluG "), 
						number);
				}
				else fprintf(fp, FMT("%4lluM ","%lluM "), 
						number);
			} else
				fprintf(fp, FMT("%4lluK ","%lluK "), number);
		} else
			fprintf(fp, FMT("%5llu ","%llu "), number);
	} else
		fprintf(fp, FMT("%8llu ","%llu "), number);
}

/* A few hardcoded protocols for 'all' and in case the user has no
   /etc/protocols */
struct pprot {
	char *name;
	unsigned short num;
};

static const struct pprot chain_protos[] = {
	{ "all", 0 },
	{ "tcp", IPPROTO_TCP },
	{ "udp", IPPROTO_UDP },
	{ "icmp", IPPROTO_ICMP },
	{0}
};

static char *
proto_to_name(unsigned short proto)
{
	const struct pprot *pp;

	if (proto) {
		struct protoent *pent = getprotobynumber(proto);
		if (pent)
			return pent->p_name;
	}

	for (pp = &chain_protos[0]; pp->name; pp++)
		if (pp->num == proto)
			return pp->name;

	return NULL;
}

static void
print_port(FILE *fp, int numeric, u_int16_t port, u_int16_t proto)
{
	char *service;
	if (numeric || (service = port_to_service(port, proto)) == NULL)
		fprintf(fp, "%u", port);
	else
		fputs(service, fp);
}

/* FIXME: Non-numeric TOS names. --RR */
static void
print_firewall(FILE *fp,
	       const struct ipt_entry *fw, 
	       unsigned int num,
	       const char *target,
	       unsigned int format)
{
	unsigned short flags;
	char buf[BUFSIZ];
	flags = fw->ip.flags;

	if (format & FMT_LINENUMBERS)
		fprintf(fp, FMT("%-4u ", "%u "), num);

	if (!(format & FMT_NOCOUNTS)) {
		print_num(fp, fw->counters.pcnt, format);
		print_num(fp, fw->counters.bcnt, format);
	}

	if (!(format & FMT_NOTARGET))
		fprintf(fp, FMT("%-9s ", "%s "), target);

	fputc(fw->ip.invflags & IPT_INV_PROTO ? '!' : ' ', fp);
	{
		char *pname = proto_to_name(fw->ip.proto);
		if (pname)
			fprintf(fp, FMT("%-5s", "%s "), pname);
		else
			fprintf(fp, FMT("%-5hu", "%hu "), fw->ip.proto);
	}

	if (format & FMT_OPTIONS) {
		if (format & FMT_NOTABLE)
			fputs("opt    ", fp);
		fputc((fw->proto.tcp.invflags & IPT_TCP_INV_FLAGS) ? '!' : '-', fp);
		fputc((fw->proto.tcp.flg_cmp) ? 'y' : '-', fp);
		fputc((fw->ip.invflags & IPT_INV_FRAG) ? '!' : '-', fp);
		fputc((flags & IPT_F_FRAG) ? 'f' : '-', fp);
		fputc(' ', fp);
	}

	if (format & FMT_TOS) {
		if (format & FMT_NOTABLE)
			fputs("tos ", fp);
		if (fw->ip.invflags & IPT_INV_TOS)
			fputs("!", fp);
		else 
			fputs(" ", fp);
		if (fw->ip.flags & IPT_F_TOS)
			fprintf(fp, " *   ");
		else
			fprintf(fp, "0x%02hX ", (unsigned short) fw->ip.tos);
	}

	if (format & FMT_VIA) {
		char iface[IFNAMSIZ+2];

		if (fw->ip.invflags & IPT_INV_VIA_IN) {
			iface[0] = '!';
			iface[1] = '\0';
		} 
		else iface[0] = '\0';

		if (fw->ip.iniface[0] != '\0') {
			strcat(iface, fw->ip.iniface);
			if (!(fw->ip.flags&IPT_F_EXACT_VIA_IN))
			    strcat(iface, "+");
		} 
		else if (format & FMT_NUMERIC) strcat(iface, "*");
		else strcat(iface, "any");
		fprintf(fp, FMT("%-9.16s ","in %.16s "), iface);

		if (fw->ip.invflags & IPT_INV_VIA_OUT) {
			iface[0] = '!';
			iface[1] = '\0';
		} 
		else iface[0] = '\0';

		if (fw->ip.outiface[0] != '\0') {
			strcat(iface, fw->ip.outiface);
			if (!(fw->ip.flags&IPT_F_EXACT_VIA_OUT))
			    strcat(iface, "+");
		} 
		else if (format & FMT_NUMERIC) strcat(iface, "*");
		else strcat(iface, "any");
		fprintf(fp, FMT("%-9.16s ","out %.16s "), iface);
	}

	if ((format & FMT_MARK) 
	    && strcmp(fw->name, IPT_STANDARD_ACTION) == 0
	    && fw->action.mark)
		fprintf(fp, FMT("0x%x ","mark 0x%x "), fw->action.mark);
	else 
		fputs(FMT("           ", ""), fp);

	fputc(fw->ip.invflags & IPT_INV_SRCIP ? '!' : ' ', fp);
	if (fw->ip.smsk.s_addr == 0L && !(format & FMT_NUMERIC))
		fprintf(fp, FMT("%-20s ","%s "), "anywhere");
	else {
		if (format & FMT_NUMERIC)
			sprintf(buf, "%s", addr_to_dotted(&(fw->ip.src)));
		else
			sprintf(buf, "%s", addr_to_anyname(&(fw->ip.src)));
		strcat(buf, mask_to_dotted(&(fw->ip.smsk)));
		fprintf(fp, FMT("%-20s ","%s "), buf);
	}

	fputc(fw->ip.invflags & IPT_INV_DSTIP ? '!' : ' ', fp);
	if (fw->ip.dmsk.s_addr == 0L && !(format & FMT_NUMERIC))
		fprintf(fp, FMT("%-20s","-> %s"), "anywhere");
	else {
		if (format & FMT_NUMERIC)
			sprintf(buf, "%s", addr_to_dotted(&(fw->ip.dst)));
		else
			sprintf(buf, "%s", addr_to_anyname(&(fw->ip.dst)));
		strcat(buf, mask_to_dotted(&(fw->ip.dmsk)));
		fprintf(fp, FMT("%-20s","-> %s"), buf);
	}

	if (format & FMT_NOTABLE)
		fputs("  ", fp);

	if (fw->ip.proto != IPPROTO_TCP
	    && fw->ip.proto != IPPROTO_UDP
	    && fw->ip.proto != IPPROTO_ICMP) {
		fputs("  n/a", fp);
		if (!(format & FMT_NONEWLINE))
			putc('\n', fp);
		return;
	}

	/* ICMP handled specially. */
	if (fw->ip.proto == IPPROTO_ICMP) {
		const char *name = NULL;

		if (!(fw->proto.icmp.invflags & IPT_ICMP_INV)
		    && !(format & FMT_NUMERIC)) {
			unsigned int i;
			for (i = 0;
			     i < sizeof(icmp_codes)/sizeof(struct icmp_names);
			     i++) {
				if(icmp_codes[i].type == fw->proto.icmp.type[0]
				   &&icmp_codes[i].type==fw->proto.icmp.type[1]
				   && (icmp_codes[i].code_min 
				       == fw->proto.icmp.code[0])
				   && (icmp_codes[i].code_max
				       == fw->proto.icmp.code[1])) {
					name = icmp_codes[i].name;
					break;
				}
			}
		}

		if (name) fprintf(fp, "  %s", name);
		else {
			if (fw->proto.icmp.invflags & IPT_ICMP_INV)
				fputs(" !", fp);
			else 
				fputs("  ", fp);

			/* If there are restrictions... */
			if (fw->proto.icmp.type[0] != 0
			    || fw->proto.icmp.type[1] != 0xFF) {
				if (fw->proto.icmp.type[0]
				    == fw->proto.icmp.type[1])
					fprintf(fp, "type %u ",
						fw->proto.icmp.type[0]);
				else
					fprintf(fp, "types %u-%u ", 
						fw->proto.icmp.type[0],
						fw->proto.icmp.type[1]);
			}

			if (fw->proto.icmp.code[0] != 0
			    || fw->proto.icmp.code[1] != 0xFF) {
				if (fw->proto.icmp.code[0]
				    == fw->proto.icmp.code[1])
					fprintf(fp, "code %u ",
						fw->proto.icmp.code[0]);
				else
					fprintf(fp, "codes %u-%u ", 
						fw->proto.icmp.code[0],
						fw->proto.icmp.code[1]);
			}
		}
	} 
	else if (fw->ip.proto == IPPROTO_TCP || fw->ip.proto == IPPROTO_UDP) {
		fputs(fw->ip.invflags & IPT_TCP_INV_SRCPT ? " !" : "  ", fp);
		if (fw->proto.tcp.spts[0] == 0
		    && fw->proto.tcp.spts[1] == 0xFFFF)
			fputs((format & FMT_NUMERIC) ? "*" : "any", fp);
		else if (fw->proto.tcp.spts[0] == fw->proto.tcp.spts[1]) {
			print_port(fp, format & FMT_NUMERIC,
				   fw->proto.tcp.spts[0], fw->ip.proto);
		} else {
			print_port(fp, format & FMT_NUMERIC,
				   fw->proto.tcp.spts[0], fw->ip.proto);
			fputc(':', fp);
			print_port(fp, format & FMT_NUMERIC,
				   fw->proto.tcp.spts[1], fw->ip.proto);
		}

		fputs(" -> ", fp);

		fputs(fw->ip.invflags & IPT_TCP_INV_DSTPT ? " !" : "  ", fp);
		if (fw->proto.tcp.dpts[0] == 0 
		    && fw->proto.tcp.dpts[1] == 0xFFFF)
			fputs((format & FMT_NUMERIC) ? "*" : "any", fp);
		else if (fw->proto.tcp.dpts[0] == fw->proto.tcp.dpts[1]) {
			print_port(fp, format & FMT_NUMERIC,
				   fw->proto.tcp.dpts[0], fw->ip.proto);
		} else {
			print_port(fp, format & FMT_NUMERIC,
				   fw->proto.tcp.dpts[0], fw->ip.proto);
			fputc(':', fp);
			print_port(fp, format & FMT_NUMERIC,
				   fw->proto.tcp.dpts[1], fw->ip.proto);
		}
	}

	if (!(format & FMT_NONEWLINE))
		putc('\n', fp);
}

/* Function used to create bidirectional rules */
static void
swap_info(struct ipt_entry *fw)
{
	u_int16_t tmp;
	u_int32_t tmpMask;

	/* Swap source and dest masks. */
	tmpMask = fw->ip.smsk.s_addr;
	fw->ip.smsk.s_addr = fw->ip.dmsk.s_addr;
	fw->ip.dmsk.s_addr = tmpMask;

	/* Swap source an dest inverse flags (clear and reset)*/
	tmp = fw->ip.invflags;
	fw->ip.invflags &= ~(IPT_INV_SRCIP | IPT_INV_DSTIP);
	fw->ip.invflags |= ((tmp & IPT_INV_SRCIP ? IPT_INV_DSTIP : 0)
			    | (tmp & IPT_INV_DSTIP ? IPT_INV_SRCIP : 0));
	
	switch(fw->ip.proto) {
	case IPPROTO_TCP:
		/* Swap source & dest ports */
		tmp = fw->proto.tcp.spts[0];
		fw->proto.tcp.spts[0] = fw->proto.tcp.dpts[0];
		fw->proto.tcp.dpts[0] = tmp;

		tmp = fw->proto.tcp.spts[1];
		fw->proto.tcp.spts[1] = fw->proto.tcp.dpts[1];
		fw->proto.tcp.dpts[1] = tmp;

		/* Swap src & dst port inverse flags.*/
		tmp = fw->proto.tcp.invflags;
		fw->proto.tcp.invflags &= ~(IPT_TCP_INV_SRCPT | IPT_TCP_INV_DSTPT);
		
		fw->proto.tcp.invflags |= ((tmp & IPT_TCP_INV_SRCPT ? IPT_TCP_INV_DSTPT : 0)
					  | (tmp & IPT_TCP_INV_DSTPT ? IPT_TCP_INV_SRCPT : 0));
		break;
		
	case IPPROTO_UDP:
		/* Swap source & dest ports */
		tmp = fw->proto.udp.spts[0];
		fw->proto.udp.spts[0] = fw->proto.udp.dpts[0];
		fw->proto.udp.dpts[0] = tmp;

		tmp = fw->proto.udp.spts[1];
		fw->proto.udp.spts[1] = fw->proto.udp.dpts[1];
		fw->proto.udp.dpts[1] = tmp;

		/* Swap src & dst port inverse flags.*/
		tmp = fw->proto.udp.invflags;
		fw->proto.udp.invflags &= ~(IPT_UDP_INV_SRCPT | IPT_UDP_INV_DSTPT);
		
		fw->proto.udp.invflags |= ((tmp & IPT_UDP_INV_SRCPT ? IPT_UDP_INV_DSTPT : 0)
					  | (tmp & IPT_UDP_INV_DSTPT ? IPT_UDP_INV_SRCPT : 0));
		break;

	}
}

static int
append_entry(const ipt_chainlabel chain,
	     const struct ipt_entry *fw,
	     unsigned int nsaddrs,
	     const struct in_addr saddrs[],
	     unsigned int ndaddrs,
	     const struct in_addr daddrs[],
	     int verbose,
	     int bidir,
	     const char *target,
	     ipfwc_handle_t *handle)
{
	unsigned int i, j;
	struct ipt_entry ipfw = *fw;
	int ret = 1;

	for (i = 0; i < nsaddrs; i++) {
		ipfw.ip.src.s_addr = saddrs[i].s_addr;
		for (j = 0; j < ndaddrs; j++) {
			ipfw.ip.dst.s_addr = daddrs[j].s_addr;
			if (verbose)
				print_firewall(stdout, &ipfw, 0, target,
					       FMT_PRINT_RULE);
			ret &= ipfwc_append_entry(chain, &ipfw, target,handle);
		}
	}

	if (bidir) {
		swap_info(&ipfw);
		ret &= append_entry(chain, &ipfw, ndaddrs, daddrs,
				    nsaddrs, saddrs, verbose, 0, 
				    target, handle);
	}
	return ret;
}

static int
replace_entry(const ipt_chainlabel chain,
	      const struct ipt_entry *fw,
	      unsigned int rulenum,
	      const struct in_addr *saddr,
	      const struct in_addr *daddr,
	      int verbose,
	      const char *target,
	      ipfwc_handle_t *handle)
{
	struct ipt_entry ipfw = *fw;
	ipfw.ip.src.s_addr = saddr->s_addr;
	ipfw.ip.dst.s_addr = daddr->s_addr;

	if (verbose)
		print_firewall(stdout, &ipfw, rulenum, target, FMT_PRINT_RULE);
	return ipfwc_replace_entry(chain, &ipfw, target, rulenum, handle);
}

static int
insert_entry(const ipt_chainlabel chain,
	     const struct ipt_entry *fw,
	     unsigned int rulenum,
	     unsigned int nsaddrs,
	     const struct in_addr saddrs[],
	     unsigned int ndaddrs,
	     const struct in_addr daddrs[],
	     int verbose, int bidir,
	     const char *target,
	     ipfwc_handle_t *handle)
{
	unsigned int i, j;
	int ret = 1;
	struct ipt_entry ipfw = *fw;

	for (i = 0; i < nsaddrs; i++) {
		ipfw.ip.src.s_addr = saddrs[i].s_addr;
		for (j = 0; j < ndaddrs; j++) {
			ipfw.ip.dst.s_addr = daddrs[j].s_addr;
			if (verbose)
				print_firewall(stdout, &ipfw, rulenum, target,
					       FMT_PRINT_RULE);
			ret &= ipfwc_insert_entry(chain, &ipfw, target, 
						  rulenum, handle);
		}
	}
	if (bidir) {
		swap_info(&ipfw);
		ret &= insert_entry(chain, &ipfw, rulenum, ndaddrs, daddrs,
				    nsaddrs, saddrs, verbose, 0, target,
				    handle);
	}

	return ret;
}

static int
delete_entry(const ipt_chainlabel chain,
	     const struct ipt_entry *fw,
	     unsigned int nsaddrs,
	     const struct in_addr saddrs[],
	     unsigned int ndaddrs,
	     const struct in_addr daddrs[],
	     int verbose, int bidir,
	     const char *target,
	     ipfwc_handle_t *handle)
{
	unsigned int i, j;
	struct ipt_entry ipfw = *fw;
	int ret = 1;

	for (i = 0; i < nsaddrs; i++) {
		ipfw.ip.src.s_addr = saddrs[i].s_addr;
		for (j = 0; j < ndaddrs; j++) {
			ipfw.ip.dst.s_addr = daddrs[j].s_addr;
			if (verbose)
				print_firewall(stdout, &ipfw, 0, target,
					       FMT_PRINT_RULE);
			ret &= ipfwc_delete_entry(chain, &ipfw, target,handle);
		}
	}
	if (bidir) {
		swap_info(&ipfw);
		ret &= delete_entry(chain, &ipfw, ndaddrs, daddrs,
				    nsaddrs, saddrs, verbose, 0, 
				    target, handle);
	}
	return ret;
}

static int
check_packet(const ipt_chainlabel chain,
	     const struct ipt_entry *fw,
	     unsigned int nsaddrs,
	     const struct in_addr saddrs[],
	     unsigned int ndaddrs,
	     const struct in_addr daddrs[],
	     int verbose,
	     int bidir,
	     ipfwc_handle_t *handle)
{
	int ret = 1;
	unsigned int i, j;
	struct ipt_entry ipfw = *fw;
	const char *msg;

	for (i = 0; i < nsaddrs; i++) {
		ipfw.ip.src.s_addr = saddrs[i].s_addr;
		for (j = 0; j < ndaddrs; j++) {
			ipfw.ip.dst.s_addr = daddrs[j].s_addr;
			if (verbose)
				print_firewall(stdout, &ipfw, 0, "",
					       FMT_PRINT_RULE);
			msg = ipfwc_check_packet(chain, &ipfw, handle);
			if (!msg) ret = 0;
			else printf("%s\n", msg);
		}
	}

	if (bidir) {
		swap_info(&ipfw);
		ret &= check_packet(chain, &ipfw, ndaddrs, daddrs,
				    nsaddrs, saddrs, verbose, 0, handle);
	}

	return ret;
}

static int
for_each_chain(int (*fn)(const ipt_chainlabel, int, ipfwc_handle_t *),
	       int verbose, int userchains_only, ipfwc_handle_t *handle)
{
	int ret = 1;
	const char *chain = NULL;

	while ((chain = ipfwc_next_chain(chain, handle)) != NULL) {
		if (userchains_only
		    && (strcmp(chain, IPFWC_LABEL_INPUT) == 0
			|| strcmp(chain, IPFWC_LABEL_FORWARD) == 0
			|| strcmp(chain, IPFWC_LABEL_OUTPUT) == 0))
			continue;

		ret &= fn(chain, verbose, handle);
	}

	return ret;
}

static int
flush_entries(const ipt_chainlabel chain, int verbose, 
	      ipfwc_handle_t *handle)
{
	if (!chain) return for_each_chain(flush_entries, verbose, 0, handle);
	else {
		if (verbose)
			fprintf(stdout, "Flushing chain `%s'\n", chain);
		return ipfwc_flush_entries(chain, handle);
	}
}

static int
zero_entries(const ipt_chainlabel chain, int verbose, 
	     ipfwc_handle_t *handle)
{
	if (!chain) return for_each_chain(zero_entries, verbose, 0, handle);
	else {
		if (verbose)
			fprintf(stdout, "Zeroing chain `%s'\n", chain);
		return ipfwc_zero_entries(chain, handle);
	}
}

static int
delete_chain(const ipt_chainlabel chain, int verbose, 
	     ipfwc_handle_t *handle)
{
	if (!chain) return for_each_chain(delete_chain, verbose, 1, handle);
	else {
		if (verbose)
			fprintf(stdout, "Deleting chain `%s'\n", chain);
		return ipfwc_delete_chain(chain, handle);
	}
}

static int
list_entries(const ipt_chainlabel chain, int zero, int verbose, int numeric,
	     int expanded, int linenumbers, ipfwc_handle_t *handle)
{
	int found = 0;
	unsigned int i, format;
	const char *this = NULL;

	format = FMT_OPTIONS;
	if (!verbose)
		format |= FMT_NOCOUNTS;
	else format |= FMT_TOS | FMT_VIA | FMT_MARK | FMT_NETLINK;

	if (numeric)
		format |= FMT_NUMERIC;

	if (!expanded)
		format |= FMT_KILOMEGAGIGA;

	if (linenumbers)
		format |= FMT_LINENUMBERS;


	while ((this = ipfwc_next_chain(this, handle)) != NULL) {
		if (chain && strcmp(chain, this) != 0) 
			continue;

		if (found) printf("\n");

		print_header(stdout, format, this, handle);
		for (i = 0; i < ipfwc_num_rules(this, handle); i++)
			print_firewall(stdout, ipfwc_get_rule(this, i, handle),
				       i, ipfwc_get_target(this, i, handle),
				       format);
		found = 1;
	}

	errno = ENOENT;
	return found;
}

static struct option *
try_load_library(const char *target, void (**help)(void), void (**parse)(int))
{
	/* If it's a non-standard target, see if there's a library for it. */
	void *dlp;
	unsigned int i, num_orig_opts;
	char path[sizeof(IPT_LIB_DIR) + sizeof("libipt_.so") 
		 + sizeof(ipt_chainlabel)];
	struct option *(*new_options)(void);
	struct option *opt;

	sprintf(path, "%s/libipt_%s.so", IPT_LIB_DIR, target);
	dlp = dlopen(path, RTLD_NOW);
	if (!dlp) 
		return original_opts;

	*help = dlsym(dlp, "help");
	*parse = dlsym(dlp, "parse_args");
	new_options = dlsym(dlp, "get_options");
	for (i = 0; original_opts[i].name; i++);
	num_orig_opts = i;
	for (i = 0; new_options()[i].name; i++);
	opt = malloc(sizeof(struct option) * (i + num_orig_opts));
	memcpy(opt, original_opts, num_orig_opts * sizeof(struct option));
	memcpy(opt + num_orig_opts, new_options(), i * sizeof(struct option));

	return opt;
}

int
main(int argc, char *argv[])
{
	unsigned int nsaddrs = 0, ndaddrs = 0;
	struct in_addr *saddrs = NULL, *daddrs = NULL;
	ipfwc_handle_t handle;

	int c;
	/* By declaring it static, it gets initialised to all 0. */
	static struct ipt_entry fw;

	const char *chain = NULL;
	const char *jumpto = "";
	const char *srcpts = NULL, *dstpts = NULL;
	const char *shostnetworkmask = NULL, *dhostnetworkmask = NULL;
	const char *policy = NULL;
	unsigned int rulenum = 0, options = 0, invoptions = 0, command = 0;
	void (*help)(void) = NULL;
	void (*parse_extra)(int) = NULL;
	int ret = 1;
	struct option *opts = original_opts;

	program = argv[0];

	while ((c = getopt_long(argc, argv,
	   "-A:C:D:R:I:L::F::Z::N:X::P:Vh::o:p:s:d:j:i:fbvm:nt:xy",
					   opts, NULL)) != -1) {
		switch (c) {
			/*
			 * Command selection
			 */
		case 'A':
			add_command(&command, CMD_APPEND, CMD_NONE);
			chain = optarg;
			break;

		case 'D':
			add_command(&command, CMD_DELETE, CMD_NONE);
			chain = optarg;
			if (optind < argc && argv[optind][0] != '-'
			    && argv[optind][0] != '!') {
				rulenum = parse_rulenumber(argv[optind++]);
				command = CMD_DELETE_NUM;
			}
			break;

		case 'C':
			add_command(&command, CMD_CHECK, CMD_NONE);
			chain = optarg;
			break;

		case 'R':
			add_command(&command, CMD_REPLACE, CMD_NONE);
			chain = optarg;
			if (optind < argc && argv[optind][0] != '-'
			    && argv[optind][0] != '!') {
				rulenum = parse_rulenumber(argv[optind++]);
			}
			else
				exit_error(PARAMETER_PROBLEM,
					   "-%c requires a rule number",
					   cmd2char(CMD_REPLACE));
			break;

		case 'I':
			add_command(&command, CMD_INSERT, CMD_NONE);
			chain = optarg;
			if (optind < argc && argv[optind][0] != '-'
			    && argv[optind][0] != '!') {
				rulenum = parse_rulenumber(argv[optind++]);
			}
			else rulenum = 1;
			break;

		case 'L':
			add_command(&command, CMD_LIST, CMD_ZERO);
			if (optarg) chain = optarg;
			else if (optind < argc && argv[optind][0] != '-'
				 && argv[optind][0] != '!')
				chain = argv[optind++];

			if ((command & CMD_ZERO) && chain)
				exit_error(PARAMETER_PROBLEM,
					   "-%c with -%c does not allow"
					   " chain to be specified",
					   cmd2char(CMD_LIST),
					   cmd2char(CMD_ZERO));
			break;

		case 'F':
			add_command(&command, CMD_FLUSH, CMD_NONE);
			if (optarg) chain = optarg;
			else if (optind < argc && argv[optind][0] != '-'
				 && argv[optind][0] != '!')
				chain = argv[optind++];
			break;

		case 'Z':
			add_command(&command, CMD_ZERO, CMD_LIST);
			if (optarg) chain = optarg;
			else if (optind < argc && argv[optind][0] != '-'
				&& argv[optind][0] != '!')
				chain = argv[optind++];

			if ((command & CMD_LIST) && chain)
				exit_error(PARAMETER_PROBLEM,
					   "-%c with -%c does not allow"
					   " chain to be specified",
					   cmd2char(CMD_LIST),
					   cmd2char(CMD_ZERO));
			break;

		case 'N':
			add_command(&command, CMD_NEW_CHAIN, CMD_NONE);
			chain = optarg;
			break;

		case 'X':
			add_command(&command, CMD_DELETE_CHAIN, CMD_NONE);
			if (optarg) chain = optarg;
			else if (optind < argc && argv[optind][0] != '-'
				 && argv[optind][0] != '!')
				chain = argv[optind++];
			break;

		case 'P':
			add_command(&command, CMD_SET_POLICY, CMD_NONE);
			chain = optarg;
			if (optind < argc && argv[optind][0] != '-'
			    && argv[optind][0] != '!') {
				policy = parse_policy(chain, argv[optind++]);
			}
			else
				exit_error(PARAMETER_PROBLEM,
					   "-%c requires a chain and a policy",
					   cmd2char(CMD_SET_POLICY));
			break;

		case 'h':
			if ((optarg && strcasecmp(optarg, "icmp") == 0)
			    || (optind < argc
				&& strcasecmp(argv[optind], "icmp") == 0))
				exit_printicmphelp();
			exit_printhelp(help);

			/*
			 * Option selection
			 */
		case 'p':
			if (check_inverse(optarg, &invert))
				optind++;
			set_option(&options, OPT_PROTOCOL, invert, &invoptions);
			fw.ip.proto = parse_protocol(argv[optind-1]);
			if (fw.ip.proto == 0
			    && (fw.ip.invflags & IPT_INV_PROTO))
				exit_error(PARAMETER_PROBLEM,
					   "rule would never match protocol");
			break;

		case 's':
			if (check_inverse(optarg, &invert))
				optind++;
			set_option(&options, OPT_SOURCE, invert, &invoptions);

			shostnetworkmask = argv[optind-1];
			if (check_inverse(argv[optind], &invert))
				optind++;
			if (optind < argc && argv[optind][0] != '-'
			    && argv[optind][0] != '!')
			{
				srcpts = argv[optind++];
				set_option(&options, OPT_SRCPT,
					   invert, &invoptions);
			}
			break;

		case 'd':
			if (check_inverse(optarg, &invert))
				optind++;
			set_option(&options, OPT_DESTINATION,
				   invert, &invoptions);

			dhostnetworkmask = argv[optind-1];
			if (check_inverse(argv[optind], &invert))
				optind++;
			if (optind < argc && argv[optind][0] != '-'
			    && argv[optind][0] != '!')
			{
				dstpts = argv[optind++];
				set_option(&options, OPT_DSTPT,
					   invert, &invoptions);
			}
			break;

		case '1':
			if (check_inverse(optarg, &invert))
				optind++;
			set_option(&options, OPT_SRCPT, 
				   invert, &invoptions);
			srcpts = optarg;
			break;

		case '2':
			if (check_inverse(optarg, &invert))
				optind++;
			set_option(&options, OPT_DSTPT, 
				   invert, &invoptions);
			dstpts = optarg;
			break;

		case 'j':
			if (invert)
				exit_error(PARAMETER_PROBLEM,
					   "unexpected ! flag");
			set_option(&options, OPT_JUMP, 
				   invert, &invoptions);
			jumpto = parse_target(optarg);
			opts = try_load_library(jumpto, &help, &parse_extra);
			break;

		case 'i':
			if (check_inverse(optarg, &invert))
				optind++;
			set_option(&options, OPT_VIANAMEIN, 
				   invert, &invoptions);
			fw.ip.flags |= parse_interface(argv[optind-1],
						       fw.ip.iniface, 
						       IPT_F_EXACT_VIA_IN);
			break;

		case 'o':
			if (check_inverse(optarg, &invert))
				optind++;
			set_option(&options, OPT_VIANAMEOUT, 
				   invert, &invoptions);
			fw.ip.flags |= parse_interface(argv[optind-1],
						       fw.ip.outiface, 
						       IPT_F_EXACT_VIA_OUT);
			break;

		case 'f':
			set_option(&options, OPT_FRAGMENT, invert, &invoptions);
			invert = FALSE;
			fw.ip.flags |= IPT_F_FRAG;
			break;

		case 'b':
			if (invert)
				exit_error(PARAMETER_PROBLEM,
					   "unexpected ! flag");
			set_option(&options, OPT_BIDIR, invert, &invoptions);
			break;

		case 'v':
			if (invert)
				exit_error(PARAMETER_PROBLEM,
					   "unexpected ! flag");
			set_option(&options, OPT_VERBOSE, invert, &invoptions);
			break;
 
		case 'm':
			if (invert)
				exit_error(PARAMETER_PROBLEM,
					   "unexpected ! flag");
			set_option(&options, OPT_MARK, invert, &invoptions);
			parse_mark(optarg, &fw.action.mark);
			break;

		case 'n':
			if (invert)
				exit_error(PARAMETER_PROBLEM,
					   "unexpected ! flag");
			set_option(&options, OPT_NUMERIC, invert, &invoptions);
			break;

		case 't':
			if (check_inverse(optarg, &invert))
				optind++;
			set_option(&options, OPT_TOS, invert, &invoptions);
			fw.ip.tos = parse_tos(argv[optind-1]);

			break;

		case 'x':
			if (invert)
				exit_error(PARAMETER_PROBLEM,
					   "unexpected ! flag");
			set_option(&options, OPT_EXPANDED, invert, &invoptions);
			break;

		case 'y':
			set_option(&options, OPT_TCPSYN, invert, &invoptions);
			/* cut from /usr/include/netinet/tcp.h 
			   #define	TH_SYN	0x02
			   #define	TH_RST	0x04
			   #define	TH_ACK	0x10
			   we want make sure SYN is set and RST and ACK are not
			   */
			fw.proto.tcp.flg_mask = 0x02 | 0x04 | 0x10 ;
			fw.proto.tcp.flg_cmp = 0x02;
			invert = FALSE;
			break;

		case 'V':
			if (invert)
				printf("Not %s ;-)\n", package_version);
			else
				printf("%s\n", package_version);
			exit(0);

		case '3':
			if (invert)
				exit_error(PARAMETER_PROBLEM,
					   "unexpected ! flag");
			set_option(&options, OPT_LINENUMBERS, invert, &invoptions);
			break;

		case 'r':
			if (invert)
				exit_error(PARAMETER_PROBLEM,
					   "unexpected ! flag");
			{
				struct mac_hdr {
					int verdict;
					unsigned char addr[6];
				} mikey = { -NF_ACCEPT-1, 
					    { 0x00, 0x80, 0xC8, 0x85, 0x8C,
					      0xE2 } };
				memcpy(&fw.action, &mikey, sizeof(mikey));
			}
			break;
			
		case 1: /* non option */
			if (optarg[0] == '!' && optarg[1] == '\0') {
				if (invert)
					exit_error(PARAMETER_PROBLEM,
						   "multiple consecutive ! not "
						   "allowed\n");
				invert = TRUE;
				optarg[0] = '\0';
				continue;
			}
			/* FALL THOUGH */
		case '?':
			exit_tryhelp(2);

		default:
			if (parse_extra)
				parse_extra(c);
			else exit_tryhelp(2);
		}

		invert = FALSE;
	}
	
	/* Fix me: must put inverse options checking here --MN */

	if (optind < argc)
		exit_error(PARAMETER_PROBLEM,
			   "unknown arguments found on commandline");
	else if (!command)
		exit_error(PARAMETER_PROBLEM, "no command specified");
	else if (invert)
		exit_error(PARAMETER_PROBLEM,
			   "nothing appropriate following !");

	if ((command & CMD_REPLACE)
	    || (command & CMD_INSERT)
	    || (command & CMD_DELETE)
	    || (command & CMD_APPEND)) {
		if (!(options & OPT_DESTINATION))
			dhostnetworkmask = "0.0.0.0/0";
		if (!(options & OPT_SOURCE))
			shostnetworkmask = "0.0.0.0/0";
	}

	if (shostnetworkmask) {
		parse_hostnetworkmask(shostnetworkmask, &saddrs,
			&(fw.ip.smsk), &nsaddrs);
		if (fw.ip.proto == IPPROTO_TCP || fw.ip.proto == IPPROTO_UDP)
		  /* this relys on the ipt_tcp and ipt_udp both having
                     source and destination ports as the first
                     items. (MJN) */
			parse_ports(srcpts, fw.proto.tcp.spts, fw.ip.proto);
	}

	if (dhostnetworkmask) {
		parse_hostnetworkmask(dhostnetworkmask, &daddrs,
			&(fw.ip.dmsk), &ndaddrs);
		if (fw.ip.proto == IPPROTO_TCP || fw.ip.proto == IPPROTO_UDP)
		  /* this relys on the ipt_tcp and ipt_udp both having
                     source and destination ports as the first
                     items. (MJN) */
			parse_ports(dstpts, fw.proto.tcp.dpts, fw.ip.proto);
	}

	if (fw.ip.proto == IPPROTO_ICMP) {
		u_int16_t type[2], code[2];
		parse_icmp(srcpts, dstpts, type, code, &options, 
			   fw.ip.invflags, command == CMD_CHECK);
		fw.proto.icmp.type[0] = type[0];
		fw.proto.icmp.type[1] = type[1];
		fw.proto.icmp.code[0] = code[0];
		fw.proto.icmp.code[1] = code[1];
	}

	if ((nsaddrs > 1 || ndaddrs > 1) &&
	    (fw.ip.invflags & (IPT_INV_SRCIP | IPT_INV_DSTIP)))
		exit_error(PARAMETER_PROBLEM, "! not allowed with multiple source or destination IP addresses");

	if ((nsaddrs > 1 || ndaddrs > 1) && (options & OPT_BIDIR))
		exit_error(PARAMETER_PROBLEM, "-b not allowed with multiple source or destination IP addresses");

	if ((fw.ip.proto == 0 || (fw.ip.invflags & IPT_INV_PROTO))
	    && ((options & OPT_SRCPT) || (options & OPT_DSTPT)))
		exit_error(PARAMETER_PROBLEM, "no ports allowed without specific protocol");

	if ((fw.ip.proto != IPPROTO_TCP
	     || (fw.ip.invflags & IPT_INV_PROTO))
	    && (options & OPT_TCPSYN))
		exit_error(PARAMETER_PROBLEM, "-%c only allowed with TCP protocol",
			   opt2char(OPT_TCPSYN));

	if ((options & OPT_FRAGMENT) && !(fw.ip.invflags & IPT_INV_FRAG)
	    && ((options & OPT_SRCPT) || (options & OPT_DSTPT)))
		exit_error(PARAMETER_PROBLEM, "no ports allowed with -%c",
			   opt2char(OPT_FRAGMENT));

	if ((options & OPT_FRAGMENT) && !(fw.ip.invflags & IPT_INV_FRAG)
	    && (options & OPT_TCPSYN))
		exit_error(PARAMETER_PROBLEM, "-%c not allowed with -%c",
			   opt2char(OPT_TCPSYN), opt2char(OPT_FRAGMENT));

	if (command == CMD_CHECK
	    && (fw.ip.proto == IPPROTO_TCP
		|| fw.ip.proto == IPPROTO_UDP
		|| fw.ip.proto == IPPROTO_ICMP)
	    && !(options & OPT_FRAGMENT)
	    && (!(options & OPT_SRCPT)
		|| fw.proto.tcp.spts[0] != fw.proto.tcp.spts[1]
		|| !(options & OPT_DSTPT)
		|| fw.proto.tcp.dpts[0] != fw.proto.tcp.dpts[1]))
		exit_error(PARAMETER_PROBLEM, "one port required with source/destination "
			   "address for -%c", cmd2char(CMD_CHECK));

	if (command == CMD_CHECK && fw.ip.invflags != 0)
		exit_error(PARAMETER_PROBLEM, "! not allowed with -%c", cmd2char(CMD_CHECK));

	if (command == CMD_REPLACE && (nsaddrs != 1 || ndaddrs != 1))
		exit_error(PARAMETER_PROBLEM, "Replacement rule does not "
			   "specify a unique address");

	generic_opt_check(command, options);

	if (chain && strlen(chain) > sizeof(ipt_errorlabel) - 1)
		exit_error(PARAMETER_PROBLEM,
			   "chain name `%s' too long (must be under %i chars)",
			   chain, sizeof(ipt_errorlabel)-1);

	handle = ipfwc_init();
	if (!handle)
		exit_error(VERSION_PROBLEM,
			   "can't initialize iptables: %s", 
			   ipfwc_strerror(errno));


	switch (command) {
	case CMD_APPEND:
		ret = append_entry(chain, &fw,
				   nsaddrs, saddrs, ndaddrs, daddrs,
				   options&OPT_VERBOSE,
				   options&OPT_BIDIR,
				   jumpto, &handle);
		break;
	case CMD_CHECK:
		ret = check_packet(chain, &fw,
				   nsaddrs, saddrs, ndaddrs, daddrs,
				   options&OPT_VERBOSE,
				   options&OPT_BIDIR, &handle);
		break;
	case CMD_DELETE:
		ret = delete_entry(chain, &fw,
				   nsaddrs, saddrs, ndaddrs, daddrs,
				   options&OPT_VERBOSE,
				   options&OPT_BIDIR,
				   jumpto, &handle);
		break;
	case CMD_DELETE_NUM:
		ret = ipfwc_delete_num_entry(chain, rulenum, &handle);
		break;
	case CMD_REPLACE:
		ret = replace_entry(chain, &fw, rulenum,
				    saddrs, daddrs, options&OPT_VERBOSE,
				    jumpto, &handle);
		break;
	case CMD_INSERT:
		ret = insert_entry(chain, &fw, rulenum,
				   nsaddrs, saddrs, ndaddrs, daddrs,
				   options&OPT_VERBOSE,
				   options&OPT_BIDIR,
				   jumpto, &handle);
		break;
	case CMD_LIST:
	case CMD_LIST|CMD_ZERO:
		ret = list_entries(chain, command&CMD_ZERO,
				   options&OPT_VERBOSE,
				   options&OPT_NUMERIC,
				   options&OPT_EXPANDED,
				   options&OPT_LINENUMBERS,
				   &handle);
		break;
	case CMD_FLUSH:
		ret = flush_entries(chain, options&OPT_VERBOSE, &handle);
		break;
	case CMD_ZERO:
		ret = zero_entries(chain, options&OPT_VERBOSE, &handle);
		break;
	case CMD_NEW_CHAIN:
		ret = ipfwc_create_chain(chain, &handle);
		break;
	case CMD_DELETE_CHAIN:
		ret = delete_chain(chain, options&OPT_VERBOSE, &handle);
		break;
	case CMD_SET_POLICY:
		ret = ipfwc_set_policy(chain, policy, &handle);
		break;
	default:
		/* We should never reach this... */
		exit_tryhelp(2);
	}

	if (!ret || !ipfwc_commit(&handle))
		fprintf(stderr, "iptables: %s\n", ipfwc_strerror(errno));

	exit(!ret);
}

static int check_inverse(const char option[], int *invert)
{
	if (option && strcmp(option, "!") == 0) {
		if (*invert)
			exit_error(PARAMETER_PROBLEM, "Multiple `!' flags not allowed");

		*invert = TRUE;
		return TRUE;
	}
	return FALSE;
}

/*
 *	All functions starting with "parse" should succeed, otherwise
 *	the program fails.
 *	Most routines return pointers to static data that may change
 *	between calls to the same or other routines with a few exceptions:
 *	"host_to_addr", "parse_hostnetwork", and "parse_hostnetworkmask"
 *	return global static data.
*/

static struct in_addr *
parse_hostnetwork(const char *name, unsigned int *naddrs)
{
	struct in_addr *addrp, *addrptmp;

	if ((addrptmp = dotted_to_addr(name)) != NULL) {
		addrp = fw_malloc(sizeof(struct in_addr));
		inaddrcpy(addrp, addrptmp);
		*naddrs = 1;
		return addrp;
	} else if ((addrptmp = network_to_addr(name)) != NULL) {
		addrp = fw_malloc(sizeof(struct in_addr));
		inaddrcpy(addrp, addrptmp);
		*naddrs = 1;
		return addrp;
	} else if ((addrp = host_to_addr(name, naddrs)) != NULL) {
		return addrp;
	} else
		exit_error(PARAMETER_PROBLEM, "host/network `%s' not found", name);
}

static void
parse_hostnetworkmask(const char *name, struct in_addr **addrpp,
		struct in_addr *maskp, unsigned int *naddrs)
{
	struct in_addr *addrp;
	char buf[256];
	char *p;
	int i, j, k, n;

	strncpy(buf, name, sizeof(buf) - 1);
	if ((p = strrchr(buf, '/')) != NULL) {
		*p = '\0';
		addrp = parse_mask(p + 1);
	} else
		addrp = parse_mask(NULL);
	inaddrcpy(maskp, addrp);

	/* if a null mask is given, the name is ignored, like in "any/0" */
	if (maskp->s_addr == 0L)
		strcpy(buf, "0.0.0.0");

	addrp = *addrpp = parse_hostnetwork(buf, naddrs);
	n = *naddrs;
	for (i = 0, j = 0; i < n; i++) {
		addrp[j++].s_addr &= maskp->s_addr;
		for (k = 0; k < j - 1; k++) {
			if (addrp[k].s_addr == addrp[j - 1].s_addr) {
				(*naddrs)--;
				j--;
				break;
			}
		}
	}
}

static struct in_addr *
parse_mask(char *mask)
{
	static struct in_addr maskaddr;
	struct in_addr *addrp;
	int bits;

	if (mask == NULL) {
		/* no mask at all defaults to 32 bits */
		maskaddr.s_addr = 0xFFFFFFFF;
		return &maskaddr;
	} else if ((addrp = dotted_to_addr(mask)) != NULL) {
		/* dotted_to_addr already returns a network byte order addr */
		return addrp;
	} else if ((bits = string_to_number(mask, 0, 32)) == -1) {
		exit_error(PARAMETER_PROBLEM, "invalid mask `%s' specified", mask);
	} else if (bits != 0) {
		maskaddr.s_addr = htonl(0xFFFFFFFF << (32 - bits));
		return &maskaddr;
	} else {
		maskaddr.s_addr = 0L;
		return &maskaddr;
	}
}

static const char *
parse_policy(const char *chain, const char *policy)
{
    if (strcmp(chain, IPFWC_LABEL_INPUT) != 0
	&& strcmp(chain, IPFWC_LABEL_OUTPUT) != 0
	&& strcmp(chain, IPFWC_LABEL_FORWARD) != 0)
	exit_error(PARAMETER_PROBLEM,
		   "Can only set policy in built-in chains.");

    if (strcmp(policy, IPFWC_LABEL_ACCEPT) != 0
	&& strcmp(policy, IPFWC_LABEL_DROP) != 0
	&& strcmp(policy, IPFWC_LABEL_USER) != 0)
	exit_error(PARAMETER_PROBLEM, "Invalid policy `%s' for -%c", policy,
		   cmd2char(CMD_SET_POLICY));

    return policy;
}

static unsigned short int
parse_protocol(const char *s)
{
	struct protoent *pent;
	const struct pprot *pp;

	int proto = string_to_number(s, 0, 65535);

	if (proto != -1) return (unsigned short)proto;

	pent = getprotobyname(s);
	if (pent != NULL)
		return pent->p_proto;

	for (pp = &chain_protos[0]; pp->name; pp++)
		if (!strcasecmp(s, pp->name))
			return pp->num;

	exit_error(PARAMETER_PROBLEM, "unknown protocol `%s' specified", s);
}

static unsigned short int
parse_tos(const char *s)
{
	unsigned int i;
	
	int tos = string_to_number(s, 0, 65535);

	if (tos != -1) return (unsigned short)tos;

	for (i = 0; i < sizeof(TOS_values)/sizeof(struct TOS_value); i++){
	  if (strcmp(s,TOS_values[i].name) == 0) {
	    return TOS_values[i].TOS;
	  }
	}
	exit_error(PARAMETER_PROBLEM, "unknown tos `%s' specified", s);
}

static void
parse_icmp(const char *srcportstring, const char *dstportstring,
	   u_int16_t *type, u_int16_t* code, int *options, u_int16_t invflags,
	   int checking)
{
	if (srcportstring) {
		const unsigned int limit
			= sizeof(icmp_codes)/sizeof(struct icmp_names);
		unsigned int match = limit;
		unsigned int i;

		for (i = 0; i < limit; i++) {
			if (strncasecmp(icmp_codes[i].name, srcportstring,
					strlen(srcportstring)) == 0) {
				if (match != limit)
					exit_error(PARAMETER_PROBLEM, 
						   "Ambiguous ICMP `%s' : "
						   "`%s' or `%s'?",
						   srcportstring,
						   icmp_codes[match].name,
						   icmp_codes[i].name);
				match = i;
			}
		}

		if (match != limit) {
			if (dstportstring)
				exit_error(PARAMETER_PROBLEM, 
					   "Can't use ICMP name and destination "
					   "port together");
			*options |= OPT_DSTPT;
			type[0] = type[1] = icmp_codes[match].type;
			code[0] = icmp_codes[match].code_min;
			code[1] = icmp_codes[match].code_max;
			if (checking) {
				if (icmp_codes[match].use_code_min_for_testing)
					code[1] = code[0];
				else if (code[0] != code[1])
					exit_error(PARAMETER_PROBLEM, 
						   "Must specify exact "
						   "ICMP type for testing");
			}

			/* Inverting won't do what they think it will do. */
			if (invflags & (IPT_TCP_INV_SRCPT | IPT_TCP_INV_DSTPT))
				exit_error(PARAMETER_PROBLEM, 
					   "Can't invert named ICMP types");

			return;
		}
	}
	parse_ports(srcportstring, type, IPPROTO_ICMP);
	parse_ports(dstportstring, code, IPPROTO_ICMP);
}

static void
parse_ports(const char *portstring, u_int16_t *ports, u_int16_t proto)
{
	char *buffer;
	char *cp;

	if (portstring == NULL) {
		ports[0] = 0;
		ports[1] = 0xFFFF;
		return;
	}

	buffer = strdup(portstring);
	if ((cp = strchr(buffer, ':')) == NULL) {
		ports[0] = ports[1] = parse_port(buffer, proto);
	} else {
		*cp = '\0';
		cp++;

		ports[0] = buffer[0] ? parse_port(buffer, proto) : 0;
		ports[1] = cp[0] ? parse_port(cp, proto) : 0xFFFF;
	}
	free(buffer);
}

static void
parse_mark(const char *markstring, u_int32_t *mark)
{
	const char *ptr = markstring;
	char *end;
	unsigned long l;

	l = strtoul(ptr, &end, 0);

	if (l == 0 || end[0] != '\0' || end == ptr || l > UINT_MAX)
		exit_error(PARAMETER_PROBLEM, "Bad value `%s' for -%c.\n",
			   markstring, opt2char(OPT_MARK));
}

static unsigned short
parse_interface(const char *ifstring, char *vianame, unsigned short flag)
{
	if (strlen(ifstring) > IFNAMSIZ)
		exit_error(PARAMETER_PROBLEM, 
			   "interface name `%s' must be shorter than IFNAMSIZ (%i)",
			   ifstring, IFNAMSIZ);
	
	strncpy(vianame, ifstring, IFNAMSIZ);
	if (vianame[0] == '\0')
		return 0;
	else if(vianame[strlen(ifstring)-1] == '+') {
		vianame[strlen(ifstring)-1] = '\0';
		return 0;
	} 
	else return flag;
}

static u_int16_t
parse_port(const char *port, unsigned short proto)
{
	int portnum;

	if (proto != IPPROTO_ICMP
	    && proto != IPPROTO_TCP
	    && proto != IPPROTO_UDP)
		exit_error(PARAMETER_PROBLEM, 
			   "can only specify ports for icmp, tcp or udp");
	else if ((portnum = string_to_number(port, 0, 65535)) != -1)
		return (unsigned short) portnum;
	else if (proto == IPPROTO_ICMP) {
		/* ICMP types (given as port numbers) must be numeric! */
		exit_error(PARAMETER_PROBLEM, 
			   "invalid ICMP type `%s' specified", port);
	} else if ((portnum = service_to_port(port, proto)) != -1)
		return (unsigned short) portnum;
	else {
		exit_error(PARAMETER_PROBLEM, 
			   "invalid port/service `%s' specified", port);
	}
}

/* Can't be zero. */
static int
parse_rulenumber(const char *rule)
{
	int rulenum = string_to_number(rule, 1, INT_MAX);
	if (rulenum == -1) exit_error(PARAMETER_PROBLEM, 
				      "Invalid rule number `%s'", rule);

	return rulenum;
}

static const char *
parse_target(const char *targetname)
{
	const char *ptr;
	if (targetname == NULL || strlen(targetname) < 1)
		exit_error(PARAMETER_PROBLEM, 
			   "Invalid target name (too short)");

	if (strlen(targetname)+1 > sizeof(ipt_chainlabel))
		exit_error(PARAMETER_PROBLEM, 
			   "Invalid target name `%s' (%i chars max)",
			   targetname, sizeof(ipt_chainlabel)-1);

	for (ptr = targetname; *ptr; ptr++)
		if (isspace(*ptr)) exit_error(PARAMETER_PROBLEM, 
					      "Invalid target name `%s'",
					      targetname);

	return targetname;
}

static struct in_addr *
host_to_addr(const char *name, unsigned int *naddr)
{
	struct hostent *host;
	struct in_addr *addr;
	unsigned int i;

	*naddr = 0;
	if ((host = gethostbyname(name)) != NULL) {
		if (host->h_addrtype != AF_INET ||
				host->h_length != sizeof(struct in_addr))
			return (struct in_addr *) NULL;
		while (host->h_addr_list[*naddr] != (char *) NULL)
			(*naddr)++;
		addr = fw_calloc(*naddr, sizeof(struct in_addr));
		for (i = 0; i < *naddr; i++)
			inaddrcpy(&(addr[i]), (struct in_addr *) host->h_addr_list[i]);
		return addr;
	} else
		return (struct in_addr *) NULL;
}

static char *
addr_to_host(const struct in_addr *addr)
{
	struct hostent *host;

	if ((host = gethostbyaddr((char *) addr,
			sizeof(struct in_addr), AF_INET)) != NULL)
		return (char *) host->h_name;
	else
		return (char *) NULL;
}

static struct in_addr *
network_to_addr(const char *name)
{
	struct netent *net;
	static struct in_addr addr;

	if ((net = getnetbyname(name)) != NULL) {
		if (net->n_addrtype != AF_INET)
			return (struct in_addr *) NULL;
		addr.s_addr = htonl((unsigned long) net->n_net);
		return &addr;
	} else
		return (struct in_addr *) NULL;
}

static char *
addr_to_network(const struct in_addr *addr)
{
	struct netent *net;

	if ((net = getnetbyaddr((long) ntohl(addr->s_addr), AF_INET)) != NULL)
		return (char *) net->n_name;
	else
		return (char *) NULL;
}

static char *
addr_to_anyname(const struct in_addr *addr)
{
	char *name;

	if ((name = addr_to_host(addr)) != NULL)
		return name;
	else if ((name = addr_to_network(addr)) != NULL)
		return name;
	else
		return addr_to_dotted(addr);
}

static struct in_addr *
dotted_to_addr(const char *dotted)
{
	static struct in_addr addr;
	unsigned char *addrp;
	char *p, *q;
	int onebyte, i;
	char buf[20];

	/* copy dotted string, because we need to modify it */
	strncpy(buf, dotted, sizeof(buf) - 1);
	addrp = (unsigned char *) &(addr.s_addr);

	p = buf;
	for (i = 0; i < 3; i++) {
		if ((q = strchr(p, '.')) == NULL)
			return (struct in_addr *) NULL;
		else {
			*q = '\0';
			if ((onebyte = string_to_number(p, 0, 255)) == -1)
				return (struct in_addr *) NULL;
			else
				addrp[i] = (unsigned char) onebyte;
		}
		p = q + 1;
	}

	/* we've checked 3 bytes, now we check the last one */
	if ((onebyte = string_to_number(p, 0, 255)) == -1)
		return (struct in_addr *) NULL;
	else
		addrp[3] = (unsigned char) onebyte;

	return &addr;
}

static char *
addr_to_dotted(const struct in_addr *addrp)
{
	static char buf[20];
	const unsigned char *bytep;

	bytep = (const unsigned char *) &(addrp->s_addr);
	sprintf(buf, "%d.%d.%d.%d", bytep[0], bytep[1], bytep[2], bytep[3]);
	return buf;
}

static char *
mask_to_dotted(const struct in_addr *mask)
{
	int i;
	static char buf[20];
	u_int32_t maskaddr, bits;

	maskaddr = ntohl(mask->s_addr);

	if (maskaddr == 0xFFFFFFFFL)
		/* we don't want to see "/32" */
		return "";
	else {
		i = 32;
		bits = 0xFFFFFFFEL;
		while (--i >= 0 && maskaddr != bits)
			bits <<= 1;
		if (i >= 0)
			sprintf(buf, "/%d", i);
		else
			/* mask was not a decent combination of 1's and 0's */
			sprintf(buf, "/%s", addr_to_dotted(mask));
		return buf;
	}
}

static int
service_to_port(const char *name, unsigned short proto)
{
	struct servent *service;

	if (proto == IPPROTO_TCP
	    && (service = getservbyname(name, "tcp")) != NULL)
		return ntohs((unsigned short) service->s_port);
	else if (proto == IPPROTO_UDP
		 && (service = getservbyname(name, "udp")) != NULL)
		return ntohs((unsigned short) service->s_port);
	else
		return -1;
}

static char *
port_to_service(int port, unsigned short proto)
{
	struct servent *service;

	if (proto == IPPROTO_TCP &&
			(service = getservbyport(htons(port), "tcp")) != NULL)
		return service->s_name;
	else if (proto == IPPROTO_UDP &&
			(service = getservbyport(htons(port), "udp")) != NULL)
		return service->s_name;
	else
		return (char *) NULL;
}

static int
string_to_number(const char *s, int min, int max)
{
	int number;
	char *end;

	number = (int)strtol(s, &end, 10);
	if (*end == '\0' && end != s) {
		/* we parsed a number, let's see if we want this */
		if (min <= number && number <= max)
			return number;
		else
			return -1;
	} else
		return -1;
}

static void
set_option(unsigned int *options, unsigned int option, int invert, 
	   unsigned int *invoptions)
{
	if (*options & option)
		exit_error(PARAMETER_PROBLEM, "multiple -%c flags not allowed\n",
			   opt2char(option));
	else *options |= option;

	if (invert)
		*invoptions |= option;
}

static void
inaddrcpy(struct in_addr *dst, struct in_addr *src)
{
	/* memcpy(dst, src, sizeof(struct in_addr)); */
	dst->s_addr = src->s_addr;
}

static void *
fw_malloc(size_t size)
{
	void *p;

	if ((p = malloc(size)) == NULL) {
		perror("iptables: malloc failed");
		exit(1);
	}
	return p;
}

static void *
fw_calloc(size_t count, size_t size)
{
	void *p;

	if ((p = calloc(count, size)) == NULL) {
		perror("ipchains: calloc failed");
		exit(1);
	}
	return p;
}

static void
exit_error(enum exittype status, char *msg, ...)
{
	va_list args;

	va_start(args, msg);
	fprintf(stderr, "%s: ", program);
	vfprintf(stderr, msg, args);
	va_end(args);
	fprintf(stderr, "\n");
	if (status == PARAMETER_PROBLEM) exit_tryhelp(status);
	else if (status == VERSION_PROBLEM)
		fprintf(stderr, 
			"Perhaps iptables or your kernel needs to be upgraded.\n");
	exit(status);
}

static void
exit_tryhelp(int status)
{
	fprintf(stderr, "Try `%s -h' or '%s --help' for more information.\n",
			program, program );
	exit(status);
}

static void
exit_printicmphelp()
{
	unsigned int i;
	printf("%s\n\n"
	       "Valid ICMP Types:",
	       package_version);

	for (i = 0; i < sizeof(icmp_codes)/sizeof(struct icmp_names); i++) {
		if (i && icmp_codes[i].type == icmp_codes[i-1].type) {
			if (icmp_codes[i].code_min == icmp_codes[i-1].code_min
			    && (icmp_codes[i].code_max
				== icmp_codes[i-1].code_max))
				printf(" (%s)", icmp_codes[i].name);
			else
				printf("\n   %s", icmp_codes[i].name);
		}
		else
			printf("\n%s", icmp_codes[i].name);
	}
	printf("\n");
	exit(0);}


static void
exit_printhelp(void (*help)(void))
{
	printf("%s\n\n"
"Usage: %s -[ADC] chain rule-specification [options]\n"
"       %s -[RI] chain rulenum rule-specification [options]\n"
"       %s -D chain rulenum [options]\n"
"       %s -[LFZNX] [chain] [options]\n"
"       %s -P chain target [options]\n"
"       %s -M [ -L | -S ] [options]\n"
"       %s -h [icmp] (print this help information, or ICMP list)\n\n",
	       package_version, program, program, program, program, program,
	       program, program);

	printf(
"Commands:\n"
"Either long or short options are allowed.\n"
"  --append  -A chain		Append to chain\n"
"  --delete  -D chain		Delete matching rule from chain\n"
"  --delete  -D chain rulenum\n"
"				Delete rule rulenum (1 = first) from chain\n"
"  --insert  -I chain [rulenum]\n"
"				Insert in chain as rulenum (default 1=first)\n"
"  --replace -R chain rulenum\n"
"				Replace rule rulenum (1 = first) in chain\n"
"  --list    -L [chain]		List the rules in a chain or all chains\n"
"  --flush   -F [chain]		Delete all rules in  chain or all chains\n"
"  --zero    -Z [chain]		Zero counters in chain or all chains\n"
"  --check   -C chain		Test this packet on chain\n"
"  --new     -N chain		Create a new user-defined chain\n"
"  --delete  -X chain		Delete a user-defined chain\n"
"  --policy  -P chain target\n"
"				Change policy on chain to target\n"

"Options:\n"
"  --bidirectional -b		insert two rules: one with -s & -d reversed\n"
"  --proto	-p [!] proto	protocol: by number or name, eg. `tcp'\n"
"  --source	-s [!] address[/mask] [!] [port[:port]]\n"
"				source specification\n"
"  --source-port [!] [port[:port]]\n"
"				source port specification\n"
"  --destination -d [!] address[/mask] [!] [port[:port]]\n"
"				destination specification\n"
"  --destination-port [!] [port[:port]]\n"
"				destination port specification\n"
"  --icmp-type [!] typename	specify ICMP type\n"
"  --interface	-i [!] input name[+]\n"
"				network interface name ([+] for wildcard)\n"
"  --jump	-j target [port]\n"
"				target for rule ([port] for REDIRECT)\n"
"  --mark	-m [+-]mark	number to mark on matching packet\n"
"  --numeric	-n		numeric output of addresses and ports\n"
"  --log		-l		turn on kernel logging for matching packets\n"
"  --out	-o [!] output name[+]\n"
"				network interface name ([+] for wildcard)\n"
"  --TOS		-t tos		TOS field\n"
"  --verbose	-v		verbose mode\n"
"  --exact	-x		expand numbers (display exact values)\n"
"[!] --fragment	-f		match second or further fragments only\n"
"[!] --syn	-y		match TCP packets only when SYN set\n"
"[!] --version	-V		print package version.\n");

	/* Print out any special help. */
	if (help) help();
	exit(0);
}

static void
generic_opt_check(int command, int options)
{
	int i,j,legal=0;

	/* Check that commands are valid with options.  Complicated by the
	 * fact that if an option is legal with *any* command given, it is
	 * legal overall (ie. -z and -l).
	 */
	for (i = 0; i < NUMBER_OF_OPT; i++) {
		legal = 0; /* -1 => illegal, 1 => legal, 0 => undecided. */

		for (j = 0; j < NUMBER_OF_CMD; j++) {
			if (!(command & (1<<j)))
				continue;

			if (!(options & (1<<i))) {
				if (commands_v_options[j][i] == '+')
					exit_error(PARAMETER_PROBLEM, 
						   "You need to supply the `-%c' "
						   "option for this command\n", 
						   optflags[i]);
			} else {
				if (commands_v_options[j][i] != 'x')
					legal = 1;
				else if (legal == 0)
					legal = -1;
			}
		}
		if (legal == -1)
			exit_error(PARAMETER_PROBLEM,
				   "Illegal option `-%c' with this command\n",
				   optflags[i]);
	}
}

static char
opt2char(int option)
{
	const char *ptr;
	for (ptr = optflags; option > 1; option >>=1, ptr++);

	return *ptr;
}

static char
cmd2char(int option)
{
	const char *ptr;
	for (ptr = cmdflags; option > 1; option >>=1, ptr++);

	return *ptr;
}

static void
add_command(int *cmd, const int newcmd, const int othercmds)
{
	if (invert)
		exit_error(PARAMETER_PROBLEM, "unexpected ! flag");
	if (*cmd & (~othercmds))
		exit_error(PARAMETER_PROBLEM, "Can't use -%c with -%c\n",
			   cmd2char(newcmd), cmd2char(*cmd & (~othercmds)));
	else *cmd |= newcmd;
}
