/* Library which manipulates firewall rules.  Version 0.1. */

/* Architecture of firewall rules is as follows:
 *
 * Chains go INPUT, FORWARD, OUTPUT then user chains.
 * Each user chain starts with an ERROR node.
 * Every chain ends with an unconditional jump: a RETURN for user chains,
 * and a POLICY for built-ins.
 */

/* (C)1998 Paul ``Rusty'' Russell - Placed under the GNU GPL (See
   COPYING for details). */

#include <assert.h>
#include <string.h>
#include <errno.h>
#include <stdlib.h>
#include <stdio.h>

#include "libipfwc.h"

#define IP_VERSION	4
#define IP_OFFSET	0x1FFF

static int sockfd = -1;
static void *ipfwc_fn = NULL;

struct ipfwc_handle
{
	/* Have changes been made? */
	int changed;
	/* Number of rules in here reflects original state. */
	struct ipt_getinfo info;
	unsigned int *counter_map;
	/* Number of rules in here reflects current state. */
	struct ipt_get_entries entries;
};

static ipfwc_handle_t alloc_handle(unsigned int num_rules)
{
	size_t len;
	ipfwc_handle_t h;

	len = sizeof(struct ipfwc_handle) 
		+ num_rules * sizeof(struct ipt_entry)
		+ num_rules * sizeof(unsigned int);

	if (!(h = malloc(len))) {
		errno = ENOMEM;
		return NULL;
	}	
	h->changed = 0;
	h->entries.number = num_rules;
	h->entries.position = 0;
	h->counter_map = (unsigned int *)(&h->entries.entries[num_rules]);

	return h;
}

ipfwc_handle_t ipfwc_init(void)
{
	ipfwc_handle_t h;
	struct ipt_getinfo info;
	size_t len;
	unsigned int i;

	sockfd = socket(AF_INET, SOCK_RAW, IPPROTO_RAW);
	if (sockfd < 0)
		return NULL;

	if (setsockopt(sockfd, IPPROTO_IP, IPT_SO_GET_INFO, &info, 
		       sizeof(info)) < 0)
		return NULL;


	len = sizeof(struct ipt_getinfo) 
		+ sizeof(struct ipt_get_entries)
		+ info.num_entries * sizeof(struct ipt_entry)
		+ info.num_entries * sizeof(unsigned int);

	if (!(h = alloc_handle(info.num_entries)))
		return NULL;

	h->info = info;

	/* If we do nothing, all counters map onto the same rules. */
	for (i = 0; i < info.num_entries; i++)
		h->counter_map[i] = i;

	if (setsockopt(sockfd, IPPROTO_IP, IPT_SO_GET_ENTRIES, 
		       &h->entries, 
		       sizeof(struct ipt_get_entries)
		       + info.num_entries * sizeof(struct ipt_entry)) < 0) {
		free(h);
		return NULL;
	}
	return h;
}

static int
find_label(unsigned int *pos, 
	   const char *name, 
	   const ipfwc_handle_t handle)
{
	unsigned int i;
	const struct ipt_entry *entries = handle->entries.entries;

	if (strcmp(name, IPFWC_LABEL_INPUT) == 0)
		*pos = handle->info.hook_entry[IPT_HOOK_INPUT];
	else if (strcmp(name, IPFWC_LABEL_FORWARD) == 0)
		*pos = handle->info.hook_entry[IPT_HOOK_FORWARD];
	else if (strcmp(name, IPFWC_LABEL_OUTPUT) == 0)
		*pos = handle->info.hook_entry[IPT_HOOK_OUTPUT];
	else {
		for (i = 0; i < handle->entries.number; i++) {
			if (strcmp(entries[i].name, IPT_STANDARD_ERROR) == 0 
			    && strcmp((const char *)&entries[i].action, 
				      name) == 0) {
				*pos = i+1;
				return 1;
			}
		}
		return 0;
	}
	return 1;
}

/* Returns the position of the final (ie. unconditional) element. */
static unsigned int 
get_chain_end(const ipfwc_handle_t handle, unsigned int pos)
{
	unsigned int i;

	/* Terminate when we meet a error label or a hook entry. */
	for (i = pos+1; i < handle->info.num_entries; i++) {
		if (i == handle->info.hook_entry[IPT_HOOK_INPUT]
		    || i == handle->info.hook_entry[IPT_HOOK_FORWARD]
		    || i == handle->info.hook_entry[IPT_HOOK_OUTPUT]
		    || strcmp(handle->entries.entries[i].name,
			      IPT_STANDARD_ERROR) == 0) {
			return i - 1;
		}
	}
	/* SHOULD NEVER HAPPEN */
	fprintf(stderr, "ERROR: Off end of chain!\n");
	return handle->info.num_entries;
}

/* Iterator functions to run through the chains; prev = NULL means
   first chain.  Returns NULL at end. */
const char *
ipfwc_next_chain(const char *prev, ipfwc_handle_t *handle)
{
	unsigned int pos;

	if (!prev)
		pos = 0;
	else {
		find_label(&pos, prev, *handle);
		pos = get_chain_end(*handle, pos) + 1;
	}

	if (pos == (*handle)->info.hook_entry[IPT_HOOK_INPUT])
		return IPFWC_LABEL_INPUT;
	else if (pos == (*handle)->info.hook_entry[IPT_HOOK_FORWARD])
		return IPFWC_LABEL_FORWARD;
	else if (pos == (*handle)->info.hook_entry[IPT_HOOK_OUTPUT])
		return IPFWC_LABEL_OUTPUT;
	else if (pos == (*handle)->entries.number-1)
		return NULL;
	else
		return (const char *)&(*handle)->entries.entries[pos].action;
}

/* How many rules in this chain? */
unsigned int
ipfwc_num_rules(const char *chain, ipfwc_handle_t *handle)
{
	unsigned int pos = 0;

	find_label(&pos, chain, *handle);
	return get_chain_end(*handle, pos) - pos;
}

/* Get n'th rule in this chain. */
const struct ipt_entry *
ipfwc_get_rule(const char *chain, 
	       unsigned int n,
	       ipfwc_handle_t *handle)
{
	unsigned int pos = 0;

	find_label(&pos, chain, *handle);
	return &(*handle)->entries.entries[pos + n];
}

/* Returns a pointer to the target name of this position. */
const char *
ipfwc_get_target(const char *chain, 
		 unsigned int n,
		 ipfwc_handle_t *handle)
{
	unsigned int pos = 0;
	struct ipt_standard *act;

	find_label(&pos, chain, *handle);
	if (strcmp((*handle)->entries.entries[pos + n].name,
		   IPT_STANDARD_ACTION) != 0)
		return (*handle)->entries.entries[pos + n].name;

	/* Standard action: evaluate */
	act = &(*handle)->entries.entries[pos + n].action;
	if (act->verdict < 0) {
		switch (-act->verdict - 1) {
		case IPT_RETURN:
			return IPFWC_LABEL_RETURN;

		case NF_ACCEPT:
			return IPFWC_LABEL_ACCEPT;

		case NF_DROP:
			return IPFWC_LABEL_DROP;

		case NF_USERSPACE:
			return IPFWC_LABEL_USER;

		default:
			return "*UNKNOWN*";
		}
	} else {
		if (act->verdict == pos+n+1) 
			return "";
		else return (const char *)
			     &(*handle)->entries.entries[act->verdict].action;
	}
}

/* Get the policy of a given built-in chain */
const char *ipfwc_get_policy(const char *chain, 
			     struct ipt_counters *counters,
			     ipfwc_handle_t *handle)
{
	unsigned int start, end;

	if (strcmp(chain, IPFWC_LABEL_INPUT) == 0) 
		start = (*handle)->info.hook_entry[IPT_HOOK_INPUT];
	else if (strcmp(chain, IPFWC_LABEL_FORWARD) == 0) 
		start = (*handle)->info.hook_entry[IPT_HOOK_FORWARD];
	else if (strcmp(chain, IPFWC_LABEL_OUTPUT) == 0) 
		start = (*handle)->info.hook_entry[IPT_HOOK_OUTPUT];
	else return NULL;

	end = get_chain_end(*handle, start);
	*counters = (*handle)->entries.entries[end].counters;
	return ipfwc_get_target(chain, end-start, handle);
}

/* If prepend is set, then we are prepending to a chain: if the
 * insertion position is an entry point, keep the entry point. */
static int 
insert_rules(unsigned int num_rules, unsigned int position, int prepend,
	     const struct ipt_entry *insert, ipfwc_handle_t *handle)
{
	ipfwc_handle_t newh;
	unsigned int i;

	if (position >= (*handle)->entries.number) {
		errno = EINVAL;
		return 0;
	}

	newh = alloc_handle((*handle)->entries.number + num_rules);
	if (!newh)
		return 0;

	/* Fix up entry points. */
	for (i = 0; i < 3; i++) {
		if ((*handle)->info.hook_entry[i] > position)
			newh->info.hook_entry[i] 
				= (*handle)->info.hook_entry[i] + num_rules;
		else 
			newh->info.hook_entry[i]
				= (*handle)->info.hook_entry[i];
	}

	newh->info.num_entries = (*handle)->info.num_entries;

	/* Copy pre... */
	memcpy(newh->entries.entries, (*handle)->entries.entries, 
	       sizeof(newh->entries.entries[0]) * position);
	/* ... Insert new ... */
	memcpy(newh->entries.entries + position, insert, 
	       sizeof(newh->entries.entries[0]) * num_rules);
	/* ... copy post */
	memcpy(newh->entries.entries + position + num_rules, 
	       (*handle)->entries.entries + position, 
	       sizeof(newh->entries.entries[0]) 
	       * ((*handle)->info.num_entries - position));

	/* Move counter map. */
	/* Copy pre... */
	memcpy(newh->counter_map, (*handle)->counter_map, 
	       sizeof(unsigned int) * position);
	/* ... copy post */
	memcpy(newh->counter_map + position + num_rules, 
	       (*handle)->counter_map + position, 
	       sizeof(unsigned int)*((*handle)->info.num_entries - position));
	/* Set intermediates to no counter copy */
	for (i = 0; i < num_rules; i++)
		newh->counter_map[position+i] = 0xFFFFFFFF;

	free(*handle);
	*handle = newh;
	(*handle)->changed = 1;
	return 1;
}

static int 
replace_rule(unsigned int position, const struct ipt_entry *newrule,
	     ipfwc_handle_t *handle)
{
	if (position > (*handle)->entries.number) {
		errno = EINVAL;
		return 0;
	}

	/* Replace entry. */
	(*handle)->entries.entries[position] = *newrule;
	/* Counter of old rule no longer valid */
	(*handle)->counter_map[position] = 0xFFFFFFFF;

	(*handle)->changed = 1;
	return 1;
}

static int 
delete_rules(unsigned int position, unsigned int num_rules,
	     ipfwc_handle_t *handle)
{
	unsigned int i;

	if (position + num_rules > (*handle)->entries.number) {
		errno = EINVAL;
		return 0;
	}

	/* Fix up entry points. */
	for (i = 0; i < 3; i++) {
		if ((*handle)->info.hook_entry[i] > position + num_rules)
			(*handle)->info.hook_entry[i] -= num_rules;
		else if ((*handle)->info.hook_entry[i] > position)
			/* Shouldn't really happen. */
			(*handle)->info.hook_entry[i] = position;
	}

	/* Move the rules down. */
	memmove(&(*handle)->entries.entries[position],
		&(*handle)->entries.entries[position + num_rules],
		sizeof((*handle)->entries.entries[position])
		* ((*handle)->entries.number - num_rules));

	/* Move the counter map down. */
	memmove(&(*handle)->counter_map[position],
		&(*handle)->counter_map[position + num_rules],
		sizeof(unsigned int) * ((*handle)->entries.number -num_rules));

	/* Fix number */
	(*handle)->entries.number -= num_rules;

	(*handle)->changed = 1;
	return 1;
}

static int map_target(const char *target, struct ipt_entry *rule,
		      unsigned int pos, const ipfwc_handle_t handle,
		      int postcorrect)
{
	/* Maybe it's empty (=> fall through) */
	if (strlen(target) == 0) {
		strcpy(rule->name, IPT_STANDARD_ACTION);
		rule->action.verdict = pos+1;
	}
	/* Maybe it's a standard action name... */
	else if (strcmp(target, IPFWC_LABEL_ACCEPT) == 0) {
		strcpy(rule->name, IPT_STANDARD_ACTION);
		rule->action.verdict = -NF_ACCEPT - 1;
	}
	else if (strcmp(target, IPFWC_LABEL_DROP) == 0) {
		strcpy(rule->name, IPT_STANDARD_ACTION);
		rule->action.verdict = -NF_DROP - 1;
	}
	else if (strcmp(target, IPFWC_LABEL_USER) == 0) {
		strcpy(rule->name, IPT_STANDARD_ACTION);
		rule->action.verdict = -NF_USERSPACE - 1;
	}
	else if (strcmp(target, IPFWC_LABEL_RETURN) == 0) {
		strcpy(rule->name, IPT_STANDARD_ACTION);
		rule->action.verdict = -IPT_RETURN - 1;
	}
	else if (strcmp(target, IPFWC_LABEL_INPUT) == 0
		 || strcmp(target, IPFWC_LABEL_FORWARD) == 0
		 || strcmp(target, IPFWC_LABEL_OUTPUT) == 0) {
		/* Can't jump to builtins. */
		errno = EINVAL;
		return 0;
	} else {
		/* Maybe it's an existing chain name. */
		unsigned int exists;

		if (find_label(&exists, target, handle)) {
			strcpy(rule->name, IPT_STANDARD_ACTION);
			rule->action.verdict = exists;
			if (exists >= pos)
				rule->action.verdict += postcorrect;
		} else {
			/* Must be a module. */
			if (strlen(target)+1 > sizeof(rule->name)) {
				errno = EINVAL;
				return 0;
			}
			strcpy(rule->name, target);
		}
	}
	return 1;
}

/* Insert the entry `fw' in chain `chain' into position `rulenum'. */
int ipfwc_insert_entry(const ipt_chainlabel chain, 
                       struct ipt_entry *fw, 
		       const char *target,
                       unsigned int rulenum,
		       ipfwc_handle_t *handle)
{
	unsigned int pos;

	ipfwc_fn = ipfwc_insert_entry;
	if (!find_label(&pos, chain, *handle)) {
		errno = ENOENT;
		return 0;
	}

	if (pos + rulenum >= get_chain_end(*handle, pos)) {
		errno = EINVAL;
		return 0;
	}
	if (!map_target(target, fw, pos + rulenum, *handle, +1))
		return 0;

	return insert_rules(1, pos + rulenum, rulenum == 0, fw, handle);
}

/* Atomically replace rule `rulenum' in `chain' with `fw'. */
int ipfwc_replace_entry(const ipt_chainlabel chain, 
			struct ipt_entry *fw, 
			const char *target,
                        unsigned int rulenum,
			ipfwc_handle_t *handle)
{
	unsigned int pos;

	ipfwc_fn = ipfwc_replace_entry;

	if (!find_label(&pos, chain, *handle)) {
		errno = ENOENT;
		return 0;
	}

	if (pos + rulenum >= get_chain_end(*handle, pos)) {
		errno = EINVAL;
		return 0;
	}

	if (!map_target(target, fw, pos + rulenum, *handle, 0))
		return 0;

	return replace_rule(pos + rulenum, fw, handle);
}

/* Append entry `fw' to chain `chain'.  Equivalent to insert with
   rulenum = length of chain. */
int ipfwc_append_entry(const ipt_chainlabel chain, 
		       struct ipt_entry *fw,
		       const char *target,
		       ipfwc_handle_t *handle)
{
	unsigned int pos;

	ipfwc_fn = ipfwc_append_entry;

	if (!find_label(&pos, chain, *handle)) {
		errno = ENOENT;
		return 0;
	}

	pos = get_chain_end(*handle, pos);

	if (!map_target(target, fw, pos, *handle, +1))
		return 0;

	return insert_rules(1, pos, 0, fw, handle);
}	

/* Delete the first rule in `chain' which matches `fw'. */
int ipfwc_delete_entry(const ipt_chainlabel chain, 
		       struct ipt_entry *fw,
		       const char *target,
		       ipfwc_handle_t *handle)
{
	unsigned int pos, lastrule;

	ipfwc_fn = ipfwc_delete_entry;
	if (!find_label(&pos, chain, *handle)) {
		errno = ENOENT;
		return 0;
	}

	lastrule = get_chain_end(*handle, pos);

	for (; pos < lastrule; pos++) {
		if (!map_target(target, fw, pos, *handle, 0))
			return 0;
		/* Ensure counters are the same before comparison. */
		fw->counters = (*handle)->entries.entries[pos].counters;
		if (memcmp(&(*handle)->entries.entries[pos], fw, sizeof(*fw))
		    == 0)
			return delete_rules(pos, 1, handle);
	}

	errno = ENOENT;
	return 0;
}	

/* Delete the rule in position `rulenum' in `chain'. */
int ipfwc_delete_num_entry(const ipt_chainlabel chain, 
			   unsigned int rulenum,
			   ipfwc_handle_t *handle)
{
	unsigned int pos;

	ipfwc_fn = ipfwc_delete_num_entry;
	if (!find_label(&pos, chain, *handle)) {
		errno = ENOENT;
		return 0;
	}

	if (pos + rulenum >= get_chain_end(*handle, pos)) {
		errno = EINVAL;
		return 0;
	}

	return delete_rules(pos + rulenum, 1, handle);
}

/* Check the packet `fw' on chain `chain'.  Returns the verdict, or
   NULL and sets errno. */
const char *ipfwc_check_packet(const ipt_chainlabel chain, 
			       struct ipt_entry *entry,
			       ipfwc_handle_t *handle)
{
	errno = ENOSYS;
	return NULL;
}

/* Flushes the entries in the given chain (ie. empties chain). */
int ipfwc_flush_entries(const ipt_chainlabel chain,
			ipfwc_handle_t *handle)
{
	unsigned int pos;

	ipfwc_fn = ipfwc_flush_entries;
	if (!find_label(&pos, chain, *handle)) {
		errno = ENOENT;
		return 0;
	}

	return delete_rules(pos, get_chain_end(*handle, pos) - pos, handle);
}

/* Zeroes the counters in a chain. */
int ipfwc_zero_entries(const ipt_chainlabel chain,
		       ipfwc_handle_t *handle)
{
	unsigned int i, end;
	
	if (!find_label(&i, chain, *handle)) {
		errno = ENOENT;
		return 0;
	}
	end = get_chain_end(*handle, i);

	/* Negate counters in the chain: when they get added back,
	   they'll cancel the old counters. */
	for (; i <= end; i++) {
		(*handle)->entries.entries[i].counters.pcnt
			= -(*handle)->entries.entries[i].counters.pcnt;
		(*handle)->entries.entries[i].counters.bcnt
			= -(*handle)->entries.entries[i].counters.bcnt;
	}
	(*handle)->changed = 1;

	return 1;
}

/* Creates a new chain. */
int ipfwc_create_chain(const ipt_chainlabel chain,
		       ipfwc_handle_t *handle)
{
	struct ipt_entry newchain[2];
	unsigned int pos;

	ipfwc_fn = ipfwc_create_chain;

	if (find_label(&pos, chain, *handle)) {
		errno = EEXIST;
		return 0;
	}

	/* To create a chain, create two rules: error node and unconditional
	 * return. */
	memset(newchain, 0, sizeof(newchain));
	strcpy(newchain[0].name, IPT_STANDARD_ERROR);
	strcpy((char *)&newchain[0].action, chain);
	newchain[1].action.verdict = IPT_RETURN;

	return insert_rules(2, (*handle)->entries.number - 1, 0, newchain,
			    handle);
}

/* Get the number of references to this chain */
int ipfwc_get_references(unsigned int *ref, const ipt_chainlabel chain,
			 ipfwc_handle_t *handle)
{
	unsigned int i;
	unsigned int pos;

	if (!find_label(&pos, chain, *handle)) {
		errno = ENOENT;
		return 0;
	}

	*ref = 0;
	for (i = 0; i < (*handle)->entries.number; i++) {
		if (strcmp((*handle)->entries.entries[i].name, 
			   IPT_STANDARD_ACTION) == 0
		    && ((*handle)->entries.entries[i].action.verdict == pos))
			(*ref)++;
	}
	return 1;
}

/* Deletes a chain. */
int ipfwc_delete_chain(const ipt_chainlabel chain, 
		       ipfwc_handle_t *handle)
{
	unsigned int pos;
	unsigned int references;

	if (!ipfwc_get_references(&references, chain, handle))
		return 0;

	if (references > 0)
		return ENOTEMPTY;

	ipfwc_fn = ipfwc_flush_entries;
	if (!find_label(&pos, chain, *handle)) {
		errno = ENOENT;
		return 0;
	}

	if (get_chain_end(*handle, pos) != pos + 1) {
		errno = ENOTEMPTY;
		return 0;
	}

	return delete_rules(pos - 1, 2, handle);
}

/* Sets the policy on a built-in chain. */
int ipfwc_set_policy(const ipt_chainlabel chain, 
		     const ipt_chainlabel policy,
		     ipfwc_handle_t *handle)
{
	unsigned int hook;
	unsigned int pos;

	/* Figure out which chain. */
	if (strcmp(chain, IPFWC_LABEL_INPUT) == 0)
		hook = IPT_HOOK_INPUT;
	else if (strcmp(chain, IPFWC_LABEL_FORWARD) == 0)
		hook = IPT_HOOK_FORWARD;
	else if (strcmp(chain, IPFWC_LABEL_OUTPUT) == 0)
		hook = IPT_HOOK_OUTPUT;
	else {
		errno = EINVAL;
		return 0;
	}

	pos = get_chain_end(*handle, (*handle)->info.hook_entry[hook]);
	if (strcmp(policy, IPFWC_LABEL_ACCEPT) == 0)
		(*handle)->entries.entries[pos].action.verdict = NF_ACCEPT;
	else if (strcmp(policy, IPFWC_LABEL_DROP) == 0)
		(*handle)->entries.entries[pos].action.verdict = NF_DROP;
	else if (strcmp(policy, IPFWC_LABEL_USER) == 0)
		(*handle)->entries.entries[pos].action.verdict = NF_USERSPACE;
	else {
		errno = EINVAL;
		return 0;
	}
	(*handle)->counter_map[pos] = 0xFFFFFFFF;
	return 1;
}

int ipfwc_commit(ipfwc_handle_t *handle)
{
	/* Replace, then map back the counters. */
	struct ipt_replace repl;
	struct ipt_counters_info *newcounters;
	unsigned int i;
	size_t counterlen = sizeof(struct ipt_counters_info) 
			     + (sizeof(struct ipt_counters) 
				* (*handle)->entries.number);

	/* Don't commit if nothing changed. */
	if (!(*handle)->changed)
		goto finished;

	memcpy(repl.hook_entry, (*handle)->info.hook_entry,
	       sizeof(repl.hook_entry));
	/* The error node gets appended automatically. */
	repl.number = (*handle)->entries.number-1;
	repl.entries = (*handle)->entries.entries;
	repl.num_counters = (*handle)->info.num_entries;

	repl.counters
		= malloc(sizeof(struct ipt_counters) * repl.num_counters);
	if (!repl.counters) {
		errno = ENOMEM;
		return 0;
	}

	newcounters = malloc(counterlen);
	if (!newcounters) {
		free(repl.counters);
		errno = ENOMEM;
		return 0;
	}

	/* Zero out old counters inside rule. */
	for (i = 0; i < repl.number; i++) 
		memset(&repl.entries[i].counters, 0, 
		       sizeof(repl.entries[i].counters));

	if (setsockopt(sockfd, IPPROTO_IP, IPT_SO_REPLACE, &repl, sizeof(repl))
	    < 0) {
		free(repl.counters);
		free(newcounters);
		return 0;
	}

	/* Put counters back. */
	newcounters->position = 0;
	newcounters->number = (*handle)->entries.number;

	for (i = 0; i < newcounters->number; i++) {
		if ((*handle)->counter_map[i] == 0xFFFFFFFF) 
			newcounters->counters[i] 
				= ((struct ipt_counters){ 0, 0 });
		else {
			newcounters->counters[i] 
				= repl.counters[(*handle)->counter_map[i]];
		}
	}

	if (setsockopt(sockfd, IPPROTO_IP, IPT_SO_ADD_COUNTERS, newcounters,
		       counterlen) < 0) {
		free(repl.counters);
		free(newcounters);
		return 0;
	}

	free(repl.counters);
	free(newcounters);
 finished:
	free(*handle);
	*handle = NULL;
	return 1;
}

/* Get raw socket. */
int ipfwc_get_raw_socket()
{
	return sockfd;
}

/* Translates errno numbers into more human-readable form than strerror. */
const char *ipfwc_strerror(int err)
{
	unsigned int i;
	struct table_struct {
		void *fn;
		int err;
		const char *message;
	} table [] = 
	  { { NULL, ENOENT, "No chain by that name" },
	    { NULL, 0, "Incompatible with this kernel" },
	    { NULL, ENOSYS, "Will be implemented real soon.  I promise." },
	    { ipfwc_init, EPERM, "Permission denied (you must be root)" },
	    { ipfwc_delete_chain, ENOTEMPTY, "Chain is not empty" },
	    { ipfwc_create_chain, EEXIST, "Chain already exists" },
	    /* EINVAL for CHECK probably means bad interface. */
	    { ipfwc_check_packet, EINVAL, 
	      "bad arguments (does that interface exist?)" },
	    /* EINVAL for DELETE probably means no matching rule */
	    { ipfwc_delete_entry, EINVAL,
	      "bad rule (does a matching rule exist in that chain?)" },
	    { ipfwc_insert_entry, EINVAL,
	      "bad rule (does a matching rule exist in that chain?)" },
	    { ipfwc_delete_num_entry, EINVAL,
	      "bad rule (does a matching rule exist in that chain?)" },
	    { ipfwc_replace_entry, EINVAL,
	      "bad rule (does a matching rule exist in that chain?)" }
	  };

	for (i = 0; i < sizeof(table)/sizeof(struct table_struct); i++) {
		if ((!table[i].fn || table[i].fn == ipfwc_fn)
		    && table[i].err == err)
			return table[i].message;
	}
	return strerror(err);
}
