#ifndef _LIBIPFWC_H
#define _LIBIPFWC_H
/* Library which manipulates firewall rules.  Version 0.2. */

#include "ipfwc_kernel_headers.h"
#include "ip_tables.h"
#include "ip_tables_standard.h"

typedef ipt_errorlabel ipt_chainlabel;

#define IPFWC_LABEL_INPUT   "INPUT"
#define IPFWC_LABEL_FORWARD "FORWARD"
#define IPFWC_LABEL_OUTPUT  "OUTPUT"
#define IPFWC_LABEL_ACCEPT  "ACCEPT"
#define IPFWC_LABEL_DROP    "DROP"
#define IPFWC_LABEL_USER    "USER"
#define IPFWC_LABEL_RETURN  "RETURN"

/* Transparent handle type. */
typedef struct ipfwc_handle *ipfwc_handle_t;

/* Take a snapshot of the rules.  Returns NULL on error. */
ipfwc_handle_t ipfwc_init(void);

/* Iterator functions to run through the chains; prev = NULL means
   first chain.  Returns NULL at end. */
const char *ipfwc_next_chain(const char *prev, ipfwc_handle_t *handle);

/* How many rules in this chain? */
unsigned int ipfwc_num_rules(const char *chain, ipfwc_handle_t *handle);

/* Get n'th rule in this chain. */
const struct ipt_entry *ipfwc_get_rule(const char *chain, 
				       unsigned int n,
				       ipfwc_handle_t *handle);

/* Returns a pointer to the target name of this position. */
const char *ipfwc_get_target(const char *chain, 
			     unsigned int n,
			     ipfwc_handle_t *handle);

/* Get the policy of a given built-in chain */
const char *ipfwc_get_policy(const char *chain, 
			     struct ipt_counters *counters,
			     ipfwc_handle_t *handle);

/* These functions return TRUE for OK or 0 and set errno.  If errno ==
   0, it means there was a version error (ie. upgrade libipfwc). */
/* Rule numbers start at 1 for the first rule. */

/* Insert the entry `fw' in chain `chain' into position `rulenum'. */
int ipfwc_insert_entry(const ipt_chainlabel chain, 
                       struct ipt_entry *fw, 
		       const char *target,
                       unsigned int rulenum,
		       ipfwc_handle_t *handle);

/* Atomically replace rule `rulenum' in `chain' with `fw'. */
int ipfwc_replace_entry(const ipt_chainlabel chain, 
			struct ipt_entry *fw, 
			const char *target,
                        unsigned int rulenum,
			ipfwc_handle_t *handle);

/* Append entry `fw' to chain `chain'.  Equivalent to insert with
   rulenum = length of chain. */
int ipfwc_append_entry(const ipt_chainlabel chain, 
		       struct ipt_entry *fw,
		       const char *target,
		       ipfwc_handle_t *handle);

/* Delete the first rule in `chain' which matches `fw'. */
int ipfwc_delete_entry(const ipt_chainlabel chain, 
		       struct ipt_entry *fw,
		       const char *target,
		       ipfwc_handle_t *handle);

/* Delete the rule in position `rulenum' in `chain'. */
int ipfwc_delete_num_entry(const ipt_chainlabel chain, 
			   unsigned int rulenum,
			   ipfwc_handle_t *handle);

/* Check the packet `fw' on chain `chain'.  Returns the verdict, or
   NULL and sets errno. */
const char *ipfwc_check_packet(const ipt_chainlabel chain, 
			       struct ipt_entry *,
			       ipfwc_handle_t *handle);

/* Flushes the entries in the given chain (ie. empties chain). */
int ipfwc_flush_entries(const ipt_chainlabel chain,
			ipfwc_handle_t *handle);

/* Zeroes the counters in a chain. */
int ipfwc_zero_entries(const ipt_chainlabel chain,
		       ipfwc_handle_t *handle);

/* Creates a new chain. */
int ipfwc_create_chain(const ipt_chainlabel chain,
		       ipfwc_handle_t *handle);

/* Deletes a chain. */
int ipfwc_delete_chain(const ipt_chainlabel chain,
		       ipfwc_handle_t *handle);

/* Sets the policy on a built-in chain. */
int ipfwc_set_policy(const ipt_chainlabel chain, 
		     const ipt_chainlabel policy,
		     ipfwc_handle_t *handle);

/* Get the number of references to this chain */
int ipfwc_get_references(unsigned int *ref, const ipt_chainlabel chain,
			 ipfwc_handle_t *handle);

/* Makes the actual changes. */
int ipfwc_commit(ipfwc_handle_t *handle);

/* Get raw socket. */
int ipfwc_get_raw_socket();

/* Translates errno numbers into more human-readable form than strerror. */
const char *ipfwc_strerror(int err);
#endif /* _LIBIPFWC_H */
