/* Test program for iptables. */
#include <netinet/ip.h>
#include <netinet/ip_icmp.h>
#include <netinet/tcp.h>
#include <netinet/udp.h>
#include <net/if.h>
#include <sys/types.h>

#include <linux/ip_tables.h>
#include <linux/ip_tables_standard.h>
#include <sys/socket.h>
#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <stdlib.h>

static void print_standard(unsigned int pos, void *info)
{
	struct ipt_standard *stand = info;
	unsigned int i;

	for (i = 0; i < sizeof(struct ipt_standard); i++) {
		printf("%02X ", ((unsigned char *)stand)[i]);
	}
	printf("\n");

	printf("  Verdict: %s.\n", 
	       stand->verdict == NF_DROP ? "DROP" :
	       (stand->verdict == NF_ACCEPT ? "ACCEPT" :
		(stand->verdict == NF_USERSPACE ? "USERSPACE" :
		 (stand->verdict == IPT_RETURN ? "RETURN" :
		  (stand->verdict != IPT_STANDARD_JUMP ? "UNKNOWN" :
		   (stand->relative_pos == 1 ? "-" : "JUMP"))))));
	if (stand->verdict == IPT_STANDARD_JUMP
	    && stand->relative_pos != 1)
		printf("    To %i.\n", (int)pos+stand->relative_pos);
	if (stand->flags == IPT_STANDARD_MARKABS)
		printf("  Set Mark to %08X.\n", stand->mark);
	else if (stand->mark)
		printf("  OR Mark with %08X.\n", stand->mark);
}

static void print_ip(__u32 ipaddr)
{
	printf("%u.%u.%u.%u", 
	       ntohl(ipaddr)>>24 & 0xFF,
	       ntohl(ipaddr)>>16 & 0xFF,
	       ntohl(ipaddr)>>8 & 0xFF,
	       ntohl(ipaddr));
}

static void print_rule(const struct ipt_ip *ip, const void *voidproto)
{
	printf("src = %s", ip->invflags & IPT_INV_SRCIP ? "!" : "");
	if (ip->smsk.s_addr) {
		print_ip(ip->src.s_addr);
		printf("/");
		print_ip(ip->smsk.s_addr);
	}
	else printf("ANY");
	printf(" dst = %s", ip->invflags & IPT_INV_DSTIP ? "!" : "");
	if (ip->dmsk.s_addr) {
		print_ip(ip->dst.s_addr);
		printf("/");
		print_ip(ip->dmsk.s_addr);
	}
	else printf("ANY");

	printf(" in = %s", ip->invflags & IPT_INV_VIAPID_IN ? "!" : "");
	if (ip->in.ifname[0] || !(ip->flags&IPT_F_EXACT_IN)) 
		printf("%s%s ", ip->in.ifname, 
		       ip->flags&IPT_F_EXACT_IN ? "+" : "");
	else printf("%u ", (unsigned int)ip->in.pid);

	printf("out = %s", ip->invflags & IPT_INV_VIAPID_OUT ? "!" : "");
	if (ip->out.ifname[0] || !(ip->flags&IPT_F_EXACT_OUT))
		printf("%s%s ", ip->out.ifname,
		       ip->flags&IPT_F_EXACT_OUT ? "+" : "");
	else printf("%u ", (unsigned int)ip->out.pid);

	if (ip->flags & IPT_F_TOS) {
		printf("TOS = %s%X ", 
		       ip->invflags & IPT_INV_TOS ? "!" : "", ip->tos);
	}
	if (ip->flags & IPT_F_FRAG)
		printf("%s<frag> ", ip->invflags & IPT_INV_FRAG ? "!" : "");

	printf("%s", ip->invflags & IPT_INV_PROTO ? "!" : "");
	switch (ip->proto) {
	case 0:
		printf("protocol ANY ");
		break;
	case IPPROTO_TCP: {
		const struct ipt_tcp *tcp = voidproto;
		printf("protocol TCP ");
		printf("source ports %s%u-%u ",
		       tcp->flags & IPT_TCP_INV_SRCPT ? "!" : "",
		       tcp->spts[0], tcp->spts[1]);
		printf("dest ports %s%u-%u ",
		       tcp->flags & IPT_TCP_INV_DSTPT ? "!" : "",
		       tcp->dpts[0], tcp->dpts[1]);
		if (tcp->flg_mask)
			printf("TCP flags = %X/%X ",
			       tcp->flg_cmp, tcp->flg_mask);
		break;
	}
	case IPPROTO_UDP: {
		const struct ipt_udp *udp = voidproto;
		printf("protocol UDP ");
		printf("source ports %s%u-%u ",
		       udp->flags & IPT_UDP_INV_SRCPT ? "!" : "",
		       udp->spts[0], udp->spts[1]);
		printf("dest ports %s%u-%u ",
		       udp->flags & IPT_UDP_INV_DSTPT ? "!" : "",
		       udp->dpts[0], udp->dpts[1]);
		break;
	}
	case IPPROTO_ICMP: {
		const struct ipt_icmp *icmp = voidproto;
		printf("protocol ICMP ");
		printf("%s type %u-%u ",
		       icmp->flags & IPT_ICMP_INV ? "!" : "",
		       icmp->type[0], icmp->type[1]);
		if (icmp->code[0] != 0 || icmp->code[1] != 0xFF) 
			printf("code %u-%u ", icmp->code[0], icmp->code[1]);
		break;
	}
	default:
		printf("protocol %u ", ip->proto);
		break;
	}
	printf("\n");
}

int main(int argc, const char *argv[])
{
	int sockfd = socket(AF_INET, SOCK_RAW, IPPROTO_RAW);

	if (sockfd < 0) {
		perror("opening socket");
		return 1;
	}

	if (argv[1] && strcmp(argv[1], "show") == 0) {
		struct ipt_getinfo info;
		unsigned int i;
		struct ipt_get_entries *entries;
		size_t len;

		if (setsockopt(sockfd, IPPROTO_IP, IPT_SO_GET_INFO, &info, 
			       sizeof(info)) < 0) {
			perror("setsockopt IPT_SO_GET_INFO");
			return 1;
		}

		printf("Number of entries: %u.\n", info.num_entries);
		for (i = 0; i < NF_IP_NUMHOOKS; i++)
			printf("Hook %u = %u.\n", i, info.hook_entry[i]);

		len = sizeof(struct ipt_get_entries)
			+ (info.num_entries * sizeof(struct ipt_entry));

		entries = malloc(len);
		entries->number = info.num_entries;
		entries->position = 0;
		if (setsockopt(sockfd, IPPROTO_IP, IPT_SO_GET_ENTRIES, 
			       entries, len) < 0) {
			perror("setsockopt IPT_SO_GET_ENTRIES");
			return 1;
		}
		for (i = 0; i < entries->number; i++) {
			struct ipt_entry *e = &entries->entries[i];
			printf("Rule %u: action=%s %llu p, %llu b.\n",
			       i, e->name[0] ? e->name : "standard",
			       e->counters.pcnt, e->counters.bcnt);
			print_rule(&e->ip, &e->proto);
			if (strcmp(e->name, "") == 0) 
				print_standard(i, &e->action);
		}
	}
	else if (argv[1] && strcmp(argv[1], "icmp-user") == 0) {
		struct ipt_entry icmp;
		struct ipt_standard *info = &icmp.action;
		struct ipt_replace replace = { { 0, 1, 0, 0, 0, 0 }, 
					       1, 1, &icmp, 
					       0, NULL };

		icmp.ip = ((struct ipt_ip){ { 0 }, { 0 }, { 0 }, { 0 },
					    { 0 }, { 0 }, IPPROTO_ICMP,
					    0, 0 });
		icmp.proto.icmp = ((struct ipt_icmp){ { 0, 0xFF }, {0, 0xFF },
						      0 });
		icmp.name[0] = '\0';
		icmp.counters.bcnt = icmp.counters.pcnt = 0;
		info->verdict = NF_USERSPACE;
		info->relative_pos = 0;
		info->mark = 0;
		info->flags = 0;

		if (setsockopt(sockfd, IPPROTO_IP, IPT_SO_REPLACE, &replace,
			       sizeof(replace)) < 0) {
			perror("setsockopt IPT_SO_REPLACE");
			return 1;
		}
	}
	else if (argv[1] && strcmp(argv[1], "icmp-drop") == 0) {
		struct ipt_entry icmp;
		struct ipt_standard *info = &icmp.action;
		struct ipt_replace replace = { { 1, 0, 0, 0, 0, 0 }, 
					       1, 1, &icmp, 
					       0, NULL };

		icmp.ip = ((struct ipt_ip){ { 0 }, { 0 }, { 0 }, { 0 },
					    { 0 }, { 0 }, IPPROTO_ICMP,
					    0, 0 });
		icmp.proto.icmp = ((struct ipt_icmp){ { 0, 0xFF }, {0, 0xFF },
						      0 });

		icmp.ip = ((struct ipt_ip){ { 0 }, { 0 }, { 0 }, { 0 },
					    { 0 }, { 0 }, IPPROTO_ICMP,
					    0, 0 });
		icmp.proto.icmp = ((struct ipt_icmp){ { 0, 0xFF }, {0, 0xFF },
						       0 });
		icmp.counters.bcnt = icmp.counters.pcnt = 0;
		icmp.name[0] = '\0';
		info->verdict = NF_DROP;
		info->relative_pos = 0;
		info->mark = 0;
		info->flags = 0;

		if (setsockopt(sockfd, IPPROTO_IP, IPT_SO_REPLACE, &replace,
			       sizeof(replace)) < 0) {
			perror("setsockopt IPT_SO_REPLACE");
			return 1;
		}
	}
	else {
		fprintf(stderr, "unknown args\n");
		return 1;
	}
	
	return 0;
}
